/*
 * Romeo_Wrap.cpp : Wrapper between Move3D and HPP that provides an interface for handling Romeo
 *
 * Author : renaud Viry (renaud.viry@laas.fr)
 * Copyright LAAS-CNRS 2014
 *
 */

#include "Wrap.hpp"
#include <math.h>

#ifdef HPP
#include <hpp/core/basic-configuration-shooter.hh>
 
#define TIME_OUT 0.5 // TimeOut for Ik sol in seconds

Wrap::Wrap()
{
}

bool Wrap::loadRobot(const std::string& robotName, const std::string& rootJointType,
			   		 const std::string& packageName, const std::string& modelName,
			         const std::string& urdfSuffix, const std::string& srdfSuffix,
			         Robot *move3dRobot)
{
	hpp::model::DevicePtr_t robot = hpp::model::Device::create (robotName);
	if (!robot)
	{
		std::cout << "ERROR : creating Robot failed" << std::endl;
		return false;
	}

	hpp::model::urdf::loadRobotModel (robot, rootJointType, packageName, modelName, urdfSuffix, srdfSuffix);

	_manipulationSolver = new hpp::core::ManipulationSolver();
	if (!_manipulationSolver)
	{
		std::cout << "ERROR : loading Robot model failed" << std::endl;
		return false;
	}

	_manipulationSolver->robot(robot);

	_move3dRobot = move3dRobot;

	return true;
}

void Wrap::addLockDof(const std::string& jointName, double value)
{
//	if (!_manipulationSolver->getLockedDof(name))	
//	{
		hpp::model::JointPtr_t joint = _manipulationSolver->robot()->getJointByName (jointName);

		hpp::core::LockedDofPtr_t lockedDof (hpp::core::LockedDof::create (jointName + std::string("_constraint"), joint,
							       value,
							       0,
							       0));
		_manipulationSolver->addConstraint (lockedDof);
//	}
}

void Wrap::addPositionConstraint(const std::string& name, const std::string& jointName,
									  hpp::model::vector3_t posLocal, hpp::model::vector3_t posGlobal,
									  std::vector<bool> mask)
{
	hpp::model::JointPtr_t joint = _manipulationSolver->robot()->getJointByName (jointName);
	
	if (_manipulationSolver->getPositionConstraint(name))
		updatePositionConstraintRef(name, posLocal, posGlobal);
	else
	{
		_manipulationSolver->addPositionConstraint(name, hpp::constraints::Position::create(_manipulationSolver->robot(),
																						joint,
																						posLocal,
																						posGlobal,
																						hpp::model::matrix3_t::getIdentity (),
																						boost::assign::list_of (mask[0])(mask[1])(mask[2])));
		_manipulationSolver->addConstraintToConfigProjector(name, _manipulationSolver->numericalConstraint(name));
	}
}

void Wrap::addOrientationConstraint(const std::string& name, const std::string& jointName,
										 hpp::model::matrix3_t rotation, std::vector<bool> mask)
{
	hpp::model::JointPtr_t joint = _manipulationSolver->robot()->getJointByName (jointName);

	if (_manipulationSolver->getOrientationConstraint(name))
		updateOrientationConstraintRef(name, rotation);
	else
	{
		_manipulationSolver->addOrientationConstraint(name, hpp::constraints::Orientation::create(_manipulationSolver->robot(),
																						joint,
																						rotation,
																						boost::assign::list_of (mask[0])(mask[1])(mask[2])));
		_manipulationSolver->addConstraintToConfigProjector(name, _manipulationSolver->numericalConstraint(name));
	}
}

bool Wrap::applyConstraints(hpp::core::ConfigurationPtr_t config, double *timer)
{
	struct timeval start, end;
    gettimeofday(&start, NULL);
	bool success = _manipulationSolver->constraints()->apply(*config);
	gettimeofday(&end, NULL);
	double timeOut = (double)(((end.tv_sec - start.tv_sec) * 1000. + (end.tv_usec - start.tv_usec) / 1000.) / 1000.);
	while (!success && (timeOut < TIME_OUT))
	{
		gettimeofday(&start, NULL);
		success = _manipulationSolver->constraints()->apply(*config);
		gettimeofday(&end, NULL);
		timeOut += (double)(((end.tv_sec - start.tv_sec) * 1000. + (end.tv_usec - start.tv_usec) / 1000.) / 1000.);
	}

	*timer = timeOut;
	return success;
}

void Wrap::resetConstraints()
{
	_manipulationSolver->resetConstraints();
}

void Wrap::updatePositionConstraintRef(const std::string& constraintName, hpp::model::vector3_t globalValue)
{
	_manipulationSolver->getPositionConstraint(constraintName)->reference(globalValue);
}

void Wrap::updatePositionConstraintRef(const std::string& constraintName, hpp::model::vector3_t localValue, hpp::model::vector3_t globalValue)
{
	_manipulationSolver->getPositionConstraint(constraintName)->reference(localValue, globalValue);
}

void Wrap::updateOrientationConstraintRef(const std::string& constraintName, hpp::model::matrix3_t rot)
{
	_manipulationSolver->getOrientationConstraint(constraintName)->reference(rot);
}

bool Wrap::generateOpenGraspConf(confPtr_t inputConf, confPtr_t outputConf)
{
	int index;
	if (!inputConf)
	{
		std::cout << "ERROR : solution not found" << std::endl;
		return false;
	}

	hpp::model::JointVector_t jointVector = _manipulationSolver->robot()->getJointVector ();
	for (int i = 0; i < inputConf->getRobot()->getNumberOfJoints(); i++)
	{
		outputConf->getConfigStruct()[i] = inputConf->getConfigStruct()[i];
		if (jointVector[i]->name() == _eef_joint)
			index = jointVector[i]->rankInConfiguration();
	}
	for (int j = 0; j < _grasp.openConfig.size(); j++)
	{
		outputConf->getConfigStruct()[j + index + 6] = _grasp.openConfig[j];
	}

	return true;
}

bool Wrap::generateClosedGraspConf(confPtr_t inputConf, confPtr_t outputConf)
{
	int index;
	if (!inputConf)
	{
		std::cout << "ERROR : solution not found" << std::endl;
		return false;
	}

	hpp::model::JointVector_t jointVector = _manipulationSolver->robot()->getJointVector();
	for (int i = 0; i < jointVector.size(); i++)
	{
		outputConf->getConfigStruct()[i] = inputConf->getConfigStruct()[i];
		if (jointVector[i]->name() == _eef_joint)
			index = jointVector[i]->rankInConfiguration();
	}
	for (int j = 0; j < _grasp.config.size(); j++)
	{
		outputConf->getConfigStruct()[j + index + 6] = _grasp.config[j];
	}

	return true;
}

confPtr_t Wrap::createConfiguration(hpp::core::ConfigurationPtr_t config)
{
	// In HPP, the root joint is defined as (X, Y, Z, quaternion), so we do not consider the 7 first value to be copied 
	// In Move3D, there is a "world joint" defined as (X, Y, Z, Rx, Ry, Rz), so we do not consider the 6 first value to be set
	// However, we convert the quaternion into a rotation (Rx, Ry, Rz) and vice-versa

	if (!_move3dRobot)
	{
		std::cout << "ERROR : move3d robot is not defined in Wrapper" << std::endl;
		return std::tr1::shared_ptr<Configuration>() ;
	}

	std::tr1::shared_ptr<Configuration> curConf = _move3dRobot->getCurrentPos();

	curConf->getConfigStruct()[6] = (*config)[0]; // X
	curConf->getConfigStruct()[7] = (*config)[1]; // Y
	curConf->getConfigStruct()[8] = (*config)[2]; // Z

	p3d_matrix3 rot;
	p3d_matrix4 M;
	p3d_vector4 quat;
	double Tx, Ty, Tz, Rx, Ry, Rz;
	quat[0] = (*config)[3];
	quat[1] = (*config)[4];
	quat[2] = (*config)[5];
	quat[3] = (*config)[6];
	p3d_quaternion_to_matrix3(quat, rot);
	p3d_mat4SetRotMatrix(rot, M);
	p3d_mat4ExtractPosReverseOrder(M, &Tx, &Ty, &Tz, &Rx, &Ry, &Rz);
	curConf->getConfigStruct()[9]  = Rx;
	curConf->getConfigStruct()[10] = Ry;
	curConf->getConfigStruct()[11] = Rz;

	// As per initial comment, instead of starting at index 6 and do +6 on Move3d side and +1 on HPP side
	// We start at index 7 and do only a +5 on Move3d side
	for (int i = 7; i < config->size(); i++)
		curConf->getConfigStruct()[i + 5] = (*config)[i];

	confPtr_t q = std::tr1::shared_ptr<Configuration>(new Configuration(*curConf));

	return q;
}

hpp::core::ConfigurationPtr_t Wrap::createHppConfiguration(confPtr_t q)
{
	// See comment in createConfiguration for details of index offset
	hpp::core::ConfigurationPtr_t result (new hpp::core::Configuration_t (_manipulationSolver->robot()->configSize()));

	p3d_matrix4 M;
	p3d_matrix3 rot;
	p3d_vector4 quat;
	p3d_mat4Pos(M, (*q)[6], (*q)[7], (*q)[8], (*q)[9], (*q)[10], (*q)[11]);
	p3d_mat4ExtractRotMatrix(M, rot);

	p3d_matrix3_to_quaternion(rot, quat);
	(*result)[3] = quat[0];
	(*result)[4] = quat[1];
	(*result)[5] = quat[2];
	(*result)[6] = quat[3];

	(*result)[0] = (*q)[6]; // X
	(*result)[1] = (*q)[7]; // Y
	(*result)[2] = (*q)[8]; // Z

	for (int i = 7; i < result->size(); i++)
	{
		(*result)[i] = (*q)[i + 5];
	}

	return result;
}

double Wrap::getJointDofValue(const std::string& jointName)
{
	double value = 0.0;

	hpp::model::JointVector_t joints = _manipulationSolver->robot()->getJointVector();
	for (int i = 0; i < joints.size(); i++)
	{
		if (joints[i]->name() == jointName)
			return _manipulationSolver->robot()->currentConfiguration()[joints[i]->rankInConfiguration()];
	}
	return value;
}

bool Wrap::findTrajectory(confPtr_t &initConf, confPtr_t &goalConf, p3d_traj *outputTraj, double *timer)
{
	// Set the start config to the problem solver
	hpp::core::ConfigurationPtr_t initConfig = createHppConfiguration(initConf);
	_manipulationSolver->initConfig(initConfig);

	// Set the goal config to the problem solver
	hpp::core::ConfigurationPtr_t goalConfig = createHppConfiguration(goalConf);
	_manipulationSolver->resetGoalConfigs();
	_manipulationSolver->addGoalConfig(goalConfig);

	struct timeval start, end;
	gettimeofday(&start, NULL);

	_manipulationSolver->solve();

	gettimeofday(&end, NULL);
	double timeOut = (double)(((end.tv_sec - start.tv_sec) * 1000. + (end.tv_usec - start.tv_usec) / 1000.) / 1000.);
	*timer = timeOut;

	hpp::core::PathVectors_t paths = _manipulationSolver->paths();
	if (paths.size() == 0)
	{
		std::cout << "ERROR : problem has no solution !" << std::endl;
		return false;
	}

	hpp::core::PathPtr_t path = _manipulationSolver->paths()[0];

	return createTrajectory(path, outputTraj); // pathId 0 for result of RRT
}

bool Wrap::createTrajectory(hpp::core::PathPtr_t inputPath, p3d_traj *outputTrajectory)
{
	double pathLength = inputPath->length();
	double step = 0.01;
	double t = 0;
	int i = 0;

	outputTrajectory = p3d_create_empty_trajectory(_move3dRobot->getRobotStruct());

	outputTrajectory->nlp = pathLength/step;
	// std::cout << "RV -- number of localpath in traj : " << outputTrajectory->nlp << std::endl;

	p3d_localpath *prev, *next, *last, *first;

	while (t < pathLength)
	{
		// std::cout << "RV -- Step " << t/step << std::endl;
		hpp::core::Configuration_t start_config = (*inputPath) (t);
		hpp::core::ConfigurationPtr_t start_conf = boost::shared_ptr<hpp::model::Configuration_t> (new hpp::model::Configuration_t(start_config));
		confPtr_t q_start = createConfiguration(start_conf);
		// _move3dRobot->setAndUpdate(*q_start);
		// g3d_draw_allwin_active();
		// sleep(1);

		hpp::core::Configuration_t goal_config = (*inputPath) (t+step);
		hpp::core::ConfigurationPtr_t goal_conf = boost::shared_ptr<hpp::model::Configuration_t> (new hpp::model::Configuration_t(goal_config));
		confPtr_t q_goal = createConfiguration(goal_conf);
		// _move3dRobot->setAndUpdate(*q_goal);
		// g3d_draw_allwin_active();
		// sleep(1);

		// Create a Move3D LocalPath and set q in it
		p3d_localpath *lp;
		lp->type_lp = LINEAR;
		pp3d_lin_data data;
		data->q_init = q_start->getConfigStruct();
		data->q_end = q_goal->getConfigStruct();
		p3d_lm_specific spec;
		spec.lin_data = data;
		lp->specific = spec;
		lp->prev_lp = prev;
		lp->lp_id = i++;

		prev = lp;
		last = lp;

		t += step;
	}

	last->next_lp = NULL;
	next = last;
	first = last;

	while (t >= 0)
	{
		p3d_localpath *lp;
		lp = next->prev_lp;
		lp->next_lp = next;

		next = lp;
		first = lp;

		t -= step;
	}
	first->prev_lp = NULL;

	outputTrajectory->courbePt = first;

	return true;
}

void Wrap::randomizeFullConfig(hpp::core::ConfigurationPtr_t config)
{
	hpp::core::BasicConfigurationShooter shooter (_manipulationSolver->robot());
	hpp::core::ConfigurationPtr_t configuration = shooter.shoot();
	for (int i = 0; i < configuration->size(); i++)
		(*config)[i] = (*configuration)[i];
}

void Wrap::setJointBound(std::string jointName, std::vector<double> bounds)
{
	hpp::model::JointPtr_t joint = _manipulationSolver->robot()->getJointByName(jointName);
	if (bounds.size() % 2 != 0)
	{
		std::cout << "ERROR : bounds should be specified as pairs" << std::endl;
		return;
	}
	for (int i = 0; i < bounds.size() / 2; i++)
	{
		joint->isBounded (i, true);
		joint->lowerBound(i, bounds[i*2]);
		joint->upperBound(i, bounds[i*2 + 1]);
	}
}

#endif /* HPP */
