#ifndef SOLUTIONDATA_H
#define SOLUTIONDATA_H

#include <string>
#include <map>
#include <jsoncpp/json/value.h>
#include "Logging/Logger.h"




/**
 * Info storage class.
 * An instance of SolutionData is used to cache the data retrived during computation (especially of GTP)
 * It stores the data on maps, it has three different basic maps -> int, double, and string maps,
 * and a recursive one of pointers on SolutionData.
 * To save and entry, use setMapEntry, and to retrieve it get<>MapEntry.
 */
class SolutionData
{
    MOVE3D_STATIC_LOGGER;
public:
    SolutionData();
    SolutionData(SolutionData* SD);
    ~SolutionData();

    /**
     * copy the data of a solutionData pointer into this one
     * @param SD: the solution data to copy
     */
    void copySolutionData(SolutionData* SD);



    /**
     * @return the int map
     */
    const std::map<std::string, int> &getIntMap();

    /**
     * @return the double map
     */
    const std::map<std::string, double> &getDoubleMap();

    /**
     * @return the string map
     */
    const std::map<std::string, std::string> &getStringMap();

    /**
     * @return the solution data recursive map
     */
    const std::map<std::string, SolutionData*> &getSDMap();



    /**
     * transform the string map into a multimap of strings
     * @return string multimap
     */
    std::multimap<std::string, std::string> getStringAsMultimapMap();



    /**
     * @param key: the key of the int value val
     * @param val: the value to store in the map
     */
    void setMapEntry(std::string key, int val);


    /**
     * @param key: the key of the double value val
     * @param val: the value to store in the map
     */
    void setMapEntry(std::string key, double val);

    /**
     * @param key: the key of the string value val
     * @param val: the value to store in the map
     */
    void setMapEntry(std::string key, std::string val);

    /**
     * @param key: the key of the SolutionData* value val
     * @param val: the value to store in the map
     */
    void setMapEntry(std::string key, SolutionData* val);



    /**
     * search in the int (first) and the double maps the key param and if found increment by 1 its value.
     * @param key: the key of the value to increment
     */
    void incrementMapEntry(std::string key);

    /**
     * search in the int (first) and the double maps the key param and if found increment by val its value.
     * @param key: the key of the value to increment
     * @param val: the value with which to increment the value corresponding to the key
     */
    void incrementMapEntry(std::string key,int val);



    /**
     * @param key: the key to look for in the int map
     * @return: the value corresponding to the key in the int map.
     */
    int getIntMapEntry(std::string key);

    /**
     * @param key: the key to look for in the double map
     * @return: the value corresponding to the key in the double map.
     */
    double getDoubleMapEntry(std::string key);

    /**
     * @param key: the key to look for in the string map
     * @return: the value corresponding to the key in the string map.
     */
    std::string getStringMapEntry(std::string key);

    /**
     * @param key: the key to look for in the SolutionData map
     * @return: the value corresponding to the key in the SolutionData map.
     */
    SolutionData* getSDMapEntry(std::string key);



    /**
     * Transform the data into human readable strings
     *  @return: a human readable string
     */
    std::string getHumanReadableData();

    /**
     * Call getHumanReadableData() and print the return value
     */
    void print();

    /**
     * Transform the data into json
     * @return: a json value containing all the data in the maps
     */
    Json::Value getJsonValue();

    /**
     * Fill the maps with data retrived from root, a json value
     * @param root: a json value containing solutionData info
     */
    void loadFromJsonValue(Json::Value root);



    /**
     * @return: the space to leave for this specific map for human readable string.
     */
    std::string getSpace();

    /**
     * Set the space to leave for this specific map when creating a human readable string
     */
    void setSpace(std::string s);



    /**
     * @return: the indentation to leave between the line begining and the start of the values
     */
    int getSize();

    /**
     * @param s: the indentation to leave between the line begining and the start of the values
     */
    void setSize(int s);


    /**
     * clean all the maps.
     */
    void clear();

private:

    /**
     * The int map
     */
    std::map<std::string, int> _intMap;

    /**
     * The double map
     */
    std::map<std::string, double> _doubleMap;

    /**
     * The string map
     */
    std::map<std::string, std::string> _stringMap;

    /**
     * The recursive SolutionData pointer map
     */
    std::map<std::string, SolutionData*> _SDMap;


    /**
     * The space to leave before writing the key in the human readable string
     * used to inform the user about the position of the information (if it is in a recursive map or not)
     * |key1 ...
     * | key11 ...
     * | key12 ...
     * |key2 ...
     * |key3 ...
     * | key31 ...
     * |  key311 ...
     */
    std::string _space;

    /**
     * |key         = val
     * |longerKey   = otherVal
     * |<---------->
     *    _size
     */
    int _size;


};

#endif // SOLUTIONDATA_H
