#include "navigateAndPick.hpp"
#include "API/project.hpp"

#include "GTP/GTPTools/solutionTools/taskAllSolutions.hpp"
#include "GTP/GTPTools/solutionTools/taskSolution.hpp"

#include "planner/plannerFunctions.hpp"
#include "planner/cost_space.hpp"

#ifdef LIGHT_PLANNER
#include "LightPlanner-pkg.h"
#endif

#define APPROACHDIST 0.1
#define ESCAPEDIST 0.1


using namespace std;


INIT_MOVE3D_STATIC_LOGGER(NavigateAndPick, "planners.gtp.tasks.navigateAndPick");

NavigateAndPick::NavigateAndPick() : Pick()
{
    type = "navigateAndPick";
    typeId = navigateAndPick;
    maxTests = 100;
    nbMaxLoops = 200;
    SD->getSDMapEntry("statistics")->setMapEntry("nbMaxLoops", nbMaxLoops);
}

bool NavigateAndPick::initAll()
{
    result = new TaskAllSolutions(this);

    SD->getSDMapEntry("statistics")->setMapEntry("nbParameters", 4);
    SD->getSDMapEntry("statistics")->setMapEntry("nbTests", 0);
    _armId =0;
    _maxDistFromObj = std::numeric_limits<double>::max();
    return true;
}

bool NavigateAndPick::findCandidateSolutions()
{
    bool test = Pick::findCandidateSolutions();

    if (test)
    {
        _maxDistFromObj = mainAgent->getHriAgent()->manip->reach_tolerance;
//        int nb = SD->getSDMapEntry("statistics")->getIntMapEntry("nbTotPossibleTests");
//        SD->getSDMapEntry("statistics")->setMapEntry("nbTotPossibleTests",nb*maxTests);
    }

    return test;
}

bool NavigateAndPick::getRandomSol()
{

    bool ok = true;

    confPtr_t q_obj = _WS->getRobotConf(mainObject);
//    confPtr_t q_ag = _WS->getRobotConf(mainObject);
    confPtr_t qInit = _WS->getRobotConf(mainAgent);

    if (ok)
    {
        int nbTests = 0;
        bool positionFound = false;
        while (nbTests < maxTests && !positionFound )
        {
            ok = true;
            double distToObj = p3d_random(-_maxDistFromObj,_maxDistFromObj);
            double rotation = p3d_random(-M_PI,M_PI);

            _graspingConf = (confPtr_t) new Configuration(*qInit);
            (*_graspingConf)[6] = (*q_obj)[6] + distToObj*cos(rotation);
            (*_graspingConf)[7] = (*q_obj)[7] + distToObj*sin(rotation);
            (*_graspingConf)[11] = atan2((*q_obj)[7]-(*_graspingConf)[7],(*q_obj)[6]-(*_graspingConf)[6]);

            mainAgent->setAndUpdate(*_graspingConf);
            p3d_update_this_robot_pos_without_cntrt(mainAgent->getRobotStruct());

            p3d_col_activate_robot(mainAgent->getRobotStruct());

            _returnReportcount++;
            vector<p3d_rob*> colRob = pqp_robot_robot_list_collision_test(mainAgent->getRobotStruct());
            if (colRob.size()>0)
            {
                if (_returnReportcountMax < _returnReportcount)
                {
                    _returnReportcountMax = _returnReportcount;
                    SD->setMapEntry("returnReport","no_conf_for_grasp");
                }
                ok = false;
                for (unsigned int i = 0; i < colRob.size(); i++)
                {
                    SD->getSDMapEntry("colision")->incrementMapEntry(colRob.at(i)->name);
                }
            }
            else
                positionFound = true;
            nbTests++;
        }
        SD->getSDMapEntry("statistics")->incrementMapEntry("nbTests", nbTests);
    }

    if (ok)
        ok = Pick::getRandomSol();
    return ok;
}

bool NavigateAndPick::findConfigurations()
{
    confPtr_t qInit = _WS->getRobotConf(mainAgent);
    _curGrasp = graspLists[_armId].at(current_grasp_id);


    bool ok = true;
    if(isHandInCollision())
        ok = false;

    openGripper(_graspingConf,true,_armId);

    mainAgent->setAndUpdate(*_graspingConf);
    p3d_update_this_robot_pos_without_cntrt(mainAgent->getRobotStruct());


    p3d_vector3 v1, v2;
    if (ok)
    {
        activateFingertipCol(false, mainAgent, mainObject, NULL);

        _curGrasp.direction(v1);

        v2[0] = 0;
        v2[1] = 0;
        v2[2] = 1;
    }



    if (ok)
    {
        ok = computeManipulationConfs(approachConf, graspConfClosed, graspConfOpen, escapeConf, mainObject->getJoint(0)->getJointStruct()->abs_pos,
                                         _curGrasp, v1, APPROACHDIST, v2, ESCAPEDIST);
        correctConfigs();
    }

    mainAgent->setAndUpdate(*qInit);
    return ok;
}



bool NavigateAndPick::findTrajectories()
{
    bool ok=true;
    int status = -1;
    _WS->load();

    std::vector<p3d_rob*> col_deactivated;
    for (unsigned int i=0;i<global_Project->getActiveScene()->getNumberOfRobots(); ++i){
        p3d_rob *r1,*r2;
        r1=mainObject->getRobotStruct();
        r2=global_Project->getActiveScene()->getRobot(i)->getRobotStruct();
        bool col=p3d_col_test_robot_other(r1,r2,0);
        if(col){
            col_deactivated.push_back(r2);
            M3D_DEBUG("object "<<r2->name<<" is in collision with mainObject, will be deactivated for escape");
        }
    }


    if(ok){
        graspTraj = compute_traj_using_localPath(approachConf, graspConfClosed, mainObject, status);

        _returnReportcount++;
        if (status != PLAN_STATUS_FOUND && status != PLAN_STATUS_START_EQUALS_GOAL)

        {
            cout << "no grasp traj" << endl;
            if (_returnReportcountMax < _returnReportcount)
            {
                _returnReportcountMax = _returnReportcount;
                if (status == PLAN_STATUS_START_INVALID)
                    SD->setMapEntry("returnReport","grasp_traj_invalid_start");
                else if (status == PLAN_STATUS_NOT_FOUND)
                    SD->setMapEntry("returnReport","grasp_traj_not_found");
                else if (status == PLAN_STATUS_GOAL_INVALID)
                    SD->setMapEntry("returnReport","grasp_traj_invalid_stop");
            }
            ok = false;
        }
    }


    if(ok){
        _WS->load();
        mainAgent->setAndUpdate(*graspConfClosed);
        attachObject();


        vector<SurfaceAttachement> sAtt = createAndLoadSurfaceAttachements(mainObject);
        activateFingertipCol(false, mainAgent, mainObject, NULL);
        for(unsigned int i = 0 ; i<col_deactivated.size();i++)
            p3d_col_deactivate_robot_robot(mainObject->getRobotStruct(), col_deactivated[i]);
        escapeTraj = compute_traj_using_localPath(graspConfClosed, escapeConf, mainObject, status);
        //p3d_col_activate_robot(mainObject->getRobotStruct());
        activateFingertipCol(true, mainAgent, mainObject, NULL);
        for(unsigned int i = 0 ; i<col_deactivated.size();i++)
            p3d_col_activate_robot_robot(mainObject->getRobotStruct(), col_deactivated[i]);

        removeSurfaceAttachements(sAtt);


        _returnReportcount++;
        if (status != PLAN_STATUS_FOUND && status != PLAN_STATUS_START_EQUALS_GOAL)
        {
            cout << "no escape traj found" << endl;
            if (_returnReportcountMax < _returnReportcount)
            {
                _returnReportcountMax = _returnReportcount;
                if (status == PLAN_STATUS_START_INVALID)
                    SD->setMapEntry("returnReport","escape_traj_invalid_start");
                else if (status == PLAN_STATUS_NOT_FOUND)
                    SD->setMapEntry("returnReport","escape_traj_not_found");
                else if (status == PLAN_STATUS_GOAL_INVALID)
                    SD->setMapEntry("returnReport","escape_traj_invalid_stop");
            }
            ok = false;
        }
    }

    detachObject();
    if(ok){
        _WS->load();
        mainAgent->setAndUpdate(*_graspingConf);
        approachTraj = compute_traj(_graspingConf, approachConf, mainObject, status);

        //activateFingertipCol(true, mainAgent, mainObject, NULL);
        _returnReportcount++;
        if (status != PLAN_STATUS_FOUND && status != PLAN_STATUS_START_EQUALS_GOAL)
        {
            cout << "no approach traj" << endl;
            if (_returnReportcountMax < _returnReportcount)
            {
                _returnReportcountMax = _returnReportcount;
                if (status == PLAN_STATUS_START_INVALID)
                    SD->setMapEntry("returnReport","approach_traj_invalid_start");
                else if (status == PLAN_STATUS_NOT_FOUND)
                    SD->setMapEntry("returnReport","approach_traj_not_found");
                else if (status == PLAN_STATUS_GOAL_INVALID)
                    SD->setMapEntry("returnReport","approach_traj_invalid_stop");
            }
            ok = false;
        }else{
            setGraph(new Graph(*API_activeGraph));
            SD->getSDMapEntry("graph")->setMapEntry("nbGraphNode",(int)getGraph()->getNodes().size());
            SD->getSDMapEntry("graph")->setMapEntry("nbComputedNodes",approachTraj->nlp+1);
            SD->getSDMapEntry("graph")->setMapEntry("trajSize",approachTraj->range_param);
        }
    }

    if (ok)
    {
        int status = -1;
        p3d_traj* tmpTraj = compute_traj_using_localPath(_WS->getRobotConf(mainAgent), _graspingConf, NULL, status,false);

        _returnReportcount++;
        if (status != PLAN_STATUS_FOUND && status != PLAN_STATUS_START_EQUALS_GOAL)

        {
            cout << "no nav traj" << endl;
            if (_returnReportcountMax < _returnReportcount)
            {
                _returnReportcountMax = _returnReportcount;
                if (status == PLAN_STATUS_START_INVALID)
                    SD->setMapEntry("returnReport","grasp_traj_invalid_start");
                else if (status == PLAN_STATUS_NOT_FOUND)
                    SD->setMapEntry("returnReport","grasp_traj_not_found");
                else if (status == PLAN_STATUS_GOAL_INVALID)
                    SD->setMapEntry("returnReport","grasp_traj_invalid_stop");
            }
            ok = false;
        }
        if (tmpTraj)
        {
            p3d_destroy_traj(mainAgent->getRobotStruct(),tmpTraj);

            p3d_point p1,p2;
            p1.x = (*_WS->getRobotConf(mainAgent))[6];
            p1.y = (*_WS->getRobotConf(mainAgent))[7];
            p1.z = (*_WS->getRobotConf(mainAgent))[11];
            p2.x = (*_graspingConf)[6];
            p2.y = (*_graspingConf)[7];
            p2.z = (*_graspingConf)[11];
            _navTraj.push_back(p1);
            _navTraj.push_back(p2);
        }
    }

    closeHandTraj = NULL;
    return ok;
}

bool NavigateAndPick::computeSolutionTrajectories(TaskSolution *TS)
{
    bool res=true;
    //cout << getDescr() << endl;
    if(!TS)
    {
        res = false;
    }

    if(res){
        TS->getStartWS()->load();
        TaskSubSolution* TSS0 = TS->getSubSolutionWithId(0);
        TaskSubSolution* TSS1 = TS->getSubSolutionWithId(1);
        TaskSubSolution* TSS2 = TS->getSubSolutionWithId(2);
        TaskSubSolution* TSS4 = TS->getSubSolutionWithId(4);

        _graspingConf   = TSS0->getEndState()->getRobotConf(mainAgent);
        approachConf    = TSS1->getEndState()->getRobotConf(mainAgent);
        graspConfClosed = TSS2->getEndState()->getRobotConf(mainAgent);
        escapeConf      = TSS4->getEndState()->getRobotConf(mainAgent);

        _armId = TSS1->getArmId();
        _planPartToUse = TSS1->getPlanningPart();

        res = findTrajectories();
    }


    if(res){
        TS->getSubSolutionWithId(0)->setNavTraj(_navTraj);
        TS->getSubSolutionWithId(1)->setTraj(approachTraj);
        TS->getSubSolutionWithId(2)->setTraj(graspTraj);
        TS->getSubSolutionWithId(4)->setTraj(escapeTraj);

        TS->setTrajComputed(true);
    }

    return res;
}

TaskSolution *NavigateAndPick::createSolution(bool computeMP)
{
    correctConfigs();
    TaskSolution* TS = new TaskSolution();
    TS->setTrajComputed(computeMP);
    TS->setTaskConstraints(_currConstraints);

    TaskSubSolution* TSS0 = new TaskSubSolution(TS, "navigate", 0, NULL, mainAgent, 0);
    TaskSubSolution* TSS1 = new TaskSubSolution(TS, "approach", 1, approachTraj, mainAgent, 0);
    TaskSubSolution* TSS2 = new TaskSubSolution(TS, "engage", 2, graspTraj, mainAgent, 0);
    TaskSubSolution* TSS3 = new TaskSubSolution(TS, "grasp", 3, closeHandTraj, mainAgent, 0);
    TaskSubSolution* TSS4 = new TaskSubSolution(TS, "escape", 4, escapeTraj, mainAgent, 0);

    TSS0->setNavigation(true);;
    TSS0->setNavTraj(_navTraj);

    TSS1->setArmId(_armId);
    TSS2->setArmId(_armId);
    TSS3->setArmId(_armId);
    TSS4->setArmId(_armId);

    TSS1->setPlanningPart(_planPartToUse);
    TSS2->setPlanningPart(_planPartToUse);
    TSS3->setPlanningPart(_planPartToUse);
    TSS4->setPlanningPart(_planPartToUse);

    detachObject();

    vector<TaskSubSolution*> v1;
    v1.push_back(TSS1);
    TSS0->setNext(v1);

    vector<TaskSubSolution*> v2;
    v2.push_back(TSS2);
    TSS1->setNext(v2);

    vector<TaskSubSolution*> v3;
    v3.push_back(TSS3);
    TSS2->setNext(v3);

    vector<TaskSubSolution*> v4;
    v4.push_back(TSS4);
    TSS3->setNext(v4);

    TSS0->setStartState(_WS);
    TS->setStartWS(_WS);
    WorldState* WS0 = new WorldState(_WS);
    WS0->replaceConf(_graspingConf);

    TSS0->setEndState(WS0);
    TSS1->setStartState(WS0);

    WorldState* WS1 = new WorldState(_WS);
    bool t = WS1->replaceConf(approachConf);
    if (!t)
    {
        cout << "ERROR: error with robots names" << endl;
        delete TS;
        return NULL;
    }

    TSS1->setEndState(WS1);
    TSS2->setStartState(WS1);

    WorldState* WS2 = new WorldState(WS1);
    t = WS2->replaceConf(graspConfOpen);
    if (!t)
    {
        cout << "ERROR: error with robots names" << endl;
        delete TS;
        return NULL;
    }

    TSS2->setEndState(WS2);
    TSS3->setStartState(WS2);


    WorldState* WS3 = new WorldState(WS2);
    t = WS3->replaceConf(graspConfClosed);
    if (!t)
    {
        cout << "ERROR: error with robots names" << endl;
        delete TS;
        return NULL;
    }

    gpGrasp gr = _curGrasp;
    WS3->addAttachement(mainAgent, mainObject, gr, getArmId());
    TSS3->setEndState(WS3);
    TSS4->setStartState(WS3);

    WorldState* WS4 = new WorldState(WS3);
    WS4->load();
    vector<SurfaceAttachement> sss = createAndLoadSurfaceAttachements(mainObject);
    WS3->setSurfaceAttachements(sss);
//    WS4->setSurfaceAttachements(sss);
    t = WS4->replaceConf(escapeConf);
    if (!t)
    {
        cout << "ERROR: error with robots names" << endl;
        delete TS;
        return NULL;
    }
    mainAgent->setAndUpdate(*escapeConf);

    for (unsigned int i = 0; i<sss.size(); i++)
        WS4->replaceConf(sss.at(i).attache->getCurrentPos());

    _WS->load();
    //    mainAgent->setAndUpdate(*graspConfClosed);
    ////    p3d_mat4Print(mainObject->getJoint(0)->getJointStruct()->abs_pos,"avant");
    //    double avZ = mainObject->getJoint(0)->getJointStruct()->abs_pos[2][3];
    //    attachObject();
    //    mainAgent->setAndUpdate(*escapeConf);
    //    p3d_mat4Print(mainObject->getJoint(0)->getJointStruct()->abs_pos,"apres");

    confPtr_t q = mainObject->getCurrentPos();
    //    if (avZ == mainObject->getJoint(0)->getJointStruct()->abs_pos[2][3])
    //TODO adding the direction to the escapeDist
    (*q)[8] =  (*q)[8] + ESCAPEDIST;


    WS4->replaceConf(q);


//    WS4->addAttachement(mainAgent, mainObject, gr, getArmId());
    TSS4->setEndState(WS4);


    WS4->setTSThatEndWithThisWS(TS);
    WS4->setTCForNextTask(_currConstraints);

    TS->setStopWS(WS4);


    TS->addConf(_WS->getRobotConf(mainAgent));
    TS->addConf(approachConf);
    TS->addConf(graspConfOpen);
    TS->addConf(graspConfClosed);
    TS->addConf(escapeConf);

    _WS->updateRobotVirtualObj(mainAgent);
    WS0->updateRobotVirtualObj(mainAgent);
    WS1->updateRobotVirtualObj(mainAgent);
    WS2->updateRobotVirtualObj(mainAgent);
    WS3->updateRobotVirtualObj(mainAgent);
    WS4->updateRobotVirtualObj(mainAgent);


    return TS;

}
