#include "pickPlaceQuad.hpp"
#include "API/project.hpp"

#include "planner/plannerFunctions.hpp"
#include "planner/TrajectoryOptim/Classic/smoothing.hpp"

#include "GTP/GTPTools/solutionTools/taskAllSolutions.hpp"
#include "GTP/GTPTools/solutionTools/taskSolution.hpp"

#ifdef LIGHT_PLANNER
#include "LightPlanner-pkg.h"
#endif

#define OBJECT_RAY 0.012
#define APPROACH_DIST 0.15
#define SPHERE_RAY 0.75

#define ARM_EF_JNT 8
#define ARM_NB_DOF 8

#define VERBOSE 1
#define WITH_TASK_ID 0

#define PLACE_NB_TRY 20

#define TO_LOG 0

#define MAX_TIME_PICK_ARM       1
#define MAX_TIME_PICK_GOTO_LIN  5
#define MAX_TIME_PICK_GOTO_DYN 10
#define MAX_TIME_PLACE_ARM      2
#define MAX_TIME_PLACE_GOTO    10

//####################### PickPlaceQuad ########################

PickPlaceQuad::PickPlaceQuad() : PrivateTask() {

    type="pickplacequad";
    typeId = pickPlaceQuad;
    nbMaxLoops=20;
    result=NULL;
    _idCurrentSol=-1;



    SD->getSDMapEntry("statistics")->setMapEntry("nbMaxLoops", nbMaxLoops);
    SD->getSDMapEntry("statistics")->setMapEntry("nbParameters", 3);
    SD->getSDMapEntry("statistics")->setMapEntry("nbGikCalled", 0);
    SD->getSDMapEntry("statistics")->setMapEntry("nbMPCalled", 0);
    SD->getSDMapEntry("statistics")->setMapEntry("nbTests", 0);
    SD->getSDMapEntry("statistics")->setMapEntry("nbTotPossibleTests", nbMaxLoops);
    SD->setMapEntry("returnReport", "ERROR");

    _returnReportcount=0;
    _returnReportcountMax=0;

    _pick_goto_traj=NULL;
    _pick_arm_traj=NULL;
    _pick_approach_traj=NULL;
    _pick_traj=NULL;
    _pick_extract_traj=NULL;

    _place_goto_traj=NULL;
    _place_arm_traj=NULL;
    _place_traj=NULL;
    _place_extract_traj=NULL;

    _mainObject=NULL;
    _supportObject=NULL;
    _ghost=NULL;

    _param=NULL;

    is_pick=true;
    is_place=true;
    strcpy(_className,"PickPlaceQuad");

}

void PickPlaceQuad::print_msg(const char *fct,const char *msg,int verbose_level){
    if(VERBOSE>0){
        if(TO_LOG){
            char ad[100];
            sprintf(ad,"%s/log_move3d",getenv("HOME"));
            FILE *log=fopen(ad,"a");
            if(verbose_level==1)
                fprintf(log,"error: ");
            if(verbose_level<=VERBOSE)
                fprintf(log,"%s::%s: %s",_className,fct,msg);
            if(WITH_TASK_ID)
                fprintf(log," (task_id: %d)",task_id);
            fprintf(log,"\n");
            fclose(log);
        }
        else{
            if(verbose_level==1)
                printf("error: ");
            if(verbose_level<=VERBOSE){
                printf("%s::%s: %s",_className,fct,msg);
                if(WITH_TASK_ID)
                    printf(" (task_id: %d)",task_id);
                printf("\n");
            }
        }
    }
}

bool PickPlaceQuad::initAll(){
    const char *fct="initAll";
    result=new TaskAllSolutions(this);
    if(!result){
        print_msg(fct,"TaskAllSolutions instantiation has failed");
        return false;
    }
    if(TO_LOG){
        char ad[100];
        sprintf(ad,"%s/log_move3d",getenv("HOME"));
        FILE *log=fopen(ad,"w");
        fprintf(log,"# move3d log file\n");
        fclose(log);
    }
    print_msg(fct,"finished with success",3);
    return true;
}

bool PickPlaceQuad::setAgents(std::multimap<std::string,std::string> agents_name){
    const char *fct="setAgents";
    if(!setMainAgent(agents_name)){
        print_msg(fct,"setMainAgent has failed");
        return false;
    }
    if(!mainAgent){
        print_msg(fct,"mainAgent not found");
        return false;
    }
    if(!mainAgent->getRobotStruct()){
        print_msg(fct,"mainAgent has no robot structure");
        return false;
    }
    _param=lm_get_quadspline_lm_param(mainAgent->getRobotStruct());
    if(!_param){
        print_msg(fct,"mainAgent has no quadspline structure");
        return false;
    }
    print_msg(fct,"finished with success",3);
    return true;
}

bool PickPlaceQuad::setObjects(std::multimap<std::string,std::string> objects_name){
    const char *fct="setObjects";
    _mainObject=global_Project->getActiveScene()->getRobotByName(objects_name.find("mainObject")->second);
    if(!_mainObject){
        print_msg(fct,"_mainObject not found");
        return false;
    }
    if(!_mainObject->getRobotStruct()){
        print_msg(fct,"_mainObject has no robot structure");
        return false;
    }
    if(is_place){
        _supportObject=global_Project->getActiveScene()->getRobotByName(objects_name.find("supportObject")->second);
        if(!_supportObject){
            print_msg(fct,"_supportObject not found");
            return false;
        }
        if(!_supportObject->getRobotStruct()){
            print_msg(fct,"_supportObject has no robot structure");
            return false;
        }
    }
    print_msg(fct,"finished with success",3);
    return true;
}

bool PickPlaceQuad::initialize(){

    const char *fct="initialize";

    if(!_WS){
        print_msg(fct,"WorldState is NULL");
        return false;
    }
    p3d_rob *robotPt=mainAgent->getRobotStruct();
    if(!robotPt){
        print_msg(fct,"mainAgent has no robot structure");
        return false;
    }
    p3d_grasp *grasp=robotPt->grasped_objects;
    if(is_pick){
        if(grasp || _WS->hasAttachement(mainAgent)){
            print_msg(fct,"mainAgent is already grasping an object");
            return false;
        }
    }
    else{
        if(is_place){
            if(!grasp || !_WS->hasAttachement(mainAgent)){
                print_msg(fct,"mainAgent is not grasping any object");
                return false;
            }
            bool found=false;
            while(grasp && !found){
                found=(grasp->objectPt==_mainObject->getRobotStruct() && _mainObject->getRobotStruct()->grasp->robotPt==mainAgent->getRobotStruct());
                grasp=grasp->next;
            }
            if(!found || !_WS->hasAttachement(mainAgent,_mainObject)){
                std::cout<<"error: PlaceQuad::initialize(): mainAgent "<<mainAgent->getName()<<" is not graping object "<<_mainObject->getName()<<" for task "<<type<<" with id "<<task_id<<std::endl;
                return false;
            }
        }
        else{
            print_msg(fct,"bad class instantiation: this is neither pick nor place");
            return false;
        }
    }

    configPt init_conf=p3d_get_robot_config(robotPt);
    if(!init_conf){print_msg(fct,"cannot allocate init_conf");return false;}
    _initConf=(confPtr_t)new Configuration(mainAgent,init_conf,true);
    if(!_initConf){print_msg(fct,"cannot allocate _initConf");return false;}

    if(is_pick){
        p3d_rob *ghost=p3d_get_robot_by_name("ROBOT_bonebreaker_sphere_ghost");
        if(ghost){
            _ghost=new Robot(ghost);
            configPt ghost_conf=p3d_alloc_config(ghost);
            if(!ghost_conf){
                print_msg(fct,"cannot allocate configPt ghost_conf");
                use_ghost=false;
            }
            else{
                ghost_conf[6]=init_conf[_param->x_dof];
                ghost_conf[7]=init_conf[_param->y_dof];
                ghost_conf[8]=init_conf[_param->z_dof];
                ghost_conf[11]=init_conf[_param->rz_dof];
                p3d_set_col_off(robotPt);
                p3d_set_col_on(ghost);
                p3d_set_and_update_this_robot_conf(ghost,ghost_conf);
                use_ghost=(bool)!p3d_col_test_robot(ghost,0);
                p3d_set_col_on(robotPt);
                p3d_set_col_off(ghost);
            }
        }
        else{
            _ghost=NULL;
            use_ghost=false;
        }
    }
    else{
        _ghost=NULL;
        use_ghost=false;
    }

    PlanEnv->setBool(PlanParam::planWithTimeLimit,true);

    print_msg(fct,"finished with success",3);
    return true;
}

bool PickPlaceQuad::findPickConf(){

    const char *fct="findPickConf";
    if(!is_pick){
        print_msg(fct,"is_pick=false");
        return false;
    }
    p3d_rob* robotPt=mainAgent->getRobotStruct();
    p3d_rob* objectPt=_mainObject->getRobotStruct();
    configPt init_conf=_initConf->getConfigStruct();

    SD->getSDMapEntry("statistics")->incrementMapEntry("nbTests");
    configPt pick_conf=p3d_get_random_bonebreaker_grasp(robotPt,objectPt,OBJECT_RAY);
    if(!pick_conf){
        print_msg(fct,"p3d_get_random_bonebreaker_grasp failed",2);
        _returnReportcount++;
        if(_returnReportcountMax<_returnReportcount){
            _returnReportcountMax=_returnReportcount;
            SD->setMapEntry("returnReport","no_pick_position");
        }
        p3d_set_and_update_this_robot_conf(robotPt,init_conf);
        return false;
    }
    p3d_set_and_update_this_robot_conf(robotPt,pick_conf);
    _pickConf=(confPtr_t)new Configuration(mainAgent,pick_conf,true);
    if(p3d_col_test_robot(robotPt,0)){
        print_msg(fct,"pick configuration in collision",2);
        _returnReportcount++;
        if(_returnReportcountMax<_returnReportcount){
            _returnReportcountMax=_returnReportcount;
            SD->setMapEntry("returnReport","no_pick_position");
        }
        p3d_set_and_update_this_robot_conf(robotPt,init_conf);
        return false;
    }
    _pick_graspDir[0]=robotPt->joints[ARM_EF_JNT]->abs_pos[0][2];
    _pick_graspDir[1]=robotPt->joints[ARM_EF_JNT]->abs_pos[1][2];
    _pick_graspDir[2]=robotPt->joints[ARM_EF_JNT]->abs_pos[2][2];

    p3d_set_and_update_this_robot_conf(robotPt,init_conf);
    print_msg(fct,"finished with success",3);
    return true;

}

bool PickPlaceQuad::findPickApproachConf(){

    const char *fct="findPickApproachConf";
    if(!is_pick){
        print_msg(fct,"is_pick=false");
        return false;
    }
    p3d_rob* robotPt=mainAgent->getRobotStruct();
    configPt init_conf=_initConf->getConfigStruct();
    configPt pick_conf=_pickConf->getConfigStruct();

    _pick_approachConf=(confPtr_t)new Configuration(mainAgent,pick_conf);
    configPt pick_approach_conf=_pick_approachConf->getConfigStruct();
    pick_approach_conf[_param->x_dof]-=APPROACH_DIST*_pick_graspDir[0];
    pick_approach_conf[_param->y_dof]-=APPROACH_DIST*_pick_graspDir[1];
    pick_approach_conf[_param->z_dof]-=APPROACH_DIST*_pick_graspDir[2];
    p3d_set_and_update_this_robot_conf(robotPt,pick_approach_conf);
    if(p3d_col_test_robot(robotPt,0)){
        print_msg(fct,"pick approach configuration in collision",2);
        _returnReportcount++;
        if(_returnReportcountMax<_returnReportcount){
            _returnReportcountMax=_returnReportcount;
            SD->setMapEntry("returnReport","no_pick_approach_position");
        }
        p3d_set_and_update_this_robot_conf(robotPt,init_conf);
        return false;
    }
    p3d_set_and_update_this_robot_conf(robotPt,init_conf);
    print_msg(fct,"finished with success",3);
    return true;
}

bool PickPlaceQuad::findPickTraj(){

    const char *fct="findPickTraj";
    if(!is_pick){
        print_msg(fct,"is_pick=false");
        return false;
    }
    p3d_rob* robotPt=mainAgent->getRobotStruct();
    configPt init_conf=_initConf->getConfigStruct();
    configPt pick_approach_conf=_pick_approachConf->getConfigStruct();
    configPt pick_conf=_pickConf->getConfigStruct();

    robotPt->lpl_type=LINEAR;
    int nb_col_test;
    p3d_localpath *pick_lp=p3d_linear_localplanner(robotPt,pick_approach_conf,pick_conf,0);
    if(!pick_lp || p3d_col_test_localpath(robotPt,pick_lp,&nb_col_test)){
        if(pick_lp){
            print_msg(fct,"pick trajectory in collision",2);
            pick_lp->destroy(robotPt,pick_lp);
        }
        else            
            print_msg(fct,"cannot allocate pick localpath");
        _returnReportcount++;
        if(_returnReportcountMax<_returnReportcount){
            _returnReportcountMax=_returnReportcount;
            SD->setMapEntry("returnReport","no_pick_trajectory");
        }
        robotPt->lpl_type=QUADSPLINE;
        p3d_set_and_update_this_robot_conf(robotPt,init_conf);
        return false;
    }

    robotPt->lpl_type=QUADSPLINE;
    p3d_set_and_update_this_robot_conf(robotPt,init_conf);

    _pick_traj=p3d_create_traj_from_localpath(robotPt,pick_lp);
    pick_lp->destroy(robotPt,pick_lp);
    if(!_pick_traj){
        print_msg(fct,"cannot allocate pick trajectory");
        _returnReportcount++;
        if(_returnReportcountMax<_returnReportcount){
            _returnReportcountMax=_returnReportcount;
            SD->setMapEntry("returnReport","no_pick_trajectory");
        }
        return false;
    }

    print_msg(fct,"finished with success",3);
    return true;
}

bool PickPlaceQuad::findPickExtractConf(){

    const char *fct="findPickExtractConf";
    if(!is_pick){
        print_msg(fct,"is_pick=false");
        return false;
    }
    p3d_rob* robotPt=mainAgent->getRobotStruct();
    p3d_rob* objectPt=_mainObject->getRobotStruct();
    configPt init_conf=_initConf->getConfigStruct();
    configPt pick_conf=_pickConf->getConfigStruct();

    _pick_extractConf=(confPtr_t)new Configuration(mainAgent,pick_conf);
    configPt pick_extract_conf=_pick_extractConf->getConfigStruct();
    pick_extract_conf[_param->z_dof]+=APPROACH_DIST;
    p3d_set_and_update_this_robot_conf(robotPt,pick_conf);
    print_msg(fct,"PICK",3);
    p3d_pick(robotPt,ARM_EF_JNT,objectPt);
    p3d_set_and_update_this_robot_conf(robotPt,pick_extract_conf);
    if(p3d_col_test_robot(robotPt,0)){
        printf(fct,"pick extract configuration in collision",2);
        _returnReportcount++;
        if(_returnReportcountMax<_returnReportcount){
            _returnReportcountMax=_returnReportcount;
            SD->setMapEntry("returnReport","no_pick_extract_position");
        }
        p3d_set_and_update_this_robot_conf(robotPt,pick_conf);
        print_msg(fct,"PLACE",3);
        p3d_place(robotPt,objectPt);
        p3d_set_and_update_this_robot_conf(robotPt,init_conf);
        return false;
    }

    _pick_extractObjectConf=(confPtr_t)new Configuration(_mainObject,p3d_get_robot_config(objectPt));
    p3d_set_and_update_this_robot_conf(robotPt,pick_conf);
    print_msg(fct,"PLACE",3);
    p3d_place(robotPt,objectPt);
    p3d_set_and_update_this_robot_conf(robotPt,init_conf);
    print_msg(fct,"finished with success",3);
    return true;
}


bool PickPlaceQuad::findPickExtractTraj(){

    const char *fct="findPickExtractTraj";
    if(!is_pick){
        print_msg(fct,"is_pick=false");
        return false;
    }
    p3d_rob* robotPt=mainAgent->getRobotStruct();
    p3d_rob* objectPt=_mainObject->getRobotStruct();
    configPt init_conf=_initConf->getConfigStruct();
    configPt pick_conf=_pickConf->getConfigStruct();
    configPt pick_extract_conf=_pick_extractConf->getConfigStruct();

    robotPt->lpl_type=LINEAR;
    p3d_set_and_update_this_robot_conf(robotPt,pick_conf);
    print_msg(fct,"PICK",3);
    p3d_pick(robotPt,ARM_EF_JNT,objectPt);
    p3d_localpath *pick_extract_lp=p3d_linear_localplanner(robotPt,pick_conf,pick_extract_conf,0);
    int nb_col_test;
    if(!pick_extract_lp || p3d_col_test_localpath(robotPt,pick_extract_lp,&nb_col_test)){
        if(pick_extract_lp){
            print_msg(fct,"pick extract trajectory",2);
            pick_extract_lp->destroy(robotPt,pick_extract_lp);
        }
        else
            print_msg(fct,"pick extract trajectory allocation failed");
        _returnReportcount++;
        if(_returnReportcountMax<_returnReportcount){
            _returnReportcountMax=_returnReportcount;
            SD->setMapEntry("returnReport","no_pick_extract_trajectory");
        }
        robotPt->lpl_type=QUADSPLINE;
        p3d_set_and_update_this_robot_conf(robotPt,pick_conf);
        print_msg(fct,"PLACE",3);
        p3d_place(robotPt,objectPt);
        p3d_set_and_update_this_robot_conf(robotPt,init_conf);
        return false;
    }

    robotPt->lpl_type=QUADSPLINE;
    p3d_set_and_update_this_robot_conf(robotPt,pick_conf);
    print_msg(fct,"PLACE",3);
    p3d_place(robotPt,objectPt);
    p3d_set_and_update_this_robot_conf(robotPt,init_conf);

    _pick_extract_traj=p3d_create_traj_from_localpath(robotPt,pick_extract_lp);
    pick_extract_lp->destroy(robotPt,pick_extract_lp);
    if(!_pick_extract_traj){
        print_msg(fct,"cannot allocate pick extract trajectory");
        _returnReportcount++;
        if(_returnReportcountMax<_returnReportcount){
            _returnReportcountMax=_returnReportcount;
            SD->setMapEntry("returnReport","no_pick_extract_trajectory");
        }
        return false;
    }

    print_msg(fct,"finished with success",3);
    return true;
}

bool PickPlaceQuad::findPickTargetConf(){

    const char *fct="findPickTargetConf";
    if(!is_pick){
        print_msg(fct,"is_pick=false");
        return false;
    }
    p3d_rob* robotPt=mainAgent->getRobotStruct();
    p3d_rob* objectPt=_mainObject->getRobotStruct();
    configPt init_conf=_initConf->getConfigStruct();
    configPt pick_approach_conf=_pick_approachConf->getConfigStruct();

    p3d_vector3 approachDir;
    p3d_set_and_update_this_robot_conf(robotPt,pick_approach_conf);
    approachDir[0]=objectPt->joints[1]->abs_pos[0][3]-robotPt->joints[1]->abs_pos[0][3];
    approachDir[1]=objectPt->joints[1]->abs_pos[1][3]-robotPt->joints[1]->abs_pos[1][3];
    approachDir[2]=objectPt->joints[1]->abs_pos[2][3]-robotPt->joints[1]->abs_pos[2][3];
    double dist=p3d_vectNorm(approachDir);

    // TARGET CONF
    _pick_targetConf=(confPtr_t)new Configuration(mainAgent,init_conf);
    configPt pick_target_conf=_pick_targetConf->getConfigStruct();
    pick_target_conf[_param->x_dof]=pick_approach_conf[_param->x_dof];
    pick_target_conf[_param->y_dof]=pick_approach_conf[_param->y_dof];
    pick_target_conf[_param->z_dof]=pick_approach_conf[_param->z_dof];
    if(dist<OBJECT_RAY+SPHERE_RAY+APPROACH_DIST){
        approachDir[0]/=dist;
        approachDir[1]/=dist;
        approachDir[2]/=dist;
        dist=OBJECT_RAY+SPHERE_RAY+APPROACH_DIST-dist;
        pick_target_conf[_param->x_dof]-=dist*approachDir[0];
        pick_target_conf[_param->y_dof]-=dist*approachDir[1];
        pick_target_conf[_param->z_dof]-=dist*approachDir[2];
    }
    pick_target_conf[_param->rx_dof]=0.0;
    pick_target_conf[_param->ry_dof]=0.0;
    pick_target_conf[_param->rz_dof]=pick_approach_conf[_param->rz_dof];
    for(int k=0;k<6;k++){
        pick_target_conf[_param->vx_dof+k]=0.0;
    }

    // GHOST CONF
    if(use_ghost){
        p3d_rob *ghost=_ghost->getRobotStruct();
        configPt ghost_conf=p3d_alloc_config(ghost);
        if(!ghost_conf){
            print_msg(fct,"cannot allocate configPt ghost_conf");
            use_ghost=false;
        }
        else{
            ghost_conf[6]=pick_target_conf[_param->x_dof];
            ghost_conf[7]=pick_target_conf[_param->y_dof];
            ghost_conf[8]=pick_target_conf[_param->z_dof];
            ghost_conf[11]=pick_target_conf[_param->rz_dof];
            p3d_set_col_off(robotPt);
            p3d_set_col_on(ghost);
            p3d_set_and_update_this_robot_conf(ghost,ghost_conf);
            use_ghost=(bool)!p3d_col_test_robot(ghost,0);
            p3d_set_col_on(robotPt);
            p3d_set_col_off(ghost);
            _pick_targetGhostConf=(confPtr_t)new Configuration(_ghost,ghost_conf,true);
        }
    }

    if(!use_ghost){
        p3d_set_and_update_this_robot_conf(robotPt,pick_target_conf);
        if(p3d_col_test_robot(robotPt,0)){
            print_msg(fct,"pick target position in collision");
            _returnReportcount++;
            if(_returnReportcountMax<_returnReportcount)
            {
                _returnReportcountMax=_returnReportcount;
                SD->setMapEntry("returnReport","no_pick_target_position");
            }
            p3d_set_and_update_this_robot_conf(robotPt,init_conf);
            return false;
        }
    }

    p3d_set_and_update_this_robot_conf(robotPt,init_conf);
    print_msg(fct,"finished with success",3);
    return true;
}

bool PickPlaceQuad::findPickTargetPlacedArmConf(){

    const char *fct="findPickTargetPlacedArmConf";
    if(!is_pick){
        print_msg(fct,"is_pick=false");
        return false;
    }
    p3d_rob* robotPt=mainAgent->getRobotStruct();
    configPt init_conf=_initConf->getConfigStruct();
    configPt pick_target_conf=_pick_targetConf->getConfigStruct();
    configPt pick_approach_conf=_pick_approachConf->getConfigStruct();

    _pick_targetPlacedArmConf=(confPtr_t)new Configuration(mainAgent,pick_target_conf);
    configPt pick_target_conf_placed_arm=_pick_targetPlacedArmConf->getConfigStruct();
    for(int k=1;k<=ARM_NB_DOF;k++){
        pick_target_conf_placed_arm[_param->az_dof+k]=pick_approach_conf[_param->az_dof+k];
    }
    p3d_set_and_update_this_robot_conf(robotPt,pick_target_conf_placed_arm);
    if(p3d_col_test_robot(robotPt,0)){
        print_msg(fct,"pick target position with placed arm in collision");
        _returnReportcount++;
        if(_returnReportcountMax<_returnReportcount){
            _returnReportcountMax=_returnReportcount;
            SD->setMapEntry("returnReport","no_pick_target_with_placed_arm_position");
        }
        p3d_set_and_update_this_robot_conf(robotPt,init_conf);
        return false;
    }

    p3d_set_and_update_this_robot_conf(robotPt,init_conf);
    print_msg(fct,"finished with success",3);
    return true;
}

bool PickPlaceQuad::findPlaceConf(){

    const char *fct="findPlaceConf";
    if(!is_place){
        print_msg(fct,"is_place=false");
        return false;
    }
    p3d_rob* robotPt=mainAgent->getRobotStruct();
    p3d_rob* objectPt=_mainObject->getRobotStruct();
    p3d_rob* supportPt=_supportObject->getRobotStruct();
    configPt init_conf=_initConf->getConfigStruct();

    if(is_pick){
        p3d_set_and_update_this_robot_conf(robotPt,_pickConf->getConfigStruct());
        print_msg(fct,"PICK",3);
        p3d_pick(robotPt,ARM_EF_JNT,objectPt);
    }

    SD->getSDMapEntry("statistics")->incrementMapEntry("nbTests");
    configPt place_conf=p3d_get_random_bonebreaker_conf_for_given_grasp(robotPt,supportPt,objectPt->grasp->Matt);
    if(!place_conf){
        print_msg(fct,"p3d_get_random_bonebreaker_conf_for_given_grasp failed",2);
        _returnReportcount++;
        if(_returnReportcountMax<_returnReportcount){
            _returnReportcountMax=_returnReportcount;
            SD->setMapEntry("returnReport","no_place_position");
        }
        if(is_pick){
            p3d_set_and_update_this_robot_conf(robotPt,_pickConf->getConfigStruct());
            print_msg(fct,"PLACE",3);
            p3d_place(robotPt,objectPt);
        }
        p3d_set_and_update_this_robot_conf(robotPt,init_conf);
        return false;
    }
    _placeConf=(confPtr_t)new Configuration(mainAgent,place_conf,true);
    p3d_set_and_update_this_robot_conf(robotPt,place_conf);
    if(p3d_col_test_robot(robotPt,0)){
        print_msg(fct,"place configuration in collision",2);
        _returnReportcount++;
        if(_returnReportcountMax<_returnReportcount){
            _returnReportcountMax=_returnReportcount;
            SD->setMapEntry("returnReport","no_place_position");
        }
        if(is_pick){
            p3d_set_and_update_this_robot_conf(robotPt,_pickConf->getConfigStruct());
            print_msg(fct,"PLACE",3);
            p3d_place(robotPt,objectPt);
        }
        p3d_set_and_update_this_robot_conf(robotPt,init_conf);
        return false;
    }
    configPt place_object_conf=p3d_get_robot_config(objectPt);
    if(!place_object_conf){
        print_msg(fct,"cannot allocate configPt place_object_conf");
        return false;
    }
    _placeObjectConf=(confPtr_t)new Configuration(_mainObject,place_object_conf,true);

    _place_graspDir[0]=robotPt->joints[ARM_EF_JNT]->abs_pos[0][2];
    _place_graspDir[1]=robotPt->joints[ARM_EF_JNT]->abs_pos[1][2];
    _place_graspDir[2]=robotPt->joints[ARM_EF_JNT]->abs_pos[2][2];

    if(is_pick){
        p3d_set_and_update_this_robot_conf(robotPt,_pickConf->getConfigStruct());
        print_msg(fct,"PLACE",3);
        p3d_place(robotPt,objectPt);
    }
    p3d_set_and_update_this_robot_conf(robotPt,init_conf);
    print_msg(fct,"finished with success",3);
    return true;
}

bool PickPlaceQuad::findPlaceApproachConf(){

    const char *fct="findPlaceApproachConf";
    if(!is_place){
        print_msg(fct,"is_place=false");
        return false;
    }
    p3d_rob* robotPt=mainAgent->getRobotStruct();
    p3d_rob* objectPt=_mainObject->getRobotStruct();
    configPt init_conf=_initConf->getConfigStruct();
    configPt place_conf=_placeConf->getConfigStruct();

    if(is_pick){
        p3d_set_and_update_this_robot_conf(robotPt,_pickConf->getConfigStruct());
        print_msg(fct,"PICK",3);
        p3d_pick(robotPt,ARM_EF_JNT,objectPt);
    }

    _place_approachConf=(confPtr_t)new Configuration(mainAgent,place_conf);
    configPt place_approach_conf=_place_approachConf->getConfigStruct();
    place_approach_conf[_param->z_dof]+=APPROACH_DIST;
    p3d_set_and_update_this_robot_conf(robotPt,place_approach_conf);
    if(p3d_col_test_robot(robotPt,0)){
        print_msg(fct,"place approach configuration in collision",2);
        _returnReportcount++;
        if(_returnReportcountMax<_returnReportcount){
            _returnReportcountMax=_returnReportcount;
            SD->setMapEntry("returnReport","no_place_approach_position");
        }
        if(is_pick){
            p3d_set_and_update_this_robot_conf(robotPt,_pickConf->getConfigStruct());
            print_msg(fct,"PLACE",3);
            p3d_place(robotPt,objectPt);
        }
        p3d_set_and_update_this_robot_conf(robotPt,init_conf);
        return false;
    }

    configPt place_approach_object_conf=p3d_get_robot_config(objectPt);
    if(!place_approach_object_conf){
        print_msg(fct,"cannot allocate configPt place_approach_object_conf");
        return false;
    }
    _place_approachObjectConf=(confPtr_t)new Configuration(_mainObject,place_approach_object_conf,true);

    if(is_pick){
        p3d_set_and_update_this_robot_conf(robotPt,_pickConf->getConfigStruct());
        print_msg(fct,"PLACE",3);
        p3d_place(robotPt,objectPt);
    }
    p3d_set_and_update_this_robot_conf(robotPt,init_conf);

    print_msg(fct,"finished with success",3);
    return true;
}

bool PickPlaceQuad::findPlaceTraj(){

    const char *fct="findPlaceTraj";
    if(!is_place){
        print_msg(fct,"is_place=false");
        return false;
    }
    p3d_rob* robotPt=mainAgent->getRobotStruct();
    p3d_rob* objectPt=_mainObject->getRobotStruct();
    configPt init_conf=_initConf->getConfigStruct();
    configPt place_approach_conf=_place_approachConf->getConfigStruct();
    configPt place_conf=_placeConf->getConfigStruct();

    if(is_pick){
        p3d_set_and_update_this_robot_conf(robotPt,_pickConf->getConfigStruct());
        print_msg(fct,"PICK",3);
        p3d_pick(robotPt,ARM_EF_JNT,objectPt);
    }

    robotPt->lpl_type=LINEAR;
    p3d_localpath *place_lp=p3d_linear_localplanner(robotPt,place_approach_conf,place_conf,0);
    int nb_col_test;
    if(!place_lp || p3d_col_test_localpath(robotPt,place_lp,&nb_col_test)){
        if(place_lp){
            print_msg(fct,"place trajectory in collision",2);
            place_lp->destroy(robotPt,place_lp);
        }
        else
            print_msg(fct,"cannot allocate place localpath");
        _returnReportcount++;
        if(_returnReportcountMax<_returnReportcount){
            _returnReportcountMax=_returnReportcount;
            SD->setMapEntry("returnReport","no_place_trajectory");
        }
        if(is_pick){
            p3d_set_and_update_this_robot_conf(robotPt,_pickConf->getConfigStruct());
            print_msg(fct,"PLACE",3);
            p3d_place(robotPt,objectPt);
        }
        p3d_set_and_update_this_robot_conf(robotPt,init_conf);
        return false;
    }

    robotPt->lpl_type=QUADSPLINE;
    if(is_pick){
        p3d_set_and_update_this_robot_conf(robotPt,_pickConf->getConfigStruct());
        print_msg(fct,"PLACE",3);
        p3d_place(robotPt,objectPt);
    }
    p3d_set_and_update_this_robot_conf(robotPt,init_conf);

    _place_traj=p3d_create_traj_from_localpath(robotPt,place_lp);
    place_lp->destroy(robotPt,place_lp);
    if(!_place_traj){
        print_msg(fct,"cannot allocate place trajectory");
        _returnReportcount++;
        if(_returnReportcountMax<_returnReportcount){
            _returnReportcountMax=_returnReportcount;
            SD->setMapEntry("returnReport","no_place_trajectory");
        }
        return false;
    }

    print_msg(fct,"finished with success",3);
    return true;
}

bool PickPlaceQuad::findPlaceExtractConf(){

    const char *fct="findPlaceExtractConf";
    if(!is_place){
        print_msg(fct,"is_place=false");
        return false;
    }
    p3d_rob* robotPt=mainAgent->getRobotStruct();
    p3d_rob* objectPt=_mainObject->getRobotStruct();
    configPt init_conf=_initConf->getConfigStruct();
    configPt place_conf=_placeConf->getConfigStruct();

    if(is_pick){
        p3d_set_and_update_this_robot_conf(robotPt,_pickConf->getConfigStruct());
        print_msg(fct,"PICK",3);
        p3d_pick(robotPt,ARM_EF_JNT,objectPt);
    }

    p3d_set_and_update_this_robot_conf(robotPt,place_conf);
    print_msg(fct,"PLACE",3);
    p3d_place(robotPt,objectPt);
    _place_extractConf=(confPtr_t)new Configuration(mainAgent,place_conf);
    configPt place_extract_conf=_place_extractConf->getConfigStruct();
    place_extract_conf[_param->x_dof]-=APPROACH_DIST*_place_graspDir[0];
    place_extract_conf[_param->y_dof]-=APPROACH_DIST*_place_graspDir[1];
    place_extract_conf[_param->z_dof]-=APPROACH_DIST*_place_graspDir[2];
    p3d_set_and_update_this_robot_conf(robotPt,place_extract_conf);
    bool col=p3d_col_test_robot(robotPt,0);
    if(col){
        print_msg(fct,"extract configuration in collision",2);
        _returnReportcount++;
        if(_returnReportcountMax<_returnReportcount){
            _returnReportcountMax=_returnReportcount;
            SD->setMapEntry("returnReport","no_place_extract_position");
        }
    }
    p3d_set_and_update_this_robot_conf(robotPt,place_conf);
    print_msg(fct,"PICK",3);
    p3d_pick(robotPt,ARM_EF_JNT,objectPt);
    if(is_pick){
        p3d_set_and_update_this_robot_conf(robotPt,_pickConf->getConfigStruct());
        print_msg(fct,"PLACE",3);
        p3d_place(robotPt,objectPt);
    }
    p3d_set_and_update_this_robot_conf(robotPt,init_conf);

    if(col)
        print_msg(fct,"finished with failure",2);
    else
        print_msg(fct,"finished with success",3);

    return !col;
}

bool PickPlaceQuad::findPlaceExtractTraj(){

    const char *fct="findPlaceExtractTraj";
    if(!is_place){
        print_msg(fct,"is_place=false");
        return false;
    }
    p3d_rob* robotPt=mainAgent->getRobotStruct();
    p3d_rob* objectPt=_mainObject->getRobotStruct();
    configPt init_conf=_initConf->getConfigStruct();
    configPt place_conf=_placeConf->getConfigStruct();
    configPt place_extract_conf=_place_extractConf->getConfigStruct();

    if(is_pick){
        p3d_set_and_update_this_robot_conf(robotPt,_pickConf->getConfigStruct());
        print_msg(fct,"PICK",3);
        p3d_pick(robotPt,ARM_EF_JNT,objectPt);
    }

    p3d_set_and_update_this_robot_conf(robotPt,place_conf);
    print_msg(fct,"PLACE",3);
    p3d_place(robotPt,objectPt);
    robotPt->lpl_type=LINEAR;
    p3d_localpath *place_extract_lp=p3d_linear_localplanner(robotPt,place_conf,place_extract_conf,0);
    int nb_col_test;
    if(!place_extract_lp || p3d_col_test_localpath(robotPt,place_extract_lp,&nb_col_test)){
        if(place_extract_lp){
            print_msg(fct,"place extract trajectory in collision",2);
            place_extract_lp->destroy(robotPt,place_extract_lp);
        }
        else
            print_msg(fct,"cannot allocate place extract trajectory");
        _returnReportcount++;
        if(_returnReportcountMax<_returnReportcount){
            _returnReportcountMax=_returnReportcount;
            SD->setMapEntry("returnReport","no_place_extract_trajectory");
        }
        robotPt->lpl_type=QUADSPLINE;
        p3d_set_and_update_this_robot_conf(robotPt,place_conf);
        print_msg(fct,"PICK",3);
        p3d_pick(robotPt,ARM_EF_JNT,objectPt);
        if(is_pick){
            p3d_set_and_update_this_robot_conf(robotPt,_pickConf->getConfigStruct());
            print_msg(fct,"PLACE",3);
            p3d_place(robotPt,objectPt);
        }
        p3d_set_and_update_this_robot_conf(robotPt,init_conf);
        return false;
    }
    _place_extract_traj=p3d_create_traj_from_localpath(robotPt,place_extract_lp);

    place_extract_lp->destroy(robotPt,place_extract_lp);
    robotPt->lpl_type=QUADSPLINE;
    p3d_set_and_update_this_robot_conf(robotPt,place_conf);
    print_msg(fct,"PICK",3);
    p3d_pick(robotPt,ARM_EF_JNT,objectPt);
    if(is_pick){
        p3d_set_and_update_this_robot_conf(robotPt,_pickConf->getConfigStruct());
        print_msg(fct,"PLACE",3);
        p3d_place(robotPt,objectPt);
    }
    p3d_set_and_update_this_robot_conf(robotPt,init_conf);

    if(!_place_extract_traj){
        print_msg(fct,"can not allocate place extract trajectory");
        _returnReportcount++;
        if(_returnReportcountMax<_returnReportcount){
            _returnReportcountMax=_returnReportcount;
            SD->setMapEntry("returnReport","no_place_extract_trajectory");
        }
        return false;
    }

    print_msg(fct,"finished with success",3);
    return true;
}

bool PickPlaceQuad::findPlaceTargetConf(){

    const char *fct="findPlaceTargetConf";
    if(!is_place){
        print_msg(fct,"is_place=false");
        return false;
    }
    p3d_rob* robotPt=mainAgent->getRobotStruct();
    p3d_rob* objectPt=_mainObject->getRobotStruct();
    configPt init_conf=_initConf->getConfigStruct();
    configPt place_approach_conf=_place_approachConf->getConfigStruct();

    if(is_pick){
        p3d_set_and_update_this_robot_conf(robotPt,_pickConf->getConfigStruct());
        print_msg(fct,"PICK",3);
        p3d_pick(robotPt,ARM_EF_JNT,objectPt);
    }

    if(is_pick)
        _place_targetConf=(confPtr_t)new Configuration(mainAgent,_pick_extractConf->getConfigStruct());
    else
        _place_targetConf=(confPtr_t)new Configuration(mainAgent,init_conf);
    configPt place_target_conf=_place_targetConf->getConfigStruct();
    place_target_conf[_param->x_dof]=place_approach_conf[_param->x_dof];
    place_target_conf[_param->y_dof]=place_approach_conf[_param->y_dof];
    place_target_conf[_param->z_dof]=place_approach_conf[_param->z_dof];
    place_target_conf[_param->rz_dof]=place_approach_conf[_param->rz_dof];
    p3d_set_and_update_this_robot_conf(robotPt,place_target_conf);
    if(p3d_col_test_robot(robotPt,0)){
        print_msg(fct,"place target configuration in collision",2);
        _returnReportcount++;
        if(_returnReportcountMax<_returnReportcount){
            _returnReportcountMax=_returnReportcount;
            SD->setMapEntry("returnReport","no_place_target_position");
        }
        if(is_pick){
            p3d_set_and_update_this_robot_conf(robotPt,_pickConf->getConfigStruct());
            print_msg(fct,"PLACE",3);
            p3d_place(robotPt,objectPt);
        }
        p3d_set_and_update_this_robot_conf(robotPt,init_conf);
        return false;
    }

    configPt place_target_object_conf=p3d_get_robot_config(objectPt);
    if(!place_target_object_conf){
        print_msg(fct,"cannot allocate configPt place_target_object_conf");
        if(is_pick){
            p3d_set_and_update_this_robot_conf(robotPt,_pickConf->getConfigStruct());
            print_msg(fct,"PLACE",3);
            p3d_place(robotPt,objectPt);
        }
        p3d_set_and_update_this_robot_conf(robotPt,init_conf);
        return false;
    }
    _place_targetObjectConf=(confPtr_t)new Configuration(_mainObject,place_target_object_conf,true);

    if(is_pick){
        p3d_set_and_update_this_robot_conf(robotPt,_pickConf->getConfigStruct());
        print_msg(fct,"PLACE",3);
        p3d_place(robotPt,objectPt);
    }
    p3d_set_and_update_this_robot_conf(robotPt,init_conf);
    print_msg(fct,"finished with success",3);
    return true;
}

bool PickPlaceQuad::findConfigurations(){

    const char *fct="findConfigurations";

    if(is_pick && !findPickConf()) return false;
    if(is_place){
        if(is_pick){
            int nb_try=-1;
            do{
                nb_try++;
            }while(nb_try<PLACE_NB_TRY && !findPlaceConf());
            if(nb_try==PLACE_NB_TRY) return false;
        }
        else
            if(!findPlaceConf()) return false;
    }
    if( is_pick && !findPickApproachConf())        return false;
    if(is_place && !findPlaceApproachConf())       return false;
    if( is_pick && !findPickTraj())                return false;
    if(is_place && !findPlaceTraj())               return false;
    if( is_pick && !findPickExtractConf())         return false;
    if(is_place && !findPlaceExtractConf())        return false;
    if( is_pick && !findPickExtractTraj())         return false;
    if(is_place && !findPlaceExtractTraj())        return false;
    if( is_pick && !findPickTargetConf())          return false;
    if(is_place && !findPlaceTargetConf())         return false;
    if( is_pick && !findPickTargetPlacedArmConf()) return false;

    print_msg(fct,"finished with success",3);
    return true;
}

bool PickPlaceQuad::findPickApproachTraj(){

    const char *fct="findPickApproachTraj";
    if(!is_pick){
        print_msg(fct,"is_pick=false");
        return false;
    }
    p3d_rob* robotPt=mainAgent->getRobotStruct();
    configPt init_conf=_initConf->getConfigStruct();

    ENV.setBool(Env::isHybrid,false);
    ENV.setBool(Env::quadOnly,false);
    ENV.setBool(Env::armOnly,false);
    robotPt->lpl_type=QUADSPLINE;

    _pick_approach_traj=NULL;
    if(!_pick_targetPlacedArmConf->equal(*_pick_approachConf)){
        p3d_localpath *approach_lp=p3d_quadspline_localplanner(robotPt,
                                   _pick_targetPlacedArmConf->getConfigStruct(),
                                   _pick_approachConf->getConfigStruct(),0);
        int nb_col_test;
        if(!approach_lp || p3d_col_test_localpath(robotPt,approach_lp,&nb_col_test)){
            if(approach_lp){
                print_msg(fct,"pick approach trajectory in collision",2);
                approach_lp->destroy(robotPt,approach_lp);
            }
            else
                print_msg(fct,"cannot allocate pick approach localpath");
            _returnReportcount++;
            if(_returnReportcountMax<_returnReportcount){
                _returnReportcountMax=_returnReportcount;
                SD->setMapEntry("returnReport","no_pick_approach_traj");
            }
            p3d_set_and_update_this_robot_conf(robotPt,init_conf);
            return false;
        }
        p3d_set_and_update_this_robot_conf(robotPt,init_conf);

        _pick_approach_traj=p3d_create_traj_from_localpath(robotPt,approach_lp);
        approach_lp->destroy(robotPt,approach_lp);
        if(!_pick_approach_traj){
            print_msg(fct,"cannot allocate pick approach trajectory");
            _returnReportcount++;
            if(_returnReportcountMax<_returnReportcount){
                _returnReportcountMax=_returnReportcount;
                SD->setMapEntry("returnReport","no_pick_approach_traj");
            }
            return false;
        }
    }

    print_msg(fct,"finished with success",3);
    return true;
}

bool PickPlaceQuad::findPickArmTraj(){

    const char *fct="findPickArmTraj";
    if(!is_pick){
        print_msg(fct,"is_pick=false");
        return false;
    }
    p3d_rob* robotPt=mainAgent->getRobotStruct();
    configPt init_conf=_initConf->getConfigStruct();

    robotPt->lpl_type=LINEAR;
    ENV.setBool(Env::armOnly,true);
    SD->getSDMapEntry("statistics")->incrementMapEntry("nbMPCalled");

    if(_pick_targetConf->equal(*_pick_targetPlacedArmConf)){
        print_msg(fct,"_pick_targetConf is equal to _pick_targetPlacedArmConf");
        p3d_set_and_update_this_robot_conf(robotPt,init_conf);
        return false;
    }

    PlanEnv->setDouble(PlanParam::timeLimitPlanning,MAX_TIME_PICK_ARM);
    _pick_arm_traj=p3d_planner_function(robotPt,
                  _pick_targetConf->getConfigStruct(),
                  _pick_targetPlacedArmConf->getConfigStruct());

    if(!_pick_arm_traj){
        print_msg(fct,"pick arm trajectory planning has failed",2);
        _returnReportcount++;
        if(_returnReportcountMax<_returnReportcount){
            _returnReportcountMax=_returnReportcount;
            SD->setMapEntry("returnReport","no_pick_arm_traj");
        }
        p3d_set_and_update_this_robot_conf(robotPt,init_conf);
        return false;
    }
    ENV.setBool(Env::armOnly,false);
    p3d_switch_to_quad(robotPt,_pick_arm_traj);
    print_msg(fct,"finished with success",3);
    return true;
}

bool PickPlaceQuad::findPickGotoTraj(){

    printf("findPickGotoTraj\n");

    const char *fct="findPickGotoTraj";
    if(!is_pick){
        print_msg(fct,"is_pick=false");
        return false;
    }
    p3d_rob* robotPt=mainAgent->getRobotStruct();
    configPt init_conf=_initConf->getConfigStruct();

    _pick_goto_traj=NULL;
    if(use_ghost){
        p3d_rob *ghost=_ghost->getRobotStruct();
        configPt qi=p3d_alloc_config(ghost);
        if(!qi)
            print_msg(fct,"cannot allocate configPt qi");
        else{
            qi[6]=init_conf[_param->x_dof];
            qi[7]=init_conf[_param->y_dof];
            qi[8]=init_conf[_param->z_dof];
            qi[11]=init_conf[_param->rz_dof];
            ghost->lpl_type=LINEAR;
            p3d_set_col_on(ghost);
            p3d_set_col_off(robotPt);
            p3d_set_current_robot(ghost);
            SD->getSDMapEntry("statistics")->incrementMapEntry("nbMPCalled");
            PlanEnv->setDouble(PlanParam::timeLimitPlanning,MAX_TIME_PICK_GOTO_LIN);
            _pick_goto_traj=p3d_planner_function(ghost,qi,
                                      _pick_targetGhostConf->getConfigStruct());
            p3d_set_col_on(robotPt);
            p3d_set_col_off(ghost);
            p3d_set_current_robot(robotPt);
            p3d_switch_to_quad_from_FF(robotPt,_pick_goto_traj);
            p3d_destroy_config(ghost,qi);
            ghost->tcur=NULL;
        }
    }
    else{
        ENV.setBool(Env::quadOnly,true);
        robotPt->lpl_type=QUADSPLINE;
        SD->getSDMapEntry("statistics")->incrementMapEntry("nbMPCalled");
        PlanEnv->setDouble(PlanParam::timeLimitPlanning,MAX_TIME_PICK_GOTO_DYN);
        _pick_goto_traj=p3d_planner_function(robotPt,init_conf,
                                           _pick_targetConf->getConfigStruct());
        ENV.setBool(Env::quadOnly,false);
    }

    p3d_set_and_update_this_robot_conf(robotPt,init_conf);

    if(!_pick_goto_traj){
        print_msg(fct,"pick goto trajectory planning has failed",2);
        _returnReportcount++;
        if(_returnReportcountMax<_returnReportcount){
            _returnReportcountMax=_returnReportcount;
            SD->setMapEntry("returnReport","no_pick_goto_traj");
        }
        return false;
    }

    print_msg(fct,"finished with success",3);
    return true;
}

bool PickPlaceQuad::findPlaceArmTraj(){

    const char *fct="findPlaceArmTraj";
    if(!is_place){
        print_msg(fct,"is_place=false");
        return false;
    }
    p3d_rob* robotPt=mainAgent->getRobotStruct();
    p3d_rob* objectPt=_mainObject->getRobotStruct();
    configPt init_conf=_initConf->getConfigStruct();

    if(is_pick){
        p3d_set_and_update_this_robot_conf(robotPt,_pickConf->getConfigStruct());
        print_msg(fct,"PICK",3);
        p3d_pick(robotPt,ARM_EF_JNT,objectPt);
    }

    ENV.setBool(Env::isHybrid,false);
    ENV.setBool(Env::quadOnly,false);
    ENV.setBool(Env::armOnly,true);
    robotPt->lpl_type=LINEAR;
    SD->getSDMapEntry("statistics")->incrementMapEntry("nbMPCalled");
    PlanEnv->setDouble(PlanParam::timeLimitPlanning,MAX_TIME_PLACE_ARM);
    _place_arm_traj=p3d_planner_function(robotPt,
                                         _place_targetConf->getConfigStruct(),
                                         _place_approachConf->getConfigStruct());
    p3d_switch_to_quad(robotPt,_place_arm_traj);

    if(is_pick){
        p3d_set_and_update_this_robot_conf(robotPt,_pickConf->getConfigStruct());
        print_msg(fct,"PLACE",3);
        p3d_place(robotPt,objectPt);
    }
    p3d_set_and_update_this_robot_conf(robotPt,init_conf);
    ENV.setBool(Env::armOnly,false);

    if(!_place_arm_traj){
        print_msg(fct,"place arm trajectory planning has failed",2);
        _returnReportcount++;
        if(_returnReportcountMax<_returnReportcount){
            _returnReportcountMax=_returnReportcount;
            SD->setMapEntry("returnReport","no_place_arm_traj");
        }
        return false;
    }

    print_msg(fct,"finished with success",3);
    return true;
}

bool PickPlaceQuad::findPlaceGotoTraj(){

    printf("findPlaceGotoTraj\n");

    const char *fct="findPlaceArmTraj";
    if(!is_place){
        print_msg(fct,"is_place=false");
        return false;
    }
    p3d_rob* robotPt=mainAgent->getRobotStruct();
    p3d_rob* objectPt=_mainObject->getRobotStruct();
    configPt init_conf=_initConf->getConfigStruct();

    if(is_pick){
        p3d_set_and_update_this_robot_conf(robotPt,_pickConf->getConfigStruct());
        print_msg(fct,"PICK",3);
        p3d_pick(robotPt,ARM_EF_JNT,objectPt);
    }

    ENV.setBool(Env::quadOnly,true);
    ENV.setBool(Env::armOnly,false);
    robotPt->lpl_type=QUADSPLINE;
    SD->getSDMapEntry("statistics")->incrementMapEntry("nbMPCalled");
    PlanEnv->setDouble(PlanParam::timeLimitPlanning,MAX_TIME_PLACE_GOTO);
    if(is_pick)
        _place_goto_traj=p3d_planner_function(robotPt,
                                           _pick_extractConf->getConfigStruct(),
                                          _place_targetConf->getConfigStruct());
    else
        _place_goto_traj=p3d_planner_function(robotPt,init_conf,
                                          _place_targetConf->getConfigStruct());
    if(is_pick){
        p3d_set_and_update_this_robot_conf(robotPt,_pickConf->getConfigStruct());
        print_msg(fct,"PLACE",3);
        p3d_place(robotPt,objectPt);
    }
    p3d_set_and_update_this_robot_conf(robotPt,init_conf);
    ENV.setBool(Env::quadOnly,false);

    if(!_place_goto_traj){
        print_msg(fct,"place goto trajectory planning has failed",2);
        _returnReportcount++;
        if(_returnReportcountMax<_returnReportcount){
            _returnReportcountMax=_returnReportcount;
            SD->setMapEntry("returnReport","no_goto_traj");
        }
        return false;
    }

    print_msg(fct,"finished with success",3);
    return true;
}

bool PickPlaceQuad::findTrajectories(){
    const char *fct="findTrajectories";
    if( is_pick && !findPickApproachTraj()) return false;
    if( is_pick && !findPickArmTraj())      return false;
    if(is_place && !findPlaceArmTraj())     return false;
    if( is_pick && !findPickGotoTraj())     return false;
    if(is_place && !findPlaceGotoTraj())    return false;
    print_msg(fct,"finished with success",3);
    return true;
}

bool PickPlaceQuad::computeSolutionTrajectories(int alternativeId){

    const char *fct="computeSolutionTrajectories";

    TaskSolution* TS=result->getAlternativeWithId(alternativeId);
    if(!TS){
        print_msg(fct,"no alternative for this task ");
        return false;
    }

    if(is_pick)
        _pick_targetConf=TS->getStopWS()->getRobotConf(mainAgent);
    else
        _place_targetConf=TS->getStopWS()->getRobotConf(mainAgent);

    findTrajectories();

    print_msg(fct,"finished with success",3);
    return true;
}

bool PickPlaceQuad::setToSolution(int solutionId, bool computeMP){

    const char *fct="setToSolution";

    TaskSolution* TS=new TaskSolution(result,solutionId);
    TS->setTrajComputed(computeMP);
    TS->setTaskConstraints(_currConstraints);
    TS->setStartWS(_WS);

    int next_id=0;
    TaskSubSolution* TSS5;
    WorldState* WS5;

    if(is_pick){

        next_id=5;

        TaskSubSolution* TSS1=new TaskSubSolution(TS,    "pick goto",0,    _pick_goto_traj,mainAgent,0);
        TaskSubSolution* TSS2=new TaskSubSolution(TS,     "pick arm",1,     _pick_arm_traj,mainAgent,0);
        TaskSubSolution* TSS3=NULL;
        if(_pick_approach_traj)
                         TSS3=new TaskSubSolution(TS,"pick approach",2,_pick_approach_traj,mainAgent,0);
        TaskSubSolution* TSS4=new TaskSubSolution(TS,         "pick",3,         _pick_traj,mainAgent,0);
                         TSS5=new TaskSubSolution(TS, "pick extract",4, _pick_extract_traj,mainAgent,0);

        if(!TSS3){
            TSS4->setId(2);
            TSS5->setId(3);
            next_id=4;
        }

        TSS1->setNavigation(false);
        TSS2->setNavigation(false);
        if(TSS3) TSS3->setNavigation(false);
        TSS4->setNavigation(false);
        TSS5->setNavigation(false);

        std::vector<TaskSubSolution*> v2;
        v2.push_back(TSS2);
        TSS1->setNext(v2);

        std::vector<TaskSubSolution*> v3;
        if(TSS3){
            v3.push_back(TSS3);
            TSS2->setNext(v3);
        }

        std::vector<TaskSubSolution*> v4;
        v4.push_back(TSS4);
        if(TSS3)
            TSS3->setNext(v4);
        else
            TSS2->setNext(v4);

        std::vector<TaskSubSolution*> v5;
        v5.push_back(TSS5);
        TSS4->setNext(v5);

        TSS1->setStartState(_WS);

        WorldState* WS1=new WorldState(_WS);
        if(!WS1->replaceConf(_pick_targetConf)){
            print_msg(fct,"error with robots names");
            return false;
        }
        TSS1->setEndState(WS1);
        TSS2->setStartState(WS1);

        WorldState* WS2=new WorldState(WS1);
        if(!WS2->replaceConf(_pick_targetPlacedArmConf)){
            print_msg(fct,"error with robots names");
            return false;
        }
        TSS2->setEndState(WS2);
        if(TSS3)
            TSS3->setStartState(WS2);
        else
            TSS4->setStartState(WS2);

        WorldState* WS3;
        if(TSS3){
            WS3=new WorldState(WS2);
            if(!WS3->replaceConf(_pick_approachConf)){
                print_msg(fct,"error with robots names");
                return false;
            }
            TSS3->setEndState(WS3);
            TSS4->setStartState(WS3);
        }

        WorldState* WS4;
        if(TSS3)
            WS4=new WorldState(WS3);
        else
            WS4=new WorldState(WS2);
        if(!WS4->replaceConf(_pickConf)){
            print_msg(fct,"error with robots names");
            return false;
        }
        WS4->addAttachement(mainAgent,_mainObject);
        TSS4->setEndState(WS4);
        TSS5->setStartState(WS4);

        WS5=new WorldState(WS4);
        if(!WS5->replaceConf(_pick_extractConf) || !WS5->replaceConf(_pick_extractObjectConf)){
            print_msg(fct,"error with robots names");
            return false;
        }
        TSS5->setEndState(WS5);

        if(!is_place){
            TS->setStopWS(WS5);
            WS5->setTSThatEndWithThisWS(TS);
        }

    }

    if(is_place){

        TaskSubSolution* TSS6=new TaskSubSolution(TS,   "place goto",next_id  ,   _place_goto_traj,mainAgent,0);
        TaskSubSolution* TSS7=new TaskSubSolution(TS,    "place arm",next_id+1,    _place_arm_traj,mainAgent,0);
        TaskSubSolution* TSS8=new TaskSubSolution(TS,        "place",next_id+2,        _place_traj,mainAgent,0);
        TaskSubSolution* TSS9=new TaskSubSolution(TS,"place extract",next_id+3,_place_extract_traj,mainAgent,0);

        TSS6->setNavigation(false);
        TSS7->setNavigation(false);
        TSS8->setNavigation(false);
        TSS9->setNavigation(false);

        if(is_pick){
            std::vector<TaskSubSolution*> v6;
            v6.push_back(TSS6);
            TSS5->setNext(v6);
        }

        std::vector<TaskSubSolution*> v7;
        v7.push_back(TSS7);
        TSS6->setNext(v7);

        std::vector<TaskSubSolution*> v8;
        v8.push_back(TSS8);
        TSS7->setNext(v8);

        std::vector<TaskSubSolution*> v9;
        v9.push_back(TSS9);
        TSS8->setNext(v9);

        WorldState* WS6;
        if(!is_pick){
            TSS6->setStartState(_WS);
            WS6=new WorldState(_WS);
        }
        else{
            TSS6->setStartState(WS5);
            WS6=new WorldState(WS5);
        }

        if(!WS6->replaceConf(_place_targetConf) || !WS6->replaceConf(_place_targetObjectConf)){
            print_msg(fct,"error with robots names");
            return false;
        }
        TSS6->setEndState(WS6);
        TSS7->setStartState(WS6);

        WorldState* WS7=new WorldState(WS6);
        if(!WS7->replaceConf(_place_approachConf) || !WS7->replaceConf(_place_approachObjectConf)){
            print_msg(fct,"error with robots names");
            return false;
        }
        TSS7->setEndState(WS7);
        TSS8->setStartState(WS7);

        WorldState* WS8=new WorldState(WS7);
        if(!WS8->replaceConf(_placeConf) || !WS8->replaceConf(_placeObjectConf)){
            print_msg(fct,"error with robots names");
            return false;
        }
        WS8->removeAttachements(mainAgent,_mainObject);
        TSS8->setEndState(WS8);
        TSS9->setStartState(WS8);

        WorldState* WS9=new WorldState(WS8);
        if(!WS9->replaceConf(_place_extractConf)){
            print_msg(fct,"error with robots names");
            return false;
        }
        TSS9->setEndState(WS9);

        TS->setStopWS(WS9);
        WS9->setTSThatEndWithThisWS(TS);

    }

    print_msg(fct,"finished with success",3);
    return true;
}

std::string PickPlaceQuad::getDescr()
{
    std::string s="Task: \nId \t\t= " + to_string(task_id) + "\nType \t\t= " + type;
    if(mainAgent)
        s+= "\nMain agent \t\t= " + mainAgent->getName();
    if(_mainObject)
        s+="\nMain object \t= " + _mainObject->getName();
    if(_supportObject)
        s+="\nSupport object \t= " + _mainObject->getName();
    return s;
}

std::string PickPlaceQuad::getTextFromValues()
{
    return "Task: " + type + ": mainAgent navigate to mainObject, pick it, navigate to supportObject and place mainObject there"
            + "\n    agents: "
            + "\n        mainAgent: [AGENT ID]"
            + "\n    objects: "
            + "\n        mainObject: [OBJECT ID]"
            + "\n        supportObject: [OBJECT ID]";
}

bool PickPlaceQuad::run(std::multimap<std::string,std::string> agents_name,
                   std::multimap<std::string,std::string> objects_name,
                   std::multimap<std::string,p3d_point> point3d,
                   std::multimap<std::string,std::string> additionalData,
                   WorldState *WS, bool computeMP){

    const char *fct="run";

    WS->load();
    _WS=WS;

    if( !initAll()                ||
        !setAgents(agents_name)   ||
        !setObjects(objects_name) ||
        !initialize()             )
    {return false;}


    bool res=false;
    int count=0;
    while(!res){
        if(VERBOSE>1){
            if(TO_LOG){
                char ad[100];
                sprintf(ad,"%s/log_move3d",getenv("HOME"));
                FILE *log=fopen(ad,"a");
                fprintf(log,"loop #%d\n",count);
                fclose(log);
            }
            else
                printf("loop #%d\n",count);
        }
        _returnReportcount=0;
        if(count++>nbMaxLoops)
            break;
        if(!findConfigurations())
            continue;
        if(computeMP)
            if(!findTrajectories())
                continue;
        if(!setToSolution(0,computeMP))
            continue;
        if(!testConstraints(0))
            continue;
        res=true;
    }

    if(res)
        SD->setMapEntry("returnReport","OK");

    if(res)
        print_msg(fct,"finished with success",3);
    else
        print_msg(fct,"finished with failure",2);

    return res;
}

void PickPlaceQuad::smoothSolution(int alternativeId){
    double old_cost=0;
    double new_cost=0;
    for(int i=0;i<result->getAlternativeWithId(alternativeId)->getNbsubSolPerRobot(getMainAgent());i++){
        TaskSubSolution* tss=result->getAlternativeWithId(alternativeId)->getTSSByRobot(getMainAgent()).at(i);
        if(strstr(tss->getSubSolutionName().c_str(),"goto") ||
           strstr(tss->getSubSolutionName().c_str(), "arm")){
            tss->getStartState()->load();
            p3d_traj *trajPt=tss->getTraj();
            API::Smoothing optimTraj(mainAgent,trajPt);
            old_cost+=optimTraj.cost();
            optimTraj.removeUselessNodes();
            optimTraj.runShortCut(PlanEnv->getInt(PlanParam::smoothMaxIterations));
            optimTraj.removeUselessNodes();
            new_cost+=optimTraj.cost();
            optimTraj.replaceP3dTraj(trajPt);
        }
    }
    result->getAlternativeWithId(alternativeId)->getStartWS()->load();
    double smoothing_gain=0;
    if(fabs(old_cost)>EPS10)
        smoothing_gain=100*(old_cost-new_cost)/old_cost;
    std::cout<<"smoothing gain: "<<smoothing_gain<<" % ";
}

double PickPlaceQuad::computeCost(int altId){return 0;}
bool PickPlaceQuad::findCandidateSolutions(){return true;}
bool PickPlaceQuad::getRandomSol(){return true;}
bool PickPlaceQuad::setPoints(std::multimap<std::string,p3d_point> points){return true;}
bool PickPlaceQuad::setdata(std::multimap<std::string,std::string> additionalData){return true;}

