#include "taskManager.hpp"
#include "API/project.hpp"
#include "HRI_costspace/HRICS_costspace.hpp"

using namespace std;

TaskManager::TaskManager()
{
    ENV.setBool(Env::printMightabilities,true);

}

void TaskManager::init()
{
    ENV.setBool(Env::printMightabilities, false);
    forceArmChoice = false;
    arm_l = Other;
    current_id = 0;
    current_worldState_id = 0;
    global_Project->getActiveScene()->createAgents();
    global_Project->getActiveScene()->createObjects();

    if(ENV.getBool(Env::enableHri)){
        if(!HRICS_CostFunctionCommon)
            HRICS_init();
        HRICS_CostFunctionCommon->costManager()->takeControlAndSet();
    }


    _PT = new PlanTree();
    _OP = new OnePlan(_PT);


    PlanEnv->setBool(PlanParam::withMaxIteration,true);
    PlanEnv->setBool(PlanParam::trajWithTimeLimit,true);
    PlanEnv->setDouble(PlanParam::timeLimitSmoothing,5.0);
    PlanEnv->setBool(PlanParam::withGainLimit,true);


}


TaskManager::~TaskManager()
{
    delete _PT;
    delete _OP;
//    delete MP;

}

PlanNode* TaskManager::computeTask(TasksId taskId,
                             multimap<string,string> agents_name,
                             multimap<string,string> objects_name,
                             multimap<string,p3d_point> point3d,
                             multimap<string,string> additionalData,
                             WorldState* WC, bool computeMP, TaskConstraints* taskConst)
{
    clock_t begin = clock();

    type_data td;
    td.id = taskId;
    td.additionalData = additionalData;
    td.agents = agents_name;
    td.objects = objects_name;
    td.points = point3d;

    bool isMainHuman = false;

    if (agents_name.count("mainAgent") >= 1)
    {
        if(agents_name.find("mainAgent")->second.find("HUMAN") != string::npos)
            isMainHuman = true;
    }

    Task* task = NULL;
    if (isMainHuman)
        task = createHumanTask(taskId);

    if (!task)
        task = createTask(taskId);

    if (!task)
        return NULL;

    task->setPlanTree(_PT);
    task->setCurrentDataType(td);

    if (!WC)
        return NULL;

//    WorldState* WS = new WorldState(WC);
//    addWorldState(WS);

    WC->load();

    TaskConstraints* TC = new TaskConstraints(taskConst);

    if (PlanEnv->getBool(PlanParam::usePreviousConstraint))
    {
        TC->addConstraint(WC->getTCForNextTask());
    }

    task->setConstraints(TC);


    bool test = task->run(agents_name,objects_name,point3d,additionalData,WC,computeMP);
    TaskSolution* TS = NULL;


    PlanNode* PN;

    if (test)
    {
        task->setId(current_id);
        current_id++;
        task_list.push_back(task);
        if (task->getResult())
            TS = task->getResult()->getAlternativeWithId(0);
    }

    PN = new PlanNode(_PT,TS);
    PN->setTask(task);
    PN->setGraph(task->getGraph());
    bool test2 = _PT->addPlanNode(PN, WC->getTSThatEndWithThisWS());

    clock_t endF = clock();
    task->SD->getSDMapEntry("times")->setMapEntry("timeComp",(double)(endF - begin)/CLOCKS_PER_SEC);
    setRequestToSolutionData(task,agents_name, objects_name,point3d,additionalData,taskConst);

    if (!test || !test2)
    {
        cout << "--------------------------" << endl;
        cout << "FAIL to compute task" << endl;

        if (!test2)
            cout << "faillure due to node already existing in the tree" << endl;
//        printTd(td);
        cout << "--------------------------" << endl;
        WC->load();
    }
    else
    {
        cout << "#################################" << endl;
        cout << "task computed SUCCESSfully" << endl;
//        cout << "task description: " << endl;
//        cout << task->getDescr() << endl;
//        cout << "cost = " << TS->getCost() <<endl;
        cout << "#################################" << endl;
        TS->getStopWS()->load();
        TS->addTypeData(td);
    }


    PN->SD = new SolutionData(task->SD);
    PN->SD->print();
    if (!test)
        cout << "--------------------------" << endl;
    else
        cout << "#################################" << endl;

    if (PN && PN->getTaskSolution())
    {
        popWorldStateWithId(PN->getTaskSolution()->getStartWS()->getId());
    }

    return PN;

}



double TaskManager::computeTaskHeuristic(TasksId taskId,
                                         multimap<string,string> agents_name,
                                         multimap<string,string> objects_name,
                                         multimap<string,p3d_point> point3d,
                                         multimap<string,string> additionalData,
                                         WorldState* WC, bool computeMP, TaskConstraints* taskConst)
{
//    clock_t begin = clock();

    bool isMainHuman = false;

    if (agents_name.count("mainAgent") >= 1)
    {
        if(agents_name.find("mainAgent")->second.find("HUMAN") != string::npos)
            isMainHuman = true;
    }

    Task* task = NULL;
    if (isMainHuman)
        task = createHumanTask(taskId);

    if (!task)
        task = createTask(taskId);

    if (!task)
        return P3D_HUGE;

    if (!WC)
        return P3D_HUGE;

    WorldState* WS = new WorldState(WC);

    WS->load();

    return task->heuristic(agents_name,objects_name,point3d,additionalData,WS,computeMP);

}


Task* TaskManager::createTask(TasksId taskId)
{

    Task* t = NULL;
    switch(taskId)
    {
    case drop :
        t = new Drop();
        break;
    case give :
        t = new Give();
        break;
    case pick :
        t = new Pick();
        break;
    case pickHuman :
        t = new PickHuman();
        break;
    case pick_multirrt :
        t = new Pick_MultiRRT();
        break;
    case place :
        t = new Place();
        break;
    case placeHuman :
        t = new PlaceHuman();
        break;
    case placeReachable :
        t = new PlaceReachable();
        break;
    case stackObj:
        t = new StackObj();
        break;
    case navigateAndPick:
        t = new NavigateAndPick();
        break;
    case place_multirrt :
        t = new Place_MultiRRT();
        break;
    case fakePlace:
        t = new FakePlace();
        break;
    case point :
        t = new Point();
        break;
    case fix :
        t = new fixOnWall();
        break;
    case navigateTo:
        t = new NavigateTo();
        break;
    case moveTo:
        t = new MoveTo();
        break;
    case pickPlaceQuad :
        t = new PickPlaceQuad();
        break;
    case pickQuad :
        t = new PickQuad();
        break;
    case placeQuad :
        t = new PlaceQuad();
        break;
    default :
        t = new DefaultTask();
        break;
    }
    return t;
}

Task* TaskManager::createHumanTask(TasksId taskId)
{

    Task* t = NULL;
    switch(taskId)
    {
    case pick :
    case pickHuman :
        t = new PickHuman();
        break;
    case place :
    case placeHuman :
        t = new PlaceHuman();
        break;
    case fakePlace:
        t = new FakePlace();
        break;
    case navigateTo:
        t= new NavigateTo();
        break;
    case moveTo:
        t = new MoveTo();
        break;
    default:
        t = new DefaultTask();
        break;
    }
    return t;
}

PlanNode* TaskManager::findTaskAlternative(int taskId, bool computeMP)
{
    clock_t begin = clock();
    Task* task = getTask(taskId);


    if (!task)
    {
        cout << "ERROR: no task with id " << taskId << endl;
        return NULL;
    }
//    task->timeComp = 0;
//    task->nbMPCalled =0;
//    task->nbGikCalled = 0;
//    task->nbTests =0;

    TaskSolution* TS = NULL;
    TS = task->findAlternative(computeMP);
//    if(!TS)
//    {
//        clock_t endFail = clock();
//        cout << "--------------------------" << endl;
//        cout << "FAIL to find alternative to task with id " << taskId << " within = " << (double)(endFail - begin)/CLOCKS_PER_SEC  << " seconds" << endl;
//        cout << "--------------------------" << endl;
////        return NULL;
//    }

    PlanNode* PN = new PlanNode(_PT,TS);
    PN->setTask(task);

    bool test = _PT->addPlanNode(PN, task->getResult()->getAlternativeWithId(0)->getStartWS()->getTSThatEndWithThisWS());

    clock_t endF = clock();
    task->SD->getSDMapEntry("times")->setMapEntry("timeComp",(double)(endF - begin)/CLOCKS_PER_SEC);
//    setRequestToSolutionData(task);


    PN->SD = new SolutionData(task->SD);

    if (!TS || !test)
    {
        cout << "--------------------------" << endl;
        cout << "FAIL to find alternative to task " << endl;
        if (!test)
            cout << "Fail due to not able to add it to tree" << endl;
        cout << "--------------------------" << endl;
    }
    else
    {
//        cout << "#################################" << endl;
//        cout << "find alternative to task SUCCESSfully" << endl;
//        cout << "task description: " <<endl;
//        cout << task->getDescr() <<endl;
//        cout << "cost = " << TS->getCost() <<endl;
        cout << "#################################" << endl;
        TS->addTypeData(task->getResult()->getAlternativeWithId(0)->getTypeData());
    }

    PN->SD->print();

    if (!test)
        cout << "--------------------------" << endl;
    else
        cout << "#################################" << endl;

    return PN;
}

bool TaskManager::computeTrajInTasks(int taskId, int alternativeId)
{
    clock_t begin = clock();
    Task* task = getTask(taskId);
    if (!task)
    {
        cout << "ERROR: no task with id " << taskId << endl;
        return false;
    }

    TaskSolution* TS = task->getResult()->getAlternativeWithId(alternativeId);
    if(!TS)
    {

        cout << "ERROR: in task with id " << taskId << "no alternative with id " << alternativeId << endl;
//        cout << "--------------------------" << endl;
//        cout << "FAIL to find alternative to task with id " << taskId << " within = " << (double)(endFail - begin)/CLOCKS_PER_SEC  << " seconds" << endl;
//        cout << "--------------------------" << endl;
        return false;
    }
    bool test = task->computeSolutionTrajectories(alternativeId);

    clock_t endA = clock();

        cout << "#################################" << endl;
        if (test)
            cout << "SUCCESSfully computed ";
        else
            cout << "FAIL to compute ";

        cout << "trajectories for task witthin  " << (double)(endA - begin)/CLOCKS_PER_SEC  << " seconds" << endl;
        cout << "task description: " <<endl;
        cout << task->getDescr() <<endl;
        cout << "#################################" << endl;


    return test;
}

Task* TaskManager::getTask(int id)
{
    for (unsigned int i = 0; i < task_list.size(); i++)
    {
        if (task_list.at(i)->getId() == id)
        {
            return task_list.at(i);
        }
    }
    Task* t = NULL;
    return t;
}


bool TaskManager::setArm(armLabel a)
{
    arm_l = a;
    return true;
}


void TaskManager::clearTasks()
{
    for (unsigned int i = 0; i < task_list.size(); i++)
    {
        delete task_list.at(i);
    }
    task_list.clear();
}

int TaskManager::addWorldState(WorldState* WC)
{
    worldState_list.push_back(WC);
//    for (unsigned int i = 0; i<worldState_list.size();i++)
//    {
//        WorldState* WSi = worldState_list.at(i);
//        if (WC->isEqualTo(WSi))
//            return WSi->getId();
//    }
//    int cid = current_worldState_id++;
//    WC->setId(cid);
//    worldState_list.push_back(WC);

//    return cid;
    return 0;
}

WorldState* TaskManager::getWorldStateById(int id)
{
    WorldState* WS = NULL;
    bool found = false;
    unsigned int i=0,j=0;


    vector<PlanNode*> vPN = getPlanTree()->getAllNodes();
    while (i < vPN.size() && !found)
    {
        if (vPN.at(i)->getTaskSolution())
        {
            vector<TaskSubSolution*> vTSS = vPN.at(i)->getTaskSolution()->getAllTSS();
            while (j < vTSS.size() && !found)
            {
                TaskSubSolution* TSS = vTSS.at(j);
                if (TSS->getStartState()->getId() == id)
                {
                    WS = TSS->getStartState();
                    found = true;
                }
                else if (TSS->getEndState()->getId() == id)
                {
                    WS = TSS->getEndState();
                    found = true;
                }
                j++;
            }
        }
        i++;
    }

    if (!found)
    {
        i = 0;
        while (i < worldState_list.size() && !found)
        {
            if (id == worldState_list.at(i)->getId())
            {
                found = true;
                WS = worldState_list.at(i);
            }
            i++;
        }
    }
    return WS;


//    for (unsigned int i = 0; i < worldState_list.size(); i ++)
//    {
//        if (worldState_list.at(i)->getId() == id)
//        {
//            return worldState_list.at(i);
//        }
//    }
//    WorldState* WC= NULL;
//    return WC;
}

WorldState* TaskManager::popWorldStateWithId(int id)
{
    WorldState* WS = NULL;
    unsigned int i = -1;
    std::vector<WorldState*>  tmp;

    while (i < worldState_list.size())
    {
        i++;
        if (id == worldState_list.at(i)->getId())
        {
            WS = worldState_list.at(i);
        }
        else
        {
            tmp.push_back(worldState_list.at(i));
        }
    }

    worldState_list = tmp;

    for (unsigned int i = 0; i < worldState_list.size(); i ++)
    {
        cout << "id left = " << worldState_list.at(i)->getId() << endl;
    }
    return WS;
}

void TaskManager::clearWorldStates()
{
    for (unsigned int i = 0; i< worldState_list.size(); i++)
    {
           delete worldState_list.at(i);
    }
    worldState_list.clear();
}

bool iequals(const string& a, const string& b)
{
    unsigned int sz = a.size();
    if (b.size() != sz)
        return false;
    for (unsigned int i = 0; i < sz; ++i)
        if (tolower(a[i]) != tolower(b[i]))
            return false;
    return true;
}


TasksId TaskManager::getTaskIdFromString(string value)
{
    if (iequals(value,"pick"))
        return pick;
    else if (iequals(value,"pickquad"))
        return pickQuad;
    else if (iequals(value,"placequad"))
        return placeQuad;
    else if (iequals(value,"pickplacequad"))
        return pickPlaceQuad;
    else if (iequals(value,"place"))
        return place;
    else if (iequals(value,"pickAndPlace"))
        return pickAndPlace;
    else if (iequals(value,"makeAccessible"))
        return makeAccessible;
    else if (iequals(value,"placeReachable"))
        return placeReachable;
    else if (iequals(value,"exhibit"))
        return exhibit;
    else if (iequals(value,"pickAndShow"))
        return pickAndShow;
    else if (iequals(value,"give"))
        return give;
    else if (iequals(value,"pickAndGive"))
        return pickAndGive;
    else if (iequals(value,"point"))
        return point;
    else if (iequals(value,"fix"))
        return fix;
    else if (iequals(value,"drop"))
        return drop;
    else if (iequals(value,"navigateto"))
        return navigateTo;
    else if (iequals(value,"stackObj"))
        return stackObj;
    else if (iequals(value,"navigateAndPick"))
        return navigateAndPick;
    else if (iequals(value,"fakePlace"))
        return fakePlace;
    else if (iequals(value,"moveTo"))
        return moveTo;
    else
        return noType;
}

TasksId TaskManager::getTaskIdFromString(char* value)
{
    string str = value;
    return getTaskIdFromString(str);
}

void TaskManager::printAllTasks(bool verbose)
{
    for(unsigned int i = 0; i < task_list.size(); i++)
    {
        Task* t = task_list.at(i);
        cout << "-----------------" << endl;
        cout << "task id = " << t->getId() << " and nb alternatives = " << t->getResult()->getNbAlternatives() << endl;
        if (verbose)
        {
            cout << t->getDescr() << endl;
        }
        cout << "-----------------" << endl;
    }
}

void TaskManager::printOnePlan(bool verbose)
{
    cout << "the plan is in this order : " << endl;
    for(unsigned int i = 0; i < _OP->getPlan().size(); i++)
    {
        PlanNode* PN = _OP->getPlan().at(i);
        Task* t = PN->getTaskSolution()->getTask();
        cout << "-----------------" << endl;
        cout << "task id = " << t->getId() << " with alternatives id = " << PN->getTaskSolution()->getId() << endl;
        if (verbose)
        {
            cout << t->getDescr() << endl;
        }
        cout << "-----------------" << endl;
    }
}

string TaskManager::printAllTextNeeded()
{
    string s = "";
    for(unsigned int i = 0; i < nbOfTaskId; i++)
    {
        TasksId tI = static_cast<TasksId>(i);
        Task* t = createTask(tI);
        if (t)
        {
            string tmp = t->getTextFromValues();
            if (tmp != "")
            {
                s += tmp;
                s += "\n  \n";
            }
        }
        delete t;
    }
    return s;
}

 void TaskManager::printTd(type_data td)
 {
     cout << "agents = " << endl;
     for (multimap<string, string>::iterator itr = td.agents.begin(); itr != td.agents.end(); itr++ ) {
         cout << "key = " << itr->first << " value = " << itr->second << endl;
     }
     cout << "objects = " << endl;
     for (multimap<string, string>::iterator itr = td.objects.begin(); itr != td.objects.end(); itr++ ) {
         cout << "key = " << itr->first << " value = " << itr->second << endl;
     }
     cout << "data = " << endl;
     for (multimap<string, string>::iterator itr = td.additionalData.begin(); itr != td.additionalData.end(); itr++ ) {
         cout << "key = " << itr->first << " value = " << itr->second << endl;
     }
     //TODO:Points
 }

 void TaskManager::setRequestToSolutionData(Task* t,
                                            multimap<string, string> agents_name,
                                            multimap<string, string> objects_name,
                                            multimap<string, p3d_point> point3d,
                                            multimap<string, string> additionalData,
                                            TaskConstraints *)
 {
     t->SD->getSDMapEntry("request")->setMapEntry("task",t->getType());
     for (multimap<string, string>::iterator itr = agents_name.begin(); itr != agents_name.end(); itr++ ) {
         t->SD->getSDMapEntry("request")->getSDMapEntry("agents")->setMapEntry(itr->first,itr->second);
     }
     for (multimap<string, string>::iterator itr = objects_name.begin(); itr != objects_name.end(); itr++ ) {
         t->SD->getSDMapEntry("request")->getSDMapEntry("objects")->setMapEntry(itr->first,itr->second);
     }
     for (multimap<string, string>::iterator itr = additionalData.begin(); itr != additionalData.end(); itr++ ) {
         t->SD->getSDMapEntry("request")->getSDMapEntry("data")->setMapEntry(itr->first,itr->second);
     }
     for (multimap<string, p3d_point>::iterator itr = point3d.begin(); itr != point3d.end(); itr++ ) {
         t->SD->getSDMapEntry("request")->getSDMapEntry("points")->setMapEntry(itr->first+"x",itr->second.x);
         t->SD->getSDMapEntry("request")->getSDMapEntry("points")->setMapEntry(itr->first+"y",itr->second.y);
         t->SD->getSDMapEntry("request")->getSDMapEntry("points")->setMapEntry(itr->first+"z",itr->second.z);
     }
 }
