#include "HRICS_CF_JsonConf.h"

using namespace std;

namespace HRICS {
namespace CF {

INIT_MOVE3D_STATIC_LOGGER(JsonConf,"planners.hrics.cf.jsonconf");


string JsonConf::getPathBodyWeights(const Sem::AgentType &agent_type)
{
    std::string path = _path;
    path.append("/HRICS/BodyWeights/");
    if(agent_type.isDefined()){
        path.append(agent_type.toString() + ".json");
    }else{
        path="";
        M3D_ERROR ("No string defined for agent type "<<agent_type.toHriAgentType());
    }
    return path;
}

string JsonConf::getPathJointWeights(const Sem::AgentType &agent_type)
{
    std::string path = _path;
    path.append("/HRICS/JointWeights/");
    if(agent_type.isDefined()){
        path.append(agent_type.toString() + ".json");
    }else{
        path="";
        M3D_ERROR ("No string defined for agent type "<<agent_type.toHriAgentType());
    }
    return path;
}

JsonConf::JsonConf(CostFunctionCommon *common, string path):
    _common(common)
{
    if(path.size()==0)
        _path = getenv("HOME_MOVE3D");
    else
        _path=path;
}

JsonConf::~JsonConf()
{

}

bool JsonConf::writeBodiesWeights()
{
    bool ok;
    const std::map<HRI_AGENT_TYPE, std::map<CostSettingType,std::map<std::string,double> > > &bodyWeights = _common->bodyWeightsByName();

    for ( std::map<HRI_AGENT_TYPE, std::map<CostSettingType,std::map<std::string,double> > >::const_iterator it=bodyWeights.begin();
          it!=bodyWeights.end() ; ++it)
    {
        Sem::AgentType agent_t(it->first);
        std::string path;
        std::ofstream ofs;
        ok=true;
        if (agent_t.toString().size() == 0){
            M3D_ERROR("no string defined for agent type "<<it->first);
            ok=false;
        }
        if(ok){
            path = getPathBodyWeights(it->first);
            ok = path.size();
        }
        if (ok){
            ofs.open(path.c_str());
            if(!ofs){
                ok=false;
                M3D_ERROR("fail to open file for writting :"<<path);
            }
        }
        if(ok){
            Json::Value agent;
            agent["agent_type"]=agent_t.toString();
            for(std::map<CostSettingType,std::map<std::string,double> >::const_iterator itcost = it->second.begin();
                itcost != it->second.end(); ++itcost)
            {
                Json::Value arr;
                for(std::map<std::string,double>::const_iterator itv = itcost->second.begin(); itv!=itcost->second.end(); ++itv){
                    arr[itv->first]=itv->second;
                }
                agent["weights"][itcost->first.toString()]=arr;
            }
            ofs << agent <<endl;
            ofs.close();
        }
    }
    return ok;
}
bool JsonConf::readBodiesWeights(){
    bool ok =true;
    std::set<Sem::AgentType> agent_types;
    for(unsigned int i=0;i<global_Project->getActiveScene()->getNumberOfRobots();i++){
        Robot *r = global_Project->getActiveScene()->getRobot(i);
        if(r->getHriAgent()){
            Sem::AgentType t(r->getHriAgent()->type);
            if(t.isDefined()){
                agent_types.insert(t);
            }else{
                M3D_WARN("no string defined for agent type "<<t.toHriAgentType());
            }
        }
    }
    for(std::set<Sem::AgentType>::const_iterator it=agent_types.begin();it!=agent_types.end();++it){
        readBodiesWeights(*it);
    }
    return ok;
}

bool JsonConf::readBodiesWeights(Sem::AgentType const &agent_type)
{
    bool ok=false;
    std::ifstream ifs;
    std::string path=getPathBodyWeights(agent_type);
    ifs.open(path.c_str());
    ok=(ifs);
    if(ok){
        std::map<CostSettingType,std::map<std::string, double > > bodyweights;
        Json::Value agent;
        try{
            ifs>>agent;
        }catch(std::runtime_error e){
            M3D_ERROR("error while parsing json file "<<path<<":");
            M3D_ERROR(e.what());
            ok=false;
        }
        ifs.close();
        if(ok){
            Sem::AgentType agent_type_read(agent["agent_type"].asString());
            if(agent_type_read != agent_type){
                ok=false;
                M3D_ERROR("agent_type in file "<<path<<" differs from file name");
            }
        }
        if(ok){
            Json::Value weights = agent["weights"];
            for(Json::Value::iterator itw = weights.begin(); itw!=weights.end(); ++itw){
                Json::Value &arr=*itw;
                for(Json::Value::iterator ita = arr.begin() ; ita != arr.end() ; ++ita){
                    bodyweights[CostSettingType(itw.key().asString())].insert(std::make_pair(ita.key().asString(),(*ita).asDouble()));
                }
            }
        }
        _common->addBodyWeightsByName(agent_type.toHriAgentType(),bodyweights);
    }
    return ok;
}

bool JsonConf::writeJointWeights()
{
    bool ok =false;
    const std::map<HRI_AGENT_TYPE, std::map<CostSettingType,std::map<int,double> > > &jointWeights = _common->jointWeights();

    for ( std::map<HRI_AGENT_TYPE, std::map<CostSettingType,std::map<int,double> > >::const_iterator it=jointWeights.begin();
          it!=jointWeights.end() ; ++it)
    {
        Sem::AgentType agent_t(it->first);
        std::string path;
        std::ofstream ofs;
        ok=true;
        if (agent_t.toString().size() == 0){
            M3D_ERROR("no string defined for agent type "<<it->first);
            ok=false;
        }
        if(ok){
            path = getPathJointWeights(it->first);
            ok = path.size();
        }
        if (ok){
            ofs.open(path.c_str());
            if(!ofs){
                ok=false;
                M3D_ERROR("fail to open file for writting :"<<path);
            }
        }
        if(ok){
            Json::Value agent;
            agent["agent_type"]=agent_t.toString();
            Robot *r=_common->getRobotHavingType(agent_t.toHriAgentType());
            for(std::map<CostSettingType,std::map<int,double> >::const_iterator itcost = it->second.begin();
                itcost != it->second.end(); ++itcost)
            {
                Json::Value arr;
                for(std::map<int,double>::const_iterator itv = itcost->second.begin(); itv!=itcost->second.end(); ++itv){
                    Json::Value v;
                    v["n"]=r->getJoint(itv->first)->getName();
                    v["w"]=itv->second;
                    arr.append(v);
                }
                agent["weights"][itcost->first.toString()]=arr;
            }
            ofs << agent <<endl;
            ofs.close();
        }
    }
    return ok;
}

bool JsonConf::readJointWeights(const Sem::AgentType &agent_type){
    bool ok=false;
    std::ifstream ifs;
    std::string path=getPathJointWeights(agent_type);
    ifs.open(path.c_str());
    ok=(ifs);
    if(ok){
        std::map<CostSettingType,std::map<int,double > > jointWeights;
        Json::Value agent;
        try{
            ifs>>agent;
        }catch(std::runtime_error e){
            M3D_ERROR("error while parsing json file "<<path<<":");
            M3D_ERROR(e.what());
            ok=false;
        }
        ifs.close();
        if(ok){
            Sem::AgentType agent_type_read(agent["agent_type"].asString());
            if(agent_type_read != agent_type){
                ok=false;
                M3D_ERROR("agent_type in file "<<path<<" differs from file name");
            }
        }
        if(ok){
            Json::Value weights = agent["weights"];
            for(Json::Value::iterator itw = weights.begin(); itw!=weights.end(); ++itw){
                Json::Value &arr=*itw;
                if(arr.isArray()){
                    std::map<int,double> &jw=jointWeights[CostSettingType(itw.key().asString())];
                    Robot *r = _common->getRobotHavingType(agent_type.toHriAgentType());
                    for(Json::Value::iterator ita = arr.begin() ; ita != arr.end() ; ++ita){
                        int id=r->getJoint((*ita)["n"].asString())->getId();
                        jw[id]=(*ita)["w"].asDouble();
                    }
                }else if(arr.isMember("equals")){
                    //refer to other CostSettingType
                    CostSettingType refer(arr["equals"].asString());
                    if(jointWeights.count(refer) !=0 ){
                        jointWeights[CostSettingType(itw.key().asString())].insert(jointWeights[refer].begin(),jointWeights[refer].end());
                        M3D_DEBUG("cost setting "<<itw.key().asString()<<" equals to found setting "<<refer.toString());
                    }else{
                        M3D_ERROR("In file "<<path<<" cost setting "<<itw.key().asString() << "refers to yet undefined setting "<<refer.toString());
                    }

                }
            }
        }
        _common->addJointWeights(agent_type.toHriAgentType(),jointWeights);
    }
    return ok;
}

bool JsonConf::readJointWeights()
{
    bool ok =true;
    std::set<Sem::AgentType> agent_types;
    for(unsigned int i=0;i<global_Project->getActiveScene()->getNumberOfRobots();i++){
        Robot *r = global_Project->getActiveScene()->getRobot(i);
        if(r->getHriAgent()){
            Sem::AgentType t(r->getHriAgent()->type);
            if(t.isDefined()){
                agent_types.insert(t);
            }else{
                M3D_WARN("no string defined for agent type "<<t.toHriAgentType());
            }
        }
    }
    for(std::set<Sem::AgentType>::const_iterator it=agent_types.begin();it!=agent_types.end();++it){
        readJointWeights(*it);
    }
    return ok;

}

} // namespace CF
} // namespace HRICS

