#include "HRICS_CF_VirtualCostFunction.h"
#include "HRICS_costfunction.h"

using namespace std;
namespace HRICS {
namespace CF {

VirtualCostFunction::VirtualCostFunction(CostFunctionCommon *common,
                                         double min, double max, double low, double up, double maxToMeanRatio,
                                         VirtualCostFunction::FunctionShape shape):
    _common(common),
    _upperBound(up),_lowerBound(low),
    _minCost(min),_maxCost(max),
    _maxToMeanRatio(maxToMeanRatio),
    _functionShape(shape)
{}

CostFunctionCommon *VirtualCostFunction::commonCF() const
{
    return _common;
}

void VirtualCostFunction::setCommonCF(CostFunctionCommon *common)
{
    _common = common;
}


double VirtualCostFunction::getCost(double x)
{
    switch(functionShape()){
    case LINEAR:
        return getCostLin(x);
    case EXP:
        return getCostExp(x);
    case EXP_INV:
        return getCostExpInv(x);
    case SIGMOID:
        return getCostSigmoid(x);
    case CENTERED_GAUSS_INV:
        return getCostCenteredGaussInv(x);
    default:
        return numeric_limits<double>::quiet_NaN();
    }
}

double VirtualCostFunction::getCostExpInv(double x)
{
    assert(_maxCost> _minCost);
    double x_norm=(x-_lowerBound)/(_upperBound - _lowerBound);
    double c=( _maxCost-_minCost)*( 1 - std::exp(-x_norm*5)) + _minCost;
    return std::max(c,_minCost);
}

double VirtualCostFunction::getCostLin(double x)
{
    assert(_maxCost> _minCost);
    double x_norm=(x-_lowerBound)/(_upperBound - _lowerBound);
    double c= _minCost + (_maxCost - _minCost ) * x_norm;
    return std::min(std::max(c,_minCost),_maxCost);
}

double VirtualCostFunction::getCostExp(double x)
{
    assert(_maxCost> _minCost);
    assert(false&&"not implemented yet");
}

double VirtualCostFunction::getCostSigmoid(double x)
{
    assert(_maxCost> _minCost);
    double x_norm=(x-_lowerBound)/(_upperBound - _lowerBound);
    double c=( _maxCost-_minCost)*( 1 / (1+ std::exp(-x_norm*12+6))) + _minCost;
    assert(c<=_maxCost );
    assert(c>=_minCost);
    return c;
}

double VirtualCostFunction::getCostCenteredGaussInv(double x)
{
    assert(_maxCost> _minCost);
    double u=(_upperBound+_lowerBound)/2;
    double s=(_upperBound - u)/3;
    double c=1-std::exp(-0.5*std::pow((x-u)/s,2));
    //denormalize:
    c=(_maxCost-_minCost)*c + _minCost;
    assert(c<=_maxCost );
    assert(c>=_minCost);
    return c;
}

double VirtualCostFunction::getCostMaxMean(double max, double mean)
{
    assert(_maxToMeanRatio >=0);
    assert(_maxToMeanRatio <=1);
    return _maxToMeanRatio * max + (1-_maxToMeanRatio) * mean;
}


VirtualCostFunction::FunctionShape VirtualCostFunction::functionShape() const
{
    return _functionShape;
}
void VirtualCostFunction::setFunctionShape(const FunctionShape &functionShape)
{
    _functionShape = functionShape;
}

double VirtualCostFunction::upperBound() const
{
    return _upperBound;
}

void VirtualCostFunction::setUpperBound(double upperBound)
{
    _upperBound = upperBound;
}
double VirtualCostFunction::lowerBound() const
{
    return _lowerBound;
}

void VirtualCostFunction::setLowerBound(double lowerBound)
{
    _lowerBound = lowerBound;
}
double VirtualCostFunction::minCost() const
{
    return _minCost;
}

void VirtualCostFunction::setMinCost(double minCost)
{
    _minCost = minCost;
}
double VirtualCostFunction::maxCost() const
{
    return _maxCost;
}

void VirtualCostFunction::setMaxCost(double maxCost)
{
    _maxCost = maxCost;
}
double VirtualCostFunction::maxToMeanRatio() const
{
    return _maxToMeanRatio;
}

void VirtualCostFunction::setMaxToMeanRatio(double maxToMeanRatio)
{
    _maxToMeanRatio = maxToMeanRatio;
}


} // namespace CF
} // namespace HRICS

