#include "HRICS_costfunction.h"
#include "utils/Geometry.h"

#include "HRICS_CF.h"
#include "cost_space.hpp"
#include <boost/bind.hpp>
#include "HRICS_CF_JsonConf.h"

#include <algorithm>

namespace HRICS {
namespace CF {
using namespace std;

INIT_MOVE3D_STATIC_LOGGER(CostFunctionCommon,"planners.hrics.cf.common");

CostFunctionCommon::CostFunctionCommon(Robot* rob, vector<Robot*> humans, vector<Joint*> costJoints, CostManager *costManger):
    _pathLengthCost(0)
{
    setRobot(rob);
    setCostJoints(costJoints);
    setHumans(humans);

    if(costManger){
        setCostManager(costManger);
    }else{
        setCostManager(CostManager::mainCostManager());
    }

    readExternalConf();

    this->extractBodyWeightsForComputation();

    createCosts();
    //addCostsToCostSpace();

    //saveExternalConf();
}

CostFunctionCommon::CostFunctionCommon(bool guess_settings):
    _pathLengthCost(0)
{
    readExternalConf();
    if(guess_settings){
        updateRobot();
        updateHumans(); //it calls updateHumanPos()
        //TODO cost joints??
    }
    setCostManager(CostManager::mainCostManager());
    this->extractBodyWeightsForComputation();
    createCosts();
    //addCostsToCostSpace();

    //saveExternalConf();
}

CostFunctionCommon::~CostFunctionCommon()
{
    deleteCosts();
}

void CostFunctionCommon::readExternalConf()
{
    JsonConf json(this);
    json.readBodiesWeights();
    cout<<"found "<<_bodyWeightsByName.size() << " agent body weights settups"<<endl;
    this->cleanBodyWeightsByName();
    json.readJointWeights();
}

void CostFunctionCommon::saveExternalConf()
{
    JsonConf json(this);
    json.writeBodiesWeights();
    json.writeJointWeights();
}

void CostFunctionCommon::extractBodyWeightsForComputation()
{
    for(map<HRI_AGENT_TYPE, std::map<CostSettingType,map<string,double> > >::iterator it = _bodyWeightsByName.begin();
        it != _bodyWeightsByName.end() ; ++it)
    {

        Robot* sample_rob = getRobotHavingType(it->first);
        if(sample_rob){
            for (map<CostSettingType,map<string,double> >::iterator itcost= it->second.begin();
                 itcost != it->second.end(); ++itcost)
            {
                vector<pair<int,double> > bodies;
                for(map<string,double>::iterator itb = itcost->second.begin();
                    itb != itcost->second.end(); ++itb)
                {
                    string name=sample_rob->getName();
                    name.append(".");
                    name.append(itb->first); // reconstruct the complete name of the joint
                    int index=0; // what we are searching for!
                    double weight = itb->second;
                    while( index < sample_rob->getRobotStruct()->no
                           && name.compare(sample_rob->getRobotStruct()->o[index]->name ) !=0 )
                    {index++;}

                    if(index < sample_rob->getRobotStruct()->no){
                        bodies.push_back(make_pair(index,weight));
                    }else{
                        M3D_WARN("the body "<<name<<" was not found in the robot struct, but is given in the body weights");
                    }
                }

                _bodyWeights[it->first][itcost->first] = bodies;
            }
        }
    }
}

void CostFunctionCommon::createCosts()
{
    deleteCosts();
    _costFunctions[BodiesDistance::getName()]=new BodiesDistance(this,0,1,2,0,0.5,VirtualCostFunction::LINEAR);
    _costFunctions[BodiesDistance::getName().append("BB")]= new BodiesDistance(this,0,1,2,0,0.5,VirtualCostFunction::LINEAR,true);
    _costFunctions[TimeToCollision::getName()] = new TimeToCollision(this,0,1,1,0,0.5,VirtualCostFunction::LINEAR);
    _costFunctions[KeepItLow::getName()] = new KeepItLow(this,0,1,0,3,0.5,VirtualCostFunction::LINEAR);
    _costFunctions[KeepItLow::getName().append("Sigmoid")] = new KeepItLow(this,0,1,0,3,0.5,VirtualCostFunction::SIGMOID);
    _costFunctions[TorsoHeight::getName()] = new TorsoHeight(this,0,1,0,1,0.5,VirtualCostFunction::CENTERED_GAUSS_INV);
    _costFunctions["virtualCeiling"] = new KeepItLow(this,0,1e6,1.9,2,1,VirtualCostFunction::LINEAR);
    _costFunctions[KeepManipulatedVertical::getName()] = new KeepManipulatedVertical(this,0,1e6,0,1,1,VirtualCostFunction::SIGMOID);

    _pathLengthCost = new PathLengthCost(this,0,1,0,1,1,VirtualCostFunction::LINEAR);
    M3D_TRACE("created "<<_costFunctions.size()<<" Costs");
}

void CostFunctionCommon::deleteCosts()
{
    for(map<string,VirtualCostFunction*>::iterator it= _costFunctions.begin();
        it!=_costFunctions.end(); ++it)
    {
        global_costSpace->deleteCost(it->first);
        if (it->second){
            delete it->second;
            it->second = 0;
        }
    }
    _costFunctions.clear();

    if(_pathLengthCost){
        global_costSpace->deleteCostTraj(_pathLengthCost->getName());
        delete _pathLengthCost;
        _pathLengthCost = 0;
    }
}

void CostFunctionCommon::addCostsToCostSpace()
{

    for(map<string,VirtualCostFunction*>::const_iterator it= _costFunctions.begin();
        it!=_costFunctions.end(); ++it)
    {
        global_costSpace->addCost(it->first,boost::bind(HRICS_CF_computeCost, this, it->first,_1) );
        //global_costSpace->setCost(it->first);
        cout <<"add cost "<<it->first<<endl;
    }

    global_costSpace->addCostTrajectories(_pathLengthCost->getName(),
                                          boost::bind(static_cast<double (PathLengthCost::*)(std::vector<LocalPath*> &)>(&PathLengthCost::getCost),
                                                      _pathLengthCost,_1));
    global_costSpace->setCostTraj(_pathLengthCost->getName());
    cout <<"add path cost "<<_pathLengthCost->getName()<<endl;
    global_costSpace->addCostTrajectories("CostAggregationCostManager",
                                          boost::bind(
                                              static_cast<double (CostManager::*)(std::vector<LocalPath*> &)>(&CostManager::getCostAggregationNormalized),
                                                      this->_costManager,_1));
    cout <<"add path cost "<<"CostAggregationCostManager"<<endl;
}

double CostFunctionCommon::getCost(string name, Configuration &q)
{
    return getCost(_costFunctions[name],q);
}

double CostFunctionCommon::getCost(VirtualCostFunction *cf, Configuration &q)
{
    q.getRobot()->setAndUpdate(q);
    assert(q.getRobot()->getName() == getRobot()->getName());
    return cf->getCost();
}

vector<string> CostFunctionCommon::listCosts()
{
    vector<string> names;
    for(map<string,VirtualCostFunction*>::const_iterator it= _costFunctions.begin();
        it!=_costFunctions.end(); ++it){
        names.push_back(it->first);
    }
    return names;
}

void CostFunctionCommon::prepareForSmoothing()
{
    API::Trajectory path=getRobot()->getCurrentTraj();
    prepareForSmoothing(path);
}

void CostFunctionCommon::prepareForSmoothing(API::Trajectory &path)
{
    _pathLengthCost->setMaxLength(path);
    double len=_pathLengthCost->getTotalLength(path);
    M3D_TRACE("path length cost after norm = "<<_pathLengthCost->getCost(len)<< " - len = "<<len);
    costManager()->prepareForSmoothing(path);
}

void CostFunctionCommon::endSmoothing()
{
    API::Trajectory path=getRobot()->getCurrentTraj();
    endSmoothing(path);
}

void CostFunctionCommon::endSmoothing(API::Trajectory &path)
{
    double len=_pathLengthCost->getTotalLength(path);
    M3D_TRACE("path length cost after norm end smoothing = "<<_pathLengthCost->getCost(len)<< " - len = "<<len);
    costManager()->endSmoothing(&path);
}

void CostFunctionCommon::setHumans(const vector<Robot *> &humans)
{
    _humans = humans;
    updateHumanPos();
    updateHumanGaze();
    updateHumanBodies();
    for(unsigned int i=0;i<_humans.size();i++){
        populateJointWeights(_humans[i]);
    }
}


vector<Robot*> const & CostFunctionCommon::getHumans() const{
    return _humans;
}
vector<Eigen::Vector2d> const & CostFunctionCommon::getHumanPos() const{
    return _humanPos;
}

Robot * const &CostFunctionCommon::getRobot() const
{
    return _rob;
}

void CostFunctionCommon::setRobot(Robot *rob)
{
    _rob = rob;
    updateRobotBodies();
    populateJointWeights(rob);
}

const vector<Joint *> &CostFunctionCommon::getCostJoints() const
{
    assert(false && "deprecated function: use jointWeights() instead");
    return _costJoints;
}

void CostFunctionCommon::setCostJoints(const vector<Joint *> &cost_joints)
{
    _costJoints = cost_joints;
}

void CostFunctionCommon::updateRobot(){
    setRobot(global_Project->getActiveScene()->getActiveRobot());
}
void CostFunctionCommon::updateRobotBodies(){
    _robotBodies.clear();
    addBodyWeights(_rob);
}

void CostFunctionCommon::updateHumans(){
    _humans.clear();
    for(unsigned int i=0;i<global_Project->getActiveScene()->getNumberOfRobots();++i){
        Robot* r=global_Project->getActiveScene()->getRobot(i);
        if(r->getHriAgent() && r->getHriAgent()->is_human){
            _humans.push_back(r);
        }
    }
    updateHumanPos();
    updateHumanGaze();
    updateHumanBodies();
}

void CostFunctionCommon::updateHumanPos(){
    _humanPos.clear();
    for(vector<Robot*>::const_iterator it=_humans.begin(); it!=_humans.end(); ++it){
        _humanPos.push_back(m3dGeometry::getConfBase2DPos(*(*it)->getCurrentPos()));
    }
}

void CostFunctionCommon::updateHumanGaze(){
    _humanGaze.clear();
    for(vector<Robot*>::const_iterator it=_humans.begin(); it!=_humans.end(); ++it){
        Robot *r=*it;
        _humanGaze.push_back(r->getJoint(r->getHriAgent()->perspective->camjoint->num));
        cout << _humanGaze.back()->getName()<<endl;
    }

}

void CostFunctionCommon::updateHumanBodies(){
    _humanBodies.clear();
    _humanBodies.reserve(_humans.size());
    for(vector<Robot*>::const_iterator it=_humans.begin(); it!=_humans.end(); ++it){
        addBodyWeights(*it);
    }
}

void CostFunctionCommon::addBodyWeights(Robot *r){
    vector<p3d_obj*> bodies;
    for(int b=0;b<r->getRobotStruct()->no;++b){
        if(isBody(r->getRobotStruct()->o[b])){
            bodies.push_back(r->getRobotStruct()->o[b]);
        }
    }
    cout << "found "<<bodies.size() << " valid bodies for robot "<<r->getName()<<endl;
    map<HRI_AGENT_TYPE, map<CostSettingType, std::map<string,double> > >::iterator arr = _bodyWeightsByName.find(r->getHriAgent()->type);
    bool inBodyWeights = ( arr != _bodyWeightsByName.end() );
    if(!inBodyWeights){
        arr= _bodyWeightsByName.insert(make_pair(r->getHriAgent()->type,map<CostSettingType,map<string,double> >())).first;
        for(unsigned int i=0;i<bodies.size();i++){
            string name=bodies[i]->name;
            cout << "adding body: "<<name<<endl;
            name=name.substr(name.find('.')+1);//cut the 1st part of the name, corresponding to the robot
            arr->second[CostSettingType("default")].insert(make_pair(name,0));
        }
    }
    if(r->getHriAgent()->is_human){
        _humanBodies.push_back(bodies);
    }else{
        _robotBodies.swap(bodies);
    }

}

void CostFunctionCommon::cleanBodiesAll()
{
    cleanBodyWeightsByName();
    cleanBodiesHumanAll();
    cleanBodiesRobot();
}

void CostFunctionCommon::cleanBodyWeightsByName()
{
    for (map<HRI_AGENT_TYPE,std::map<CostSettingType, map<string,double> > >::iterator itcost=_bodyWeightsByName.begin() ;
         itcost != _bodyWeightsByName.end() ; ++itcost)
    {
         HRI_AGENT_TYPE type = itcost->first;
         Robot* sample_rob = getRobotHavingType(type);
         if(sample_rob){
             for(map<CostSettingType, map<string,double> >::iterator itbodies=itcost->second.begin();
                 itbodies != itcost->second.end(); ++itbodies)
             {
                 for (int b=0;b<sample_rob->getRobotStruct()->no;b++){
                     map<string,double> &bodies = itbodies->second;
                     p3d_obj *o=sample_rob->getRobotStruct()->o[b];
                     if(!isBody(o)){
                         string name(o->name);
                         map<string,double>::iterator in_map=bodies.find(name.substr(name.find_first_of('.')+1));
                         if(in_map!=bodies.end()){
                             bodies.erase(in_map);
                         }
                     }
                 }
             }
         }
    }
}

void CostFunctionCommon::cleanBodiesHuman(unsigned int i)
{
    if(i<_humanBodies.size())
        cleanBodies(_humanBodies[i]);
}

void CostFunctionCommon::cleanBodiesHumanAll()
{
    for (unsigned int i=0;i<_humanBodies.size();i++){
        cleanBodiesHuman(i);
    }
}

void CostFunctionCommon::cleanBodiesRobot()
{
    cleanBodies(_robotBodies);
}

void CostFunctionCommon::cleanBodies(std::vector<p3d_obj *> &bodies)
{
    remove_if(bodies.begin(),bodies.end(),p3d_col_object_is_pure_graphic);
}

bool CostFunctionCommon::isBody(p3d_obj *body)
{
    return body->pqpModel != NULL && p3d_col_object_is_pure_graphic(body) == FALSE;
}

Robot *CostFunctionCommon::getRobotHavingType(HRI_AGENT_TYPE type)
{
    Robot* sample_rob(0);
    for (unsigned int r=0;r<global_Project->getActiveScene()->getNumberOfRobots();++r){
        Robot *rob = global_Project->getActiveScene()->getRobot(r) ;
        if(rob->getHriAgent() && rob->getHriAgent()->type == type){
            sample_rob = rob;
            break;
        }
    }
    return sample_rob;
}

vector<vector<p3d_obj *> > CostFunctionCommon::humanBodies() const
{
    return _humanBodies;
}

void CostFunctionCommon::setHumanBodies(const vector<vector<p3d_obj *> > &humanBodies)
{
    _humanBodies = humanBodies;
}

const std::map<string, double> &CostFunctionCommon::bodyWeightsByNameCurrentCost(HRI_AGENT_TYPE type) const
{
    return bodyWeightsByName(type, costManager()->currentCostType());
}

const std::map<string, double> &CostFunctionCommon::bodyWeightsByName(HRI_AGENT_TYPE type, CostSettingType cost) const
{
    std::map<HRI_AGENT_TYPE,std::map<CostSettingType,std::map<std::string,double> > >::const_iterator it=_bodyWeightsByName.find(type);
    if(it!=_bodyWeightsByName.end()){
        std::map<CostSettingType,std::map<std::string,double> >::const_iterator it2=it->second.find(cost);
        if(it2!=it->second.end()){
            return it2->second;
        }else{
            throw (std::runtime_error("no bodies set for agent type "+Sem::AgentType(type).toString()+" and cost type "+cost.toString()));
        }
    }else{
        throw (std::runtime_error("no bodies set for agent type "+Sem::AgentType(type).toString()));
    }
}

const std::map<HRI_AGENT_TYPE, std::map<CostSettingType, std::map<string, double> > > &CostFunctionCommon::bodyWeightsByName() const
{
    return _bodyWeightsByName;
}

void CostFunctionCommon::addBodyWeightsByName(HRI_AGENT_TYPE agent_type, const std::map<CostSettingType,std::map<string, double> > &bodyWeights)
{
    _bodyWeightsByName[agent_type]=bodyWeights;
}

void CostFunctionCommon::setBodyWeightsByName(const map<HRI_AGENT_TYPE, std::map<CostSettingType, map<string, double> > > &bodyWeights)
{
    _bodyWeightsByName = bodyWeights;
}
const vector<p3d_obj *> &CostFunctionCommon::robotBodies() const
{
    return _robotBodies;
}

void CostFunctionCommon::setRobotBodies(const std::vector<p3d_obj *> &robotBodies)
{
    _robotBodies = robotBodies;
}

const std::vector<std::pair<int, double> > &CostFunctionCommon::bodyWeightsCurrentCost(HRI_AGENT_TYPE type) const
{
    return bodyWeights(type, costManager()->currentCostType());
}

const std::vector<std::pair<int, double> > &CostFunctionCommon::bodyWeights(HRI_AGENT_TYPE type, CostSettingType cost) const
{
    std::map<HRI_AGENT_TYPE,std::map<CostSettingType,std::vector<std::pair<int,double> > > >::const_iterator it=_bodyWeights.find(type);
    if(it!=_bodyWeights.end()){
        std::map<CostSettingType,std::vector<std::pair<int,double> > >::const_iterator it2=it->second.find(cost);
        if(it2!=it->second.end()){
            return it2->second;
        }else{
            throw (std::runtime_error("no bodies set for agent type "+Sem::AgentType(type).toString()+" and cost type "+cost.toString()));
        }
    }else{
        throw (std::runtime_error("no bodies set for agent type "+Sem::AgentType(type).toString()));
    }
}
const std::map<HRI_AGENT_TYPE, std::map<CostSettingType, std::vector<std::pair<int, double> > > > &CostFunctionCommon::bodyWeights() const
{
    return _bodyWeights;
}

void CostFunctionCommon::setBodyWeights(const std::map<HRI_AGENT_TYPE, std::map<CostSettingType, std::vector<std::pair<int, double> > > > &bodyWeights)
{
    _bodyWeights = bodyWeights;
}
CostManager *CostFunctionCommon::costManager() const
{
    return _costManager;
}

void CostFunctionCommon::setCostManager(CostManager *costManager)
{
    _costManager = costManager;
}

std::map<int, double> CostFunctionCommon::jointWeightsMovingCurrentCost(HRI_AGENT_TYPE agent_type) const
{
    const std::map<int,double> &allJoints = jointWeightsCurrentCost(agent_type);
    std::map<int,double> selectedJoints;
    Robot *r=getRobotHavingType(agent_type);
    for (std::map<int,double>::const_iterator it=allJoints.begin();it!=allJoints.end();++it){
        if(isJointMoveable(r->getJoint(it->first))){
            selectedJoints.insert(*it);
        }
    }
    return selectedJoints;
}

const std::map<int, double> &CostFunctionCommon::jointWeightsCurrentCost(HRI_AGENT_TYPE agent_type) const
{
    return jointWeights(agent_type,costManager()->currentCostType());
}

const std::map<int,double> &CostFunctionCommon::jointWeights(HRI_AGENT_TYPE agent_type, CostSettingType cost_type) const
{
    std::map<HRI_AGENT_TYPE,std::map<CostSettingType,std::map<int,double> > >::const_iterator it=_jointWeights.find(agent_type);
    if(it!=_jointWeights.end()){
        std::map<CostSettingType,std::map<int,double> >::const_iterator it2=it->second.find(cost_type);
        if(it2!=it->second.end()){
            return it2->second;
        }else{
            throw (std::runtime_error("no joints set for agent type "+Sem::AgentType(agent_type).toString()+" and cost type "+cost_type.toString()));
        }
    }else{
        throw (std::runtime_error("no joints set for agent type "+Sem::AgentType(agent_type).toString()));
    }

}
std::map<HRI_AGENT_TYPE, std::map<CostSettingType, std::map<int,double> > > CostFunctionCommon::jointWeights() const
{
    return _jointWeights;
}

void CostFunctionCommon::addJointWeights(HRI_AGENT_TYPE agent_type, const std::map<CostSettingType, std::map<int, double> > &jointWeights)
{
    _jointWeights[agent_type]=jointWeights;
}

void CostFunctionCommon::setJointWeights(const std::map<HRI_AGENT_TYPE, std::map<CostSettingType, std::map<int,double> > > &jointWeights)
{
    _jointWeights = jointWeights;
}

bool CostFunctionCommon::isJointMoveable(Joint *j)
{
    if(j==0){
        return false;
    }
    return isJointActive(j) || isJointMoveable(j->getPreviousJoint());
}

bool CostFunctionCommon::isJointActive(Joint *j){
    for(unsigned int i=0;i<j->getNumberOfDof();i++){
        if ( j->isJointDofUser(i)
             && p3d_jnt_get_dof_is_active_for_planner(j->getJointStruct(),i)
             )
        {
            return true;
        }
    }
    return false;
}

void CostFunctionCommon::populateJointWeights(Robot *r)
{
    HRI_AGENT_TYPE t=r->getHriAgent()->type;
    if(_jointWeights.count(t) == 0){
        if(_jointWeights[t].count(CostSettingType::getDefault()) == 0 ){
            for(unsigned int i=0;i<r->getNumberOfJoints();i++){
                _jointWeights[r->getHriAgent()->type][CostSettingType::getDefault()][i]=1;
            }
        }
    }
}
PathLengthCost *CostFunctionCommon::pathLengthCost() const
{
    return _pathLengthCost;
}



double HRICS_CF_computeCost(CostFunctionCommon *cfc, string name, Configuration &q)
{
    return cfc->getCost(name,q);
}


} // namespace CF
} // namespace HRICS

