#include "HRICS_CostManager.h"
#include "planner/cost_space.hpp"
#include "API/planningAPI.hpp"
#include <fstream>

using namespace std;
namespace HRICS {

INIT_MOVE3D_STATIC_LOGGER(CostManager, "planners.hrics.costmanager");

CostManager * CostManager::__mainCostManger = 0 ;

CostManager::CostManager():
    _hasControlOnPlanning(false),
    _lockFunctionsCostStack(false),
    _lockFunctionsCostStackPath(false),
    _pathCostNormalizationFactor(1)
{
    setProactive(false);
    _currentCostType= CostSettingType();
}

CostManager::CostManager(std::string const & global_path,bool proactive):
    _hasControlOnPlanning(false),
    _lockFunctionsCostStack(false),
    _lockFunctionsCostStackPath(false),
    _pathCostNormalizationFactor(1)
{
    setGlobalPath(global_path);
    setProactive(proactive);
    _currentCostType= CostSettingType();
}

CostManager::~CostManager()
{
    clear();
}

CostManager *CostManager::mainCostManager()
{
    if(__mainCostManger == 0){
        __mainCostManger  = new CostManager();
    }
    return __mainCostManger;
}

void CostManager::takeControlAndSet()
{
    setHasControlOnPlanning(true);
    M3D_DEBUG("taking control over costSpace!!! Mouhahahahahrr!");
    setCostType(currentCostType());
    global_costSpace->setCost("costManager");
}

void CostManager::setActionType(Sem::ActionType act_t)
{
    M3D_TRACE("set action type "<<act_t.toString());
    setCostType(_actionCostMapping[act_t]);
}

void CostManager::setCostType(CostSettingType cost_t){
    _currentCostType = cost_t;
    if(hasControlOnPlanning()){
        updatePlannerSettings();
    }
}

void CostManager::updatePlannerSettings()
{
    ENV.setBool(Env::isManhattan        ,false);
    ENV.setBool(Env::isMultiRRT         ,false);
    ENV.setBool(Env::costThresholdRRT   ,false);
    ENV.setBool(Env::costStarRRT        ,false);
#ifdef HRI_COSTSPACE
    ENV.setBool(Env::HRIPlannerWS       ,false);
    ENV.setBool(Env::HRIPlannerTRRT     ,false);
    ENV.setBool(Env::HRIPlannerCS       ,false);
#endif
    ENV.setBool(Env::useTRRT            ,false);
    ENV.setBool(Env::useTERRT           ,false);
    ENV.setBool(Env::isCostSpace        ,false);

    ENV.setBool(planningMethod(),true);
    if(planningMethod() == Env::useTERRT || planningMethod() == Env::useTRRT ){
        ENV.setBool(Env::isCostSpace        ,true);
    }


    //TODO set these values
    ENV.setExpansionMethod(Env::Extend);
    //OR
    //ENV.setExpansionMethod(Env::Connect);

    if (! _algoParams.count(Sem::AgentType())){
        //save current params as default
        _algoParams[Sem::AgentType()] = AlgoParams(ENV.getDouble(Env::extensionStep),ENV.getDouble(Env::minimalFinalExpansionGap));
    }

    Sem::AgentType agnt = Sem::AgentType::fromRobot(global_Project->getActiveScene()->getActiveRobot());
    if(agnt.isDefined() && _algoParams.count(agnt)){
        updatePlannerSettings(_algoParams[agnt]);
        M3D_DEBUG("update RRT settings to those defined for agentType "<<agnt.toString());
    }else{
        //use default
        updatePlannerSettings(_algoParams[Sem::AgentType()]);
        M3D_INFO("no algo settings found for current robot, falling back to defaults");
    }

}

void CostManager::updatePlannerSettings(CostManager::AlgoParams &params)
{
    ENV.setDouble(Env::extensionStep,params.extStep);
    ENV.setDouble(Env::minimalFinalExpansionGap,params.finalGap);
}

double CostManager::getCost(Configuration &q)
{
    if(!_currentCostType.isDefined()){
        M3D_WARN("no cost type defined");
        return 0;
    }
    std::map<CostSettingType,boost::function<double(Configuration&)> >::const_iterator selectedFunction = _functions.find(_currentCostType);
    assert(_lockFunctionsCostStack==false);
    _lockFunctionsCostStack=true;
    _functionCostStack.clear();
    if(selectedFunction==_functions.end()){
        M3D_WARN("no function set or valid");
        return 0;
    }
    double cost;
    try{
        cost = selectedFunction->second(q);
    }catch(std::runtime_error e){
        M3D_ERROR("caught exception: "<<e.what());
        cost=0;
    }

    _functionCostStack.clear();
    _lockFunctionsCostStack=false;

    return cost;
}

double CostManager::getCostPath(std::vector<LocalPath*> &path)
{
    double cost(0);
    if(!_currentCostType.isDefined()){
        M3D_WARN("no cost type defined");
        return 0;
    }
    boost::function<double (std::vector<LocalPath*>&)> func;
    func = getCostPathFunction(_currentCostType);
    if(func.empty()){
        func=getCostPathFunction(CostSettingType::getDefault());
        if(func){
            //M3D_TRACE("will use default traj cost function");
        }
    }
    if(func.empty()){
        M3D_WARN("no path cost function set or valid");
        return 0;
    }
    assert(_lockFunctionsCostStack==false);
    _lockFunctionsCostStackPath=true;
    _functionCostStackPath.clear();
    try{
        cost= func(path);
    }catch (std::runtime_error e){
        M3D_ERROR(e.what());
        cost=0;
    }
    _functionCostStackPath.clear();
    _lockFunctionsCostStackPath=false;

    return cost;
}

double CostManager::getCostAggregationNormalized(std::vector<LocalPath *> &path)
{
    return global_costSpace->costAggregationOverTraj(path) * pathCostNormalizationFactor();
}

boost::function<double (std::vector<LocalPath*>&)>
CostManager::getCostPathFunction(CostSettingType type)
{
    boost::function<double(std::vector<LocalPath*>&)> f_e=0;
    assert (!f_e);
    std::map<CostSettingType,
            boost::function<double(std::vector<LocalPath*>&)> >::const_iterator
            selectedFunction;

    if(isSmoothing()){
        selectedFunction = _functionsTrajSmoothing.find(type);
        if(selectedFunction==_functionsTrajSmoothing.end()){
            return f_e;
        }
    }else{
        selectedFunction = _functionsTraj.find(type);
        if(selectedFunction==_functionsTraj.end()){
            return f_e;
        }
    }

    return selectedFunction->second;

}

void CostManager::clear()
{
    _actionCostMapping.clear();
    _functions.clear();

    _functionsTraj.clear();
    _functionsTrajSmoothing.clear();
}

void CostManager::prepareForSmoothing(API::Trajectory &path)
{
    _isSmoothing = true;
    //TODO dummy constant factor to be set elsewhere (JSON)
    double c = global_costSpace->costAggregationOverTraj(path.getCourbe());
    M3D_TRACE("cost aggregation = "<<c);
    if(c!=0){
        _pathCostNormalizationFactor = 1/c;
    }else{
        _pathCostNormalizationFactor = 1;
    }
    M3D_TRACE("..\tnormalized = " << this->getCostAggregationNormalized(path.getCourbe()));
    M3D_TRACE("..\tcostManager cost = "<<this->getCostPath(path.getCourbe()));
}

void CostManager::endSmoothing(API::Trajectory *path)
{
    if(path){
        double c = global_costSpace->costAggregationOverTraj(path->getCourbe());
        M3D_TRACE("cost aggregation end smoothing = "<<c);
        M3D_TRACE("..\tnormalized = " << this->getCostAggregationNormalized(path->getCourbe()));
        M3D_TRACE("..\tcostManager cost = "<<this->getCostPath(path->getCourbe()));
    }
    _isSmoothing = false;
    _pathCostNormalizationFactor = 1;
}

bool CostManager::fetchSettings()
{
    bool ok(true);
    std::string path;

    clear();

    //_functionsTrajSmoothing[CostSettingType::getDefault()] =
    //        boost::bind(std::plus<double>(),
    //                    boost::bind(global_costSpace->getCostFunctionTrajectories("PathLength"),_1),
    //                    boost::bind(&CostManager::getCostAggregationNormalized,this,_1));

    //_functionsTraj[CostSettingType::getDefault()] =
    //        boost::bind(&CostManager::getCostAggregationNormalized,this,_1);

    path = getPath()  + "actions.json";
    ok = readActionsCostMapping(path);
    if(ok){
        for(std::map<Sem::ActionType,CostSettingType>::const_iterator it =_actionCostMapping.begin(); it!=_actionCostMapping.end(); ++it){
            if(!readCostSetting(it->second,CONF_COST)){
                M3D_WARN ( "fail to read cost settings for action "<<it->first.toString());
            }
            if(!readCostSetting(it->second,PATH_COST)){
                M3D_WARN ( "fail to read path cost settings for action "<<it->first.toString());
            }
            if(!readCostSetting(it->second,PATH_SMOOTHING_COST)){
                M3D_WARN ( "fail to read path smoothing cost settings for action "<<it->first.toString());
            }
        }
        for ( unsigned int i=0;i<global_Project->getActiveScene()->getNumberOfRobots();++i){
            Sem::AgentType agnt;
            try{
                agnt = Sem::AgentType::fromRobot(global_Project->getActiveScene()->getRobot(i));
            }catch(std::runtime_error e){
                M3D_WARN("caugth following error while getting AgentType of robot "<<global_Project->getActiveScene()->getRobot(i)->getName() << ":");
                M3D_WARN(e.what());
                continue;
            }
            readAlgoParams(agnt);
        }
    }
    return ok;
}

bool CostManager::readActionsCostMapping(string path)
{

    bool ok(true);
    Json::Value root;
    std::ifstream ifs;

    ifs.open(path.c_str());
    if (!ifs){
        ok = false;
        M3D_ERROR ( "could not open json file " << path);
    }
    if (ok){
        try{
            ifs >> root;
        }catch (std::runtime_error e){
            ok=false;
            M3D_ERROR ("error while reading file "<<path<< " : "<<e.what());
        }catch (std::exception e){
            ok=false;
            M3D_ERROR ("error while reading file "<<path <<" : "<< e.what());
        }
        ifs.close();
    }
    if(ok){
        for (Json::Value::iterator it = root.begin() ; it != root.end() ; it++){
            Sem::ActionType act_t(it.key().asString());
            if(_actionCostMapping.count(act_t) > 0 ){
                M3D_WARN ( "multiple definition of cost for action "<< it.key().asString() << ". Overwritting previous definition.");
            }
            _actionCostMapping[act_t] = (*it).asString();
        }
    }
    return ok;
}

bool CostManager::readCostSetting(CostSettingType const & cost,CostAppliesWhen_t when)
{
    return readCostSetting(getPath(when) + cost.toString() + ".json", cost,when);
}

bool CostManager::readCostSetting(std::string path, const CostSettingType &cost,CostAppliesWhen_t when)
{
    bool ok(true);
    Json::Value func_json;
    std::ifstream ifs;
    ifs.open(path.c_str());
    M3D_TRACE("reading "<<path);
    if (!ifs){
        ok = false;
        M3D_ERROR ( "could not open json file " << path);
    }
    if (ok){
        try{
            ifs >> func_json;
        }catch (std::runtime_error e){
            ok=false;
            M3D_ERROR ("error while reading file "<<path<< " : "<<e.what());
        }catch (std::exception e){
            ok=false;
            M3D_ERROR ("error while reading file "<<path <<" : "<< e.what());
        }
        ifs.close();
    }
    if(ok){
        try{
            switch(when){
            case CONF_COST:
                _functions[cost] = jsonToFunction(func_json);
                break;
            case PATH_COST:
                _functionsTraj[cost] = jsonToFunctionTraj(func_json);
                break;
            case PATH_SMOOTHING_COST:
                _functionsTrajSmoothing[cost] = jsonToFunctionTraj(func_json);
                break;
            }
        }catch (std::runtime_error e){
            ok=false;
            M3D_ERROR("error while parsing "<<path<< ":");
            M3D_ERROR(e.what());
        }
    }
    return ok;
}

bool CostManager::readAlgoParams(Sem::AgentType const &agent_type)
{
    if(agent_type.isDefined()){
        return readAlgoParams(agent_type, getPath() + "AlgoParam/" + agent_type.toString() + ".json");
    }else{
        return false;
    }
}

bool CostManager::readAlgoParams(Sem::AgentType const &agent_type, const string &path)
{
    bool ok(true);
    Json::Value param_json;
    std::ifstream ifs;
    ifs.open(path.c_str());
    M3D_TRACE("reading "<<path);
    if (!ifs){
        ok = false;
        M3D_ERROR ( "could not open json file " << path);
    }
    if (ok){
        try{
            ifs >> param_json;
        }catch (std::runtime_error e){
            ok=false;
            M3D_ERROR ("error while reading file "<<path<< " : "<<e.what());
        }catch (std::exception e){
            ok=false;
            M3D_ERROR ("error while reading file "<<path <<" : "<< e.what());
        }
        ifs.close();
    }
    if(ok){
        try{
            if (param_json.isMember("extension_step")){
                _algoParams[agent_type].extStep = param_json["extension_step"].asDouble();
            }
            if (param_json.isMember("final_gap")){
                _algoParams[agent_type].finalGap = param_json["final_gap"].asDouble();
            }
        }catch (std::runtime_error e){
            ok=false;
            M3D_ERROR("error while parsing "<<path<< ":");
            M3D_ERROR(e.what());
        }
    }
    return ok;


}

boost::function<double (Configuration &)> CostManager::jsonToFunction(Json::Value &f)
{
    if (f.isMember("op")){
        return boost::bind( elementaryOperation(f["op"].asString()),
                            boost::bind(jsonToFunction(f["left"]),_1),
                boost::bind(jsonToFunction(f["right"]),_1) );
    }else if (f.isMember("func")){
        return boost::bind( elementaryFunction(f["func"].asString(),f),
                boost::bind(jsonToFunction(f["x"]),_1) );
    }else if (f.isMember("cost")){
        return boost::bind ( &CostManager::getCostStack,this,f["cost"].asString(),_1 );
    }else if (f.isMember("const")){
        return boost::bind (std::plus<double>(), 0, f["const"].asDouble());
    }else{
        throw (std::runtime_error("error reading cost specification"));
    }
    return boost::function<double(Configuration &)>();
}

boost::function<double (std::vector<LocalPath *> &)> CostManager::jsonToFunctionTraj(Json::Value &f)
{
    if (f.isMember("op")){
        return boost::bind( elementaryOperation(f["op"].asString()),
                            boost::bind(jsonToFunctionTraj(f["left"]),_1),
                boost::bind(jsonToFunctionTraj(f["right"]),_1) );
    }else if (f.isMember("func")){
        return boost::bind( elementaryFunction(f["func"].asString(),f),
                boost::bind(jsonToFunctionTraj(f["x"]),_1) );
    }else if (f.isMember("cost")){
        return boost::bind ( &CostManager::getCostStackPath,this,f["cost"].asString(),_1 );
    }else if (f.isMember("const")){
        return boost::bind (std::plus<double>(), 0, f["const"].asDouble());
    }else{
        M3D_ERROR("can't read trajectory cost specification :\n\""<<f<<"\"");
        throw (std::runtime_error("error reading trajectory cost specification"));
    }
    return boost::function<double(std::vector<LocalPath*> &)>();

}

boost::function<double (double, double)> CostManager::elementaryOperation(const string &name)
{
    if(name == "+"){
        return boost::bind( std::plus<double>(),_1,_2);
    }else if(name == "-"){
        return boost::bind( std::minus<double>(),_1,_2);
    }else if(name == "*"){
        return boost::bind( std::multiplies<double>(),_1,_2);
    }else if(name == "/"){
        return boost::bind( std::divides<double>(),_1,_2);
    }else if (name=="^"){
        return boost::bind(static_cast<double (*)(double,double)>(&std::pow),_1,_2);
    }else if(name == "min"){
        return boost::bind( std::min<double>,_1,_2);
    }else if(name == "max"){
        return boost::bind( std::max<double>,_1,_2);
    }else{
        throw(std::runtime_error("error reading cost specification, unknown operation "+name));
    }

}


boost::function<double (double)> CostManager::elementaryFunction(const string &name,Json::Value const &json)
{
    if(name == "exp"){
        return boost::bind( static_cast<double (*)(double)>(&exp),_1);
    }else if(name == "cos"){
        return boost::bind( static_cast<double (*)(double)>(&cos),_1);
    }else if(name == "sin"){
        return boost::bind( static_cast<double (*)(double)>(&sin),_1);
    }else if(name == "sigmoid"){
        return boost::bind( CostManager::sigmoidFunction,_1,json["up"].asDouble(),json["low"].asDouble(),json["min"].asDouble(),json["max"].asDouble());
    }else if(name == "bound"){
        return boost::bind( CostManager::boundFunction,_1,json["min"].asDouble(),json["max"].asDouble());
    }else if(name=="affine"){
        return boost::bind( CostManager::affineFunction,_1,json["a"].asDouble(),json["b"].asDouble());
    }else{
        throw(std::runtime_error("error reading cost specification, unknown function "+name));
    }

}

string CostManager::getPath()
{
    return globalPath() + "/HRICS/CostManager/";

}

string CostManager::getPath(CostManager::CostAppliesWhen_t when)
{
    std::string s;
    switch (when){
    case CONF_COST:
        s="Costs/";
        break;
    case PATH_COST:
        s="PathCosts/";
        break;
    case PATH_SMOOTHING_COST:
        s="PathSmoothingCosts/";
        break;
    }
    return getPath()+s;
}
bool CostManager::proactive() const
{
    return _proactive;
}

void CostManager::setProactive(bool proactive)
{
    _proactive = proactive;
    if (_proactive){
        fetchSettings();
    }
}
std::string CostManager::globalPath() const
{
    return _globalPath;
}

void CostManager::setGlobalPath(const std::string &globalPath)
{
    _globalPath = globalPath;
}
CostSettingType CostManager::currentCostType() const
{
    return _currentCostType;
}

bool CostManager::isSmoothing() const
{
    return _isSmoothing;
}

double CostManager::pathCostNormalizationFactor() const
{
    return _pathCostNormalizationFactor;
}

void CostManager::setPathCostNormalizationFactor(double pathCostNormalizationFactor)
{
    _pathCostNormalizationFactor = pathCostNormalizationFactor;
}

double CostManager::sigmoidFunction(double val, double upper, double lower, double min, double max)
{
    if(std::isnan(val)){return val;}
    assert(upper> lower);
    assert(max>=min);
    double val_norm=(val-lower)/(upper- lower);
    double c=( max-min)*( 1 / (1+ std::exp(-val_norm*12+6))) + min;
    assert(c<=max);
    assert(c>=min);
    return c;
}

double CostManager::boundFunction(double val, double min, double max)
{
    return std::max(min,std::min(max,val));
}

double CostManager::affineFunction(double val, double a, double b)
{
    return val * a + b;
}

Env::boolParameter CostManager::planningMethod()
{
    //TODO is a parameter in json, depend on planning type
    return Env::useTRRT;
}

double CostManager::getCostStack(const string &name, Configuration &q)
{
    std::pair<std::map<std::string,double>::iterator , bool > inserted = _functionCostStack.insert(std::make_pair(name,-1.));
    double c(0);
    if(inserted.second){
        //new element inserted => compute cost
        boost::function<double (Configuration &)>  f = global_costSpace->getCostFunction(name);
        if(!f){
            M3D_ERROR("cost function \""<<name<<"\" does not exist in cost space!");
            //throw std::runtime_error("cost function \""+name+"\" does not exist in cost space");
            // f(q) will throw empty function exception
        }
        c = f(q);
        inserted.first->second = c;
    }else{
        //alredy in stack
        c = inserted.first->second;
    }
    return c;
}

double CostManager::getCostStackPath(const string &name, std::vector<LocalPath *> &path)
{
    std::pair<std::map<std::string,double>::iterator , bool > inserted = _functionCostStackPath.insert(std::make_pair(name,-1.));
    double c(-1);
    if(inserted.second){
        //new element inserted => compute cost
        boost::function<double (std::vector<LocalPath *> &)>  f = global_costSpace->getCostFunctionTrajectories(name);
        if(!f){
            M3D_ERROR("path cost function \""<<name<<"\" does not exist in cost space!");
            //throw std::runtime_error("path cost function \""+name+"\" does not exist in cost space");
            // f(path) will throw empty function exception
        }
        c = f(path);
        inserted.first->second = c;
    }else{
        //alredy in stack
        c = inserted.first->second;
    }
    assert(c!=-1);
    return c;

}

bool CostManager::hasControlOnPlanning() const
{
    return _hasControlOnPlanning;
}

void CostManager::setHasControlOnPlanning(bool hasControlOnPlanning)
{
    _hasControlOnPlanning = hasControlOnPlanning;
}







} // namespace HRICS

