#ifndef HRICS_HRICS_MULTIHANDOVER_H
#define HRICS_HRICS_MULTIHANDOVER_H

#include "HRICS_Workspace.hpp"
#include <vector>
#include "planner/planner.hpp"
#include "utils/multiHandOver/SimpleGraph.h"
#include "Grid/HRICS_MultiAgentGrid.hpp"
//#include "utils/MultiHandOverUtils.hpp"
#include "utils/MultiHandOverAlgos.h"
#include "utils/MultiHandOverDisplay.h"
#include "Logging/Logger.h"

namespace HRICS {

class MultiHandOver :public  HumanAwareMotionPlanner
{
    MOVE3D_STATIC_LOGGER;
public:

    //typedef HRICS::MultiHandOverSolution MultiHandOverSolution;
    typedef enum {NONE,RANDOM_SEARCH,LWA_SEARCH,WA_SEARCH} SearchAlgo_t;

    MultiHandOver(double cell_size=0.20);
    virtual ~MultiHandOver();

    void testForReinit();
    void createGrid();
    void dumpVar();
    void setDefaultEnvVar(void);
    bool initFromConfXML(std::istream &is);
    bool initFromConfXML(std::string path);

    /// @brief Initialize the tool set for algorithm acording to the settings
    /// @see MHO::ToolSet
    /// @see #setUseRrtInFrontierCheck(bool)
    void initToolSet();
    MHO::ToolSet* getToolSet(){return this->_tool_set;}
    /// @brief returns a pointer to newly created Heuristic class for main A* search.
    /// based on getAstarHeuristic()
    MHO::MultiHandOverHeuristic* createAstarHeuristic(void);
    void setDefaultHandOverTesters(std::vector<std::string> &some_defaults){this->_handover_testers_defaults_set=some_defaults;}
    /// reads the environment variable $MOVE3D_SCRIPT to get a script/conf file to execute
    unsigned int runScriptFromEnvVar();
    /// reads the given script and execute it
    unsigned int runScript(std::string filename);
    unsigned int run();

    /** @name initialisation
     * This groups contains initialization methods
     */
    ///@{
    unsigned int findAgents();
    void fetchAgentsConfsInitGoTo();
    unsigned int findManipulableObjects();
    unsigned int computeAgentsGraph(void);
    void initGrid(void);
    void computeGrid(void);
    void computeFrontiers(void){_grid->searchAllFrontiers();_progression_state=FRONTIERS_DISTANCE;}
    ///@}
    /// \brief reset the agents to their initial configuration
    void resetAgentsConf(void);
    void resetManipulablesConf(void);

    Configuration getAgentTargetConf(unsigned int a);
    /// returns 0 if have_final_position is set to false for that agent, a ptr to the cell otherwise.
    MultiAgentCell* getAgentTargetCell(unsigned int a);

    Configuration getAgentInitConf(unsigned int a);
    MultiAgentCell* getAgentInitCell(unsigned int a);

    void setToComputeHOTraj(bool activate,int solution=-1,int step=0);
    void computeLongestSolutionDistance();
    /**
     * @brief compute possible handover successions
     * @param cost_majoration keep a solution only if its cost is lower than the shortest(dijkstra) solution cost*cost_majoration
     * @return the number of solutions found
     *
     * This is based on the graph built by computeAgentsGraph().
     * It enumerates all possible paths without circuits from the first to the target agent,
     * and limits solutions to the ones that have a cost lower than the given factor times the
     * shortest path cost.
     * @note that here the cost are very imprecise, and a solution found here might
     * not even be feasible.
     */
    unsigned int computeRawSolutions(float cost_majoration);
    void clearSolutions(void);
    void insert_solution_sorted(MHO::MultiHandOverSolution* sol);

    /**
     * @brief searchSolutionLWAstar
     *
     * Lazy Weighted A* is a search-based algorithm.
     * We search here in the 2D grid for a path through 2d cells and
     * frontiers.
     * reference:
     * Planning Single-arm Manipulations with N-Arm Robots, Cohen B. et al.
     * RSS 2014
     * http://www.roboticsproceedings.org/rss10/p33.html
     */
    void searchSolutionLWAstar(bool no_lazy);

    void postProcessSolution(MHO::MultiHandOverSolution *sol);
    /**
     * @brief The core function of this class, it search for an optimal solution.
     *
     * This function search for new solutions, improve them, change them in order
     * to find an optimal solution.
     * Its comportement is driven by several parameters, besides the problem definition.
     * These are time and iteration limit, overheat temperature (when nothing is improved
     * for a long time, we call this overheat), and other details in algorithm selection.
     */
    void searchSolutions(void);
    bool handOverPossible(unsigned int a1, unsigned int a2);
    /**
     * @brief handOverFindConfs
     * @param solution the solution for which we will compute the transfert configurations
     * @param frontier_index is used to specify the frontier/action in the solution
     *  for which we want to compute , default is -1, for all.
     * @return false if it failed at finding a possible configuration
     */
    bool handOverFindConfs(MHO::MultiHandOverSolution * solution, int frontier_index=-1);
    bool handOverFindConfs(FrontierPtr frontier);
    MHO::MultiHandOverSolution* changeHandOverSolution( MHO::MultiHandOverSolution * solution);
    /**
     * @brief computeNavigationTrajectories
     * @param solution
     * @return
     *
     * take agent use costs into account for cost computation (adds costs to nav_dists and cost in solution)
     */
    double computeNavigationTrajectories(MHO::MultiHandOverSolution * solution, int action_index=-1);
    void computeHandOverTrajectory(MHO::MultiHandOverSolution * solution, int step=-1);
    void computeSubTaskTrajectory(MHO::SubTask* sub_task);
    MHO::MultiHandOverSolution* optimizeSolution(MHO::MultiHandOverSolution *solution);
    /**
     * @brief optimize a frontier in a solution to minimize the cost.
     * @param solution
     * @param action_index
     * @param keep_distance
     * @param min_distance
     * @return
     *
     * The optimization process can have differents comportements,
     * which are selected through class members.
     * @todo optimization comportement selection
     *
     *  -# The first one is the slowest in most cases. It's an optimization
     * process where we search for a neighbor frontier, check if we can make a handover there,
     * compute the handover, then the full cost of the solution with this new frontier,
     * computing new trajectories when needed (consequence is a call to A* for each solution
     * we want to check).
     * If the new cost is better, accept this solution and do the same on it, until no better
     * is found.
     *  -# The second one is a little bit more efficient as it precompute the navigation paths,
     * with a flooding algorithm provided by the grid, avoiding lots of calls to A*.
     *  -# The third one might be very efficient as it first search for a good solution in term
     * of distances (based on flooding algorithm as previously) with no consideration to 3D
     * environment (which spare lot of computational cost). When an optimal solution is found
     * in the 2D problem, it search from that point using the previous method.
     *  -# It's also possible to find a cell in the 2d grid that minimizes the cost for any agent
     * to travel from/to that point (like a 0-distance frontier) and then find a real frontier from
     * that point, which will be used as a starting point for the algorithm of the 2nd item
     * @see optimalCellNullDistanceFrontier()
     *
     */
    MHO::MultiHandOverSolution* optimizeFrontier(MHO::MultiHandOverSolution *solution, unsigned int action_index, double keep_distance, double min_distance);

    /**
     * @brief optimalCellNullDistanceFrontier
     * @return the cell found.
     *
     * Searches from a given starting cell a cell that optimizes
     * navigation related costs from/to this cell to defined cell,
     * for given agents
     *
     * used by optimizeFrontier() in some mode
     */
    MultiAgentCell* findOptimalCellNullDistanceFrontier(MHO::MultiHandOverSolution* solution,unsigned int action_index);

    double navigationRelatedCost(double dist,unsigned int agent);
    double navigationRelatedCostTime(double dist,unsigned int agent,double &time);
    double computeSolutionCost(MHO::MultiHandOverSolution * solution);

    bool getSolutionDataForMHP(std::vector<MHO::MultiHandOverMhpHO> &hand_overs,std::vector<Eigen::Vector3d, Eigen::aligned_allocator<Eigen::Vector3d> > &path, double &cost, double &duration, std::string &main_agent_name);

    /*!
     * \brief get a randomly selected solution, most probably with low cost
     * \param sigma
     * \return
     *
     * computes a normal distribution N(0,sigma) to select the solution in the sorted list of solutions (lower first)
     */
    MHO::MultiHandOverSolution * getLowCostRandomSolution(double sigma, int *index=0);

    /** @name getter/setters
      * getters and setters for attributes of the class
      */
    ///@{

    /// available heuristic types for [L][W]A*
    typedef enum {DIST_BEST_AGENT =0, BEST_HANDOVER_SERIES} AStarHeuristic_t;
    void setAstarHeuristic(MultiHandOver::AStarHeuristic_t heuristic){_a_star_heuristic = heuristic;}
    AStarHeuristic_t getAstarHeuristic(void){return _a_star_heuristic;}
    void setSearchAlgo(MultiHandOver::SearchAlgo_t algo){_search_algo=algo;}
    MultiHandOver::SearchAlgo_t getSearchAlgo(){return _search_algo;}
    double getEpsilonForLWAstar(){return _epsilon;}
    /// set the cell size for discretisation step in the grid
    void setGridCellSize(double size){_cellSize=size;}
    double getGridCellSize(void){return _cellSize;}
    void setEpsilonForLWAstar(double e){this->_epsilon=std::max(1.,e);}
    void setCostsFactors(double dist, double time, double agent_costs, double hri){ _f_dist=dist;_f_time=time;_f_agent_cost=agent_costs;_f_hri=hri;}
    void getCostsFactors(double &dist, double &time, double &agent_costs, double &hri){ dist=_f_dist;time=_f_time;agent_costs=_f_agent_cost;hri=_f_hri;}

    void setSearchSolutionStopCondition(int max_it,int overheat,int time_limit_ms){ _max_iterations=max_it;_overheat_temperature=overheat;_time_limit=time_limit_ms;}
    void getSearchSolutionStopCondition(int &max_it,int &overheat,int &time_limit_ms){ max_it=_max_iterations;overheat=_overheat_temperature;time_limit_ms=_time_limit;}

    FrontierCells::FrontierCheckStatus getFrontierMaxCheckStatus(){return _max_frontier_check_status;}
    void setFrontierMaxCheckStatus(FrontierCells::FrontierCheckStatus fcs){_max_frontier_check_status = fcs;}

    const std::vector<Robot*>& getAgents(void){return this->_agents;}
    const std::vector<Robot*>& getManipulableObjects(void){return this->_manipulable;}

    /// @brief the MHO::MultiHandOverAgent corresponding to agent ID a
    /// @param a the ID, must be valid (not checked)
    MHO::MultiHandOverAgent *getAgentStruct(unsigned int a){return &_mho_agents[a];}

    void setFirstAgent(Robot* f){_first_agent=f;setFirstAgentNode();}
    Robot* getFirstAgent(void){return _first_agent;}
    void setTargetAgent(Robot* f){_target_agent=f;setTargetAgentNode();}
    Robot* getTargetAgent(void){return _target_agent;}
    void setTargetAgent(std::string name);
    void setFirstAgent(std::string name);
    void setTargetAgent(p3d_rob* rob);
    void setFirstAgent(p3d_rob* rob);
    void setRandomizeFirstTarget(bool b){_random_first_target_agents=b;}
    void setObject(unsigned int i){setObject(_manipulable[i]);}
    void setObject(Robot* obj){_object=obj;_object_bb_vects.clear();}
    Robot *getObject(){return _object;}
    void setObject(p3d_rob* obj);

    bool usePostProcess(){return this->_use_post_process;}
    void usePostProcess(bool b){this->_use_post_process=b;}

    void setFirstAgentNode(void);
    void setTargetAgentNode(void);

    ///@}

    /** @name Agent parameters
     * getters and setters for the agent specific parameters
      */
    ///@{
    void setAgentUseCost(unsigned int i,double cost){this->_mho_agents[i].use_cost=cost;}//this->_agents_use_costs[i]=cost;}
    double getAgentUseCost(unsigned int i){return this->_mho_agents[i].use_cost;}//this->_agents_use_costs[i];}
    double getAgentUseCost(Robot * r);
    //std::vector<double> const & getAgentUseCost(){return this->_agents_use_costs;}
    //void setAgentUseCost(std::vector<double> const & costs){this->_agents_use_costs=costs;}
    void setAgentSpeed(unsigned int i, double speed){this->_mho_agents[i].speed=speed;}//this->_agents_speeds[i]=speed;}
    double getAgentSpeed(unsigned int i){return this->_mho_agents[i].speed;}//this->_agents_speeds[i];}
    bool isHuman(unsigned int i){return this->_mho_agents[i].is_human;}
    //std::vector<double> const & getAgentsSpeeds(){return this->_agents_speeds;}
    //void setAgentsSpeeds(std::vector<double> const & speeds){this->_agents_speeds=speeds;}

    double getTimeHO(unsigned int a1,unsigned int a2);
    void setTimesHO(double rr,double hh,double rh){_r_r_ho_time=rr;_r_h_ho_time=rh;_h_h_ho_time=hh;}

    ///@}


    /// @name 3d frontier checking
    ///@{

    void setCheckFrontierBefore(bool b){this->_check_frontier_before=b;}
    bool getCheckFrontiersBefore(void){return this->_check_frontier_before;}

    /// @brief to use an RRT on the object as free flyer to check frontier or not.
    /// @param b use RRT when true.
    ///
    /// When not using RRT, uses only a direct path check between origin and target*/
    /// (faster but not conservative)
    void setUseRrtInFrontierCheck(bool b){_use_rrt_for_frontier_check=b;}
    void setUseDummySmoothInFrontierCheck(bool b){_use_dummy_smooth_for_frontier_check=b;}
    double getShoulderHeight(unsigned int agent);
    Eigen::Vector3d getObjectMainDirection(void);
    Configuration confForBBAxisAlign(Robot * object, Eigen::Vector3d const & main_dir, Eigen::Vector3d const & second_dir);

    /*!
     * \brief set the orientation of the object for a handover
     * \param dir1 the 'direction' of the handover (can be a vector from giver agent shoulder to receiver agent shoulder)
     * \todo implement different ways of choosing the direction, current is having the smallest BB surface
     * orthogonal to the direction, and the biggest surface in a vertical plane.
     * \see confForBBAxisAlign()
     */
    void setOrientationObjectForHO(Eigen::Vector3d &dir1);
    void checkAllMetaFrontiersWithObject(void);
    bool checkMetaFrontiersWithObject(unsigned int a1,unsigned int a2);
    bool checkMetaFrontierWithObject(unsigned int a1,unsigned int a2,MetaFrontierPtr mf);
    /**
     * @brief check frontiers with object, stop when one if found to be OK, or max_it checks are made
     * @param [out] frontier_found pointer to the frontierCell found to be OK, NULL if not found
     * @param a1 first agent index
     * @param a2 second agent index
     * @param max_it max number of frontiers that will be tested if no one is found to be OK
     * @return true if a frontier is found to be OK
     * @see HRICS::MultiHandOver::checkFrontierWithObject()
     */
    bool randomCheckFrontiersWithObject(std::tr1::shared_ptr<FrontierCells> & frontier_found, unsigned int a1, unsigned int a2,unsigned int max_it);
    /**
     * @brief checkFrontierWithObject
     * @param a1
     * @param a2
     * @param frontier_index
     * uses the specified object (MultiHandOver::_object) and tries to
     * find a simple path from agent 1 shoulder to agent 2 shoulder,
     * each one of them standing on the frontier
     * The path finding search for a simple linear path.
     * @todo (improve) set step size and distance threshold for object path finding
     * @todo (improve) object orientation
     * @todo improve the path search
     */
    bool checkFrontierWithObject(FrontierCells &frontier);
    ///@}

    /** @name Display
     * methods for displaying information, setting display parameters for move3d GUI...
      */
    ///@{
    void printSolutionsForest(void);
    void printMainTimes(std::vector<double> times);
    void printLWAStats();

    void setAgentIndexToDrawGrid(unsigned int index){
        if(_grid) this->_grid->setAgentIndexToDrawGrid(index);
    }

    void setDrawAgentAccessibilityGrid(bool b,bool boundariesOnly=false){if(_grid) this->_grid->setDrawAgentAccessibilityGrid(b,boundariesOnly);}
    void setDrawFrontiers(bool b,unsigned int i=0,unsigned int j=0){if(_grid) _grid->setDrawFrontiers(b,i,j);}

    void drawObjectWhenFrontierChecking(bool b){_draw_object_frontiers_checks=b;}
    void setDrawAgentsDuringSearch(bool b){_draw_agents_holding_conf_during_search=b;}
    void setDraw2dPaths(bool b){_draw_2d_path=b;}
    void setDrawAll2dPaths(bool b){_draw_all_2d_pathes=b;}
    void setDrawLWA(bool b){_draw_lwa=b;if(_display_lwa_in_grid_obj){_display_lwa_in_grid_obj->setDraw(b);}}

    MHO::MultiHandOverSolution::CostData getSolutionDataStudio(int solution);
    /**
     * @brief set the robots to their conf computed in searchSolutions() for the given solution and step
     * @param solution the index of the solution (-1 refers to the best solution)
     * @param step the frontier/action/handover (all the same) index of the solution
     * @return the cost of the solution
     * if parameters refers to inexistent frontier, does nothing
     */
    double setSolutionStepRobotsConfs(int solution, int step, MHO::MultiHandOverSolution::CostData &cost_data);
    void getConfHandOver(int solution, int step, std::string robot_name, confPtr_t &conf);
    void draw2dPath(void);
    void draw2dPath(std::vector<MultiAgentCell*> const & path, int color_id, bool alt=false);
    /// sets the option of drawing the frontier optimization process as 2D cells
    /// representing the frontier currently being checked
    /// @note if DRAW_OPTIMIZE_FRONTIER macro is not defined, does nothing
#ifdef DRAW_OPTIMIZE_FRONTIER
    void setDrawOptimizeFrontier(bool b){_draw_optimize_frontier=b;}
#else
    void setDrawOptimizeFrontier(bool b){if(_showText){std::cout<<"DRAW_OPTIMIZE_FRONTIER macro not defined";}}
#endif

    void playTaskStepByStep(MHO::Task* task);
    void playSolutionTrajSync(MHO::Task *task);
    bool  playSolutionTaskSyncStep(int solution=-1);
    void computeTaskTrajs(MHO::Task* task);
    void setSubTaskTrajToRobot(MHO::SubTask* subtask);
    void playSolutionTask(int solution);
    void setPlaySolutionTask(int solution);
    void setPlaySolutionTaskTraj(int solution);
    void setPlaySolutionSyncStepByStep(int solution);
    bool isEndPlaySolutionTaskTraj();
    ///@}

    MultiAgentGrid *getPlanGrid(void){return _grid;}



private:

    enum {INIT, FRONTIERS_DISTANCE,FRONTIERS_OBJECT_PATH,END} _progression_state;///< this is the state of the algorithm
    std::vector<MHO::MultiHandOverAgent> _mho_agents;
    std::vector<Robot*> _agents;
    std::vector<Robot*> _manipulable;
    mho::SimpleGraph _agents_frontiers_dist_graph;
    std::vector<mho::Node*> _agents_nodes;

    double _r_r_ho_time,_r_h_ho_time,_h_h_ho_time;

    double _f_dist,_f_time,_f_agent_cost,_f_hri;
    double _max_dist, _max_time,_max_use_cost;

    bool _use_rrt_for_frontier_check;
    bool _use_dummy_smooth_for_frontier_check;

    bool _check_frontier_before;///< set it to true if you want to check all frontiers before the search

    // ///// //////////////////////////////
    // ///// FOR STATS   //////////////////
    // ///// //////////////////////////////
    enum STAT_FIELDS {SF_OK,SF_COST,SF_COST_LWA,
                      SF_AGENT_FROM, SF_AGENT_TO,
                      SF_EPSILON,SF_TIME_FACT,SF_AGENT_FACT,SF_HRI_FACT,SF_DIST_FACT,
                      SF_TIME_TOT,SF_TIME_LWA,SF_TIME_POSTPROC,
                      SF_TMP_COST_NB,  SF_TMP_COST_MEAN,   SF_TMP_COST_STDEV,
                      SF_TRUE_COST_NB, SF_TRUE_COST_MEAN,  SF_TRUE_COST_STDEV,
                      SF_HO_COST_NB,   SF_HO_COST_MEAN,    SF_HO_COST_STDEV,
                      SF_HEURISTIC_NB, SF_HEURISTIC_MEAN,  SF_HEURISTIC_STDEV,
                      SF_AGENT_NB,SF_HO_NB,SF_OBJ_DIST,SF_EXEC_TIME,// < stats about the solution
                      SF_SIZE};
    std::vector< std::vector < double > > _lwa_stats;


    FrontierCells::FrontierCheckStatus _max_frontier_check_status;
    MHO::ToolSet *_tool_set;
    std::vector<std::string> _handover_testers_defaults_set;

    std::vector<Configuration> _agents_init_confs;
    std::vector<Configuration> _agents_target_confs;
    std::vector<Configuration> _manipulables_init_confs;

    Robot *_first_agent;
    Robot *_target_agent;
    bool _random_first_target_agents;
    Robot *_object;
    std::vector<Eigen::Vector3d> _object_bb_vects;///< the vectors of the bounding box of the object
    std::vector<int> _object_directions_sorted_for_ho;///< the indices of the axis of the object (x,y,z) sorted by
                                                ///< descending order of the object size on each dimension
                                                ///< @todo wrong definition, is by increasing area of normal
                                                ///< surface of BB
    mho::Node *_first_agent_node;
    mho::Node *_target_agent_node;

    vector<pair<double,vector<mho::NodeInterface*> > > _raw_solution;

    bool _use_post_process;

    std::vector<MHO::MultiHandOverSolution*> _solutions;
    std::vector<MHO::Task*> _solutions_as_tasks;
    MHO::MultiHandOverSolution * _best_solution;
    //solution search parameters
    SearchAlgo_t _search_algo;
    AStarHeuristic_t _a_star_heuristic;
    //for LWA*
    double _epsilon;
    //for random search
    int _time_limit;
    int _max_iterations;
    int _overheat_temperature;

    bool _first_run;
    bool _compute_for_all_robots;
    bool _compute_traj_only;
    MHO::MultiHandOverSolution * _solution_to_compute_traj;
    unsigned int _step_to_compute_traj;

    //GRID
    HRICS::MultiAgentGrid * _grid;
    double _cellSize;

    bool _showText;
    bool _draw_object_frontiers_checks;
    bool _draw_agents_holding_conf_during_search;
    bool _draw_2d_path;
    unsigned int _solution_index_to_draw;
    unsigned int _solution_step_index_to_draw;
    bool _play_solution_task;
    bool _play_solution_task_traj;
    bool _play_solution_sync_step_by_step;
    bool _draw_all_2d_pathes;
    bool _draw_lwa;
    MHO::DisplayLWAstarInGrid* _display_lwa_in_grid_obj;
    MHO::MultiHandOverDisplay *_mho_display;
#ifdef DRAW_OPTIMIZE_FRONTIER
    bool _draw_optimize_frontier;
#endif
};

} // namespace HRICS

#endif // HRICS_HRICS_MULTIHANDOVER_H
