/**
 * @file Logger.h
 * @brief This file implements macros to help with the logging, in a way similar to ROS, using log4cxx.
 * @author Jules Waldhart
 * @date 22 april 2015
 *
 * The macros defined here have to be called to initialize and then use log4cxx in a .cpp file.
 * The macros allow to log for different levels, which are (ordered): TRACE < DEBUG < INFO < WARN < ERROR < FATAL
 * see http://wiki.ros.org/Verbosity%20Levels for information about the purpose of each level
 *
 * It is possible to select at execution time the lower log level that will be displayed for each logger (in each cpp file)
 * see http://logging.apache.org/log4cxx/usage.html for a short introduction to log4cxx
 *
 * You can directly use log4cxx API, but here you will find some helpers to make it easier and faster to implement.
 * Also, if for some reason the logging system needs to be changed, using the logm3d interface defined here will help in the
 * changing process.
 *
 * Names
 * -----
 *
 * The logger names have to be choosen wisely. There is a hierarchy of names, so for instance, planners is an ancestor of planners.hrics.mho and
 * the parent of planners.hrics.
 * By default, a logger inherits its parents property, but it can be overriden. So for debugging all HRICS module, one can disable debug infromation
 * for the root (by default, no debug information anywhere) and set planners.hrics logging level to debug, only the debug (and higher) level information will
 * be print.
 *
 * When naming a logger, you need to keep this in mind, so that an other developer/user can choose to display only what he needs. So do not hesitate to go deep in the hierarchy
 * (e.g. planners.hrics.mho.grid.cell)
 *
**/

#ifndef LOGGER_H
#define LOGGER_H

#ifndef MOVE3D_LOG_OFF
#include <log4cxx/logger.h>
#endif
#include <iostream>

/**
 * INIT_MOVE3D_LOG macro has to be called in each .cpp file where the logger will be used.
 * @param name is the name of the logger, according to log4cxx. e.g. "planners.api.traj" or "planners.hrics.visibility"
**/
#if !defined(MOVE3D_LOG_OFF)
#define INIT_MOVE3D_LOG(name) log4cxx::LoggerPtr __move3d_logger__  = logm3d::getLogger(name)
#else
#define INIT_MOVE3D_LOG(name)
#endif

/**
 * MOVE3D_STATIC_LOGGER macro defines a static logger for the class.
 * It the has to be initialized (defined) using the INIT_MOVE3D_STATIC_LOGGER
**/
#if !defined(MOVE3D_LOG_OFF)
#define MOVE3D_STATIC_LOGGER static  log4cxx::LoggerPtr __move3d_logger__
#else
#define MOVE3D_STATIC_LOGGER
#endif
/**
 * INIT_MOVE3D_STATIC_LOGGER() macro is to be called in the class implementation file (.cpp) to instantiate the static logger.
 * The static logger has to be defined first with MOVE3D_STATIC_LOGGER macro in the class definition (.h)
**/
#if !defined(MOVE3D_LOG_OFF)
#define INIT_MOVE3D_STATIC_LOGGER(classname,loggername) log4cxx::LoggerPtr classname::__move3d_logger__ = logm3d::getLogger(loggername)
#else
#define INIT_MOVE3D_STATIC_LOGGER(classname,loggername)
#endif

/**
 * The following macros does not need INIT_MOVE3D_LOG to be called, and can be
 * used to have different loggers in a same cpp file.
 * @param _logger_ a logm3d::LoggerPtr to use as logger
 * @param msg a std::string, char*, wchar*, the insertion operator (<<) can be used (see log4cxx doc)
 *
 * These loggers are obtained through the various logm3d::getLogger() and similar functions,
 * (or log4cxx::Logger::getLogger() : to be avoided, create a logm3d logger instead)
 **/
#if !defined(MOVE3D_LOG_ERROR_ONLY) && !defined(MOVE3D_LOG_OFF)
#define LOGM3D_TRACE(_logger_,msg) LOG4CXX_TRACE(_logger_,msg)
#define LOGM3D_DEBUG(_logger_,msg) LOG4CXX_DEBUG(_logger_,msg)
#define LOGM3D_INFO(_logger_,msg)  LOG4CXX_INFO(_logger_,msg)
#else
#define LOGM3D_TRACE(_logger_,msg)
#define LOGM3D_DEBUG(_logger_,msg)
#define LOGM3D_INFO(_logger_,msg)
#endif
#if !defined(MOVE3D_LOG_OFF)
#define LOGM3D_WARN(_logger_,msg)  LOG4CXX_WARN(_logger_,msg)
#define LOGM3D_ERROR(_logger_,msg) LOG4CXX_ERROR(_logger_,msg)
#define LOGM3D_FATAL(_logger_,msg) LOG4CXX_FATAL(_logger_,msg)
#else
#define LOGM3D_WARN(_logger_,msg)
#define LOGM3D_ERROR(_logger_,msg)
#define LOGM3D_FATAL(_logger_,msg)
#endif

/**
 * the following macros are shorter for the log4cxx macros, and uses the declarations made by INIT_MOVE3D_LOG()
 * they will log with the logger named by the call to INIT_MOVE3D_LOG(name)
**/
#if !defined(MOVE3D_LOG_ERROR_ONLY) && !defined(MOVE3D_LOG_OFF)
#define M3D_TRACE(msg) LOGM3D_TRACE(__move3d_logger__,msg)
#define M3D_DEBUG(msg) LOGM3D_DEBUG(__move3d_logger__,msg)
#define M3D_INFO(msg)  LOGM3D_INFO(__move3d_logger__,msg)
#else
#define M3D_TRACE(msg)
#define M3D_DEBUG(msg)
#define M3D_INFO(msg)
#endif
#if !defined(MOVE3D_LOG_OFF)
#define M3D_WARN(msg)  LOGM3D_WARN(__move3d_logger__,msg)
#define M3D_ERROR(msg) LOGM3D_ERROR(__move3d_logger__,msg)
#define M3D_FATAL(msg) LOGM3D_FATAL(__move3d_logger__,msg)
#else
#define M3D_WARN(msg)
#define M3D_ERROR(msg)
#define M3D_FATAL(msg)
#endif


namespace logm3d {

#ifndef MOVE3D_LOG_OFF
typedef log4cxx::LoggerPtr LoggerPtr ;
#else
typedef int LoggerPtr;
#endif

/// to be called to initialize the logging system to its defaults
/// (once in a session)
void initializePlannerLogger();


LoggerPtr getLogger(std::string name);

LoggerPtr getLoggerDataFile(std::string name, std::string file_path);
LoggerPtr getLoggerConsoleAndFile(std::string name,std::string file_path);
}

#endif // LOGGER_H

