/*
 * DirectedRRT.cpp
 *
 *  Created on: Jan 2015
 *      Author: aboeuf
 */

#include "API/Roadmap/compco.hpp"
#include "API/Roadmap/graph.hpp"
#include "planner/planEnvironment.hpp"
#include "DirectedRRT.hpp"

using namespace std;
using namespace API;

bool DirectedExpansion::expandGoalTree(){
    return _expandGoalTree;
}

void DirectedExpansion::switchTrees(){
    _expandGoalTree=!_expandGoalTree;
}

unsigned DirectedExpansion::extend(Node * fromNode,confPtr_t directionConf,Node * directionNode){

    confPtr_t fromConf,dirConf;
    if(_expandGoalTree){
        fromConf=directionConf;
        dirConf=fromNode->getConfiguration();
    }
    else{
        fromConf=fromNode->getConfiguration();
        dirConf=directionConf;
    }

    if(ENV.getBool(Env::hoveringShoot)){
        if(p3d_has_a_quad_joint(dirConf->getRobot()->getRobotStruct())){
            lm_quadspline_str *param=lm_get_quadspline_lm_param(dirConf->getRobot()->getRobotStruct());
            if(param){
                configPt q=dirConf->getConfigStruct();
                for(int k=0;k<6;k++){
                    q[param->vx_dof+k]=0;
                }
            }
        }
    }

    // path along which the extension is performed
    LocalPath directionPath(fromConf,dirConf);
    double pathLength=directionPath.getParamMax();

    // construct an 'extension path', by interpolation along the 'direction path'
    double param;
    localPathPtr_t extensionPath;
    if(_expandGoalTree){
        param=max(0.0,pathLength-this->step());
        extensionPath.reset(new LocalPath(directionPath.configAtParam(param),dirConf));
    }
    else{
        param=min(this->step(),pathLength);
        extensionPath.reset(new LocalPath(fromConf,directionPath.configAtParam(param)));
    }

    // if the extension path is valid, add a new node to the graph
    if(extensionPath->isValid()){
        this->expansionSucceeded(fromNode);
        bool existingNode=(param==pathLength && !_expandGoalTree);
        existingNode = existingNode || (param==0.0 && _expandGoalTree);
        existingNode = existingNode && directionNode;
        this->addNode(extensionPath, fromNode, existingNode ? directionNode : NULL);
        return (existingNode ? 0 : 1);
    }
    else{
        this->expansionFailed(fromNode);
        return 0;
    }
}

unsigned DirectedExpansion::connect(Node * fromNode,confPtr_t directionConf,Node * directionNode){

    confPtr_t fromConf,dirConf;
    if(_expandGoalTree){
        fromConf=directionConf;
        dirConf=fromNode->getConfiguration();
    }
    else{
        fromConf=fromNode->getConfiguration();
        dirConf=directionConf;
    }

    if(ENV.getBool(Env::hoveringShoot)){
        if(p3d_has_a_quad_joint(dirConf->getRobot()->getRobotStruct())){
            lm_quadspline_str *param=lm_get_quadspline_lm_param(dirConf->getRobot()->getRobotStruct());
            if(param){
                configPt q=dirConf->getConfigStruct();
                for(int k=0;k<6;k++){
                    q[param->vx_dof+k]=0;
                }
            }
        }
    }

    // path along which the connection is performed
    LocalPath directionPath(fromConf,dirConf);

    // size of the expansion step
    double pathLength=directionPath.getParamMax();
    double stepSize;
    if(_expandGoalTree)
        stepSize=max(0.0,pathLength-this->step());
    else
        stepSize=min(this->step(),pathLength);

    double param=stepSize;
    bool success(true);
    confPtr_t currentConf=fromConf;
    localPathPtr_t connectPath;

    // attempt the connection step by step
    while (success && ((!_expandGoalTree && param<=pathLength) || (_expandGoalTree && 0<=param)))
    {

        confPtr_t newConf=directionPath.configAtParam(param);

        if(ENV.getBool(Env::hoveringShoot)){
            if(p3d_has_a_quad_joint(newConf->getRobot()->getRobotStruct())){
                lm_quadspline_str *param=lm_get_quadspline_lm_param(newConf->getRobot()->getRobotStruct());
                if(param){
                    configPt q=newConf->getConfigStruct();
                    for(int k=0;k<6;k++){
                        q[param->vx_dof+k]=0;
                    }
                }
            }
        }

        // for quadspline, the entire path as to be tested because its different from the concatenation of all portion paths
        if(directionPath.getType()==QUADSPLINE){
            if(_expandGoalTree)
                connectPath.reset(new LocalPath(newConf,dirConf));
            else
                connectPath.reset(new LocalPath(fromConf,newConf));
        }
        // otherwise the portion path is enought
        else{
            if(_expandGoalTree)
                connectPath.reset(new LocalPath(newConf,currentConf));
            else
                connectPath.reset(new LocalPath(currentConf,newConf));
        }

        // check whether the current path is valid
        success=connectPath.get()->isValid();

        // if it is go to the next portion
        if (success)
        {
            currentConf=newConf;

            if(_expandGoalTree){
                if(param==0)
                    param=-1.;
                else if(param>stepSize+1e-3)
                    param-=stepSize;
                else
                    param=0.;
            }
            else{
                if (param == pathLength)
                    param = pathLength + 1.;
                else if (pathLength - param > stepSize + 1e-3)
                    param += stepSize;
                else
                    param = pathLength;
            }
        }
    }

    // if "reached" or "advanced", add a node to the graph
    if (success || param!=stepSize)
    {
        if(directionPath.getType()==QUADSPLINE){
            if(!success){
                if(_expandGoalTree)
                    connectPath.reset(new LocalPath(currentConf,dirConf));
                else
                    connectPath.reset(new LocalPath(fromConf,currentConf));
            }
        }
        else{
            if(_expandGoalTree)
                connectPath.reset(new LocalPath(currentConf,dirConf));
            else
                connectPath.reset(new LocalPath(fromConf,currentConf));
        }
        this->expansionSucceeded(fromNode);
        bool existingNode = ((connectPath.get()->getEnd() == directionPath.getEnd()) && directionNode);
        this->addNode(connectPath, fromNode, existingNode ? directionNode : NULL);
        return (existingNode ? 0 : 1);
    }
    else
    {
        this->expansionFailed(fromNode);
        return 0;
    }

    return 0;

}

Node* DirectedExpansion::addNode(localPathPtr_t path, Node * fromNode, Node * directionNode){

    // if the expansion was performed toward an already existing node, both components are merged
    if(directionNode){
        if(_expandGoalTree)
            m_Graph->addOrientedEdge(directionNode,fromNode,path);
        else
            m_Graph->addOrientedEdge(fromNode,directionNode,path);
        return directionNode;
    }
    // otherwise, a new node is created in the graph and linked to fromNode
    else
    {
        Node * newNode;
        if(_expandGoalTree){
            newNode=new Node(m_Graph,path->getBegin());
            m_Graph->addNode(newNode);
            if(newNode->getConfiguration()->isConnectible())
                _nbConnectibleNodes++;
            m_Graph->addOrientedEdge(newNode,fromNode,path);
        }
        else{
            newNode=new Node(m_Graph,path->getEnd());
            m_Graph->addNode(newNode);
            if(newNode->getConfiguration()->isConnectible())
                _nbConnectibleNodes++;
            m_Graph->addOrientedEdge(fromNode,newNode,path);
        }
        return newNode;
    }
}

bool DirectedExpansion::addCycles(){

    unsigned nbCycles(0);
    vector<Node *> nodes = m_Graph->getNodes();
    Node * lastNode = m_Graph->getLastNode();
    confPtr_t lastConf = lastNode->getConfiguration();
    double distanceThreshold = ENV.getDouble(Env::minimalFinalExpansionGap) * this->step();

    for(unsigned int i=0;i<nodes.size()-1;++i){

        confPtr_t nodeConf=nodes[i]->getConfiguration();

        if(_expandGoalTree){
            localPathPtr_t path(new LocalPath(nodeConf,lastConf));
            double spaceDistance=path->getParamMax();
            //if(spaceDistance<=distanceThreshold){
                double graphDistance=m_Graph->getAStarShortestPathLength(nodeConf,lastConf);
                if(spaceDistance<graphDistance && path->isValid()){
                    m_Graph->addOrientedEdge(nodes[i],lastNode,path);
                    ++nbCycles;
                }
            //}
        }
        else{
            localPathPtr_t path(new LocalPath(lastConf,nodeConf));
            double spaceDistance=path->getParamMax();
            //if(spaceDistance<=distanceThreshold){
                double graphDistance=m_Graph->getAStarShortestPathLength(lastConf,nodeConf);
                if(spaceDistance<graphDistance && path->isValid()){
                    m_Graph->addOrientedEdge(lastNode,nodes[i],path);
                    ++nbCycles;
                }
            //}
        }

    }

    return (nbCycles > 0);
}

unsigned DirectedRRT::init() {
    expansion = new DirectedExpansion(_Graph);
    unsigned int nb_init_nodes=TreePlanner::init();
    _Graph->resetConnections();
    _Start->backwardConnectToSource();
    _Goal->forwardConnectToGoal();
    this->updateConnections();
    return nb_init_nodes;
}

bool DirectedRRT::trajFound(){
    return (_Start->isForwardConnectedToGoal() && _Goal->isBackwardConnectedToSource());
}

unsigned DirectedRRT::expandOneStep(ConnectedComponent *fromComp){

    Node * directionNode(NULL);
    confPtr_t directionConf;
    p3d_rob* robotPt=_Robot->getRobotStruct();

    DirectedExpansion *exp=(DirectedExpansion*)expansion;

    // sample the expansion direction
    // taking into account the goal bias, if relevant
    bool authorizedBias = !PlanEnv->getBool(PlanParam::anytime) || !this->trajFound();
    if (ENV.getBool(Env::isGoalBiased) && authorizedBias && this->goalBias()){
        // bias toward the start configuration
        if(exp->expandGoalTree()){
            directionNode=_Start;
            directionConf=_q_start;
        }
        // bias toward the goal configuration
        else{
            directionNode=_Goal;
            directionConf=_q_goal;
        }
    }
    else
        // or random sampling in the entire C-Space
        directionConf=_Robot->shoot();

    // select, from the component to be grown, the node from which the expansion will be attempted
    Node *expandNode;
    if(ENV.getBool(Env::randomConnectionToGoal))
        expandNode=fromComp->randomNode();
    else
        expandNode=fromComp->nearestNeighbor(directionConf,exp->expandGoalTree());

    // In case of a plannification only for the quadrotor
    if(ENV.getBool(Env::quadOnly)){
        lm_quadspline_str *param=lm_get_quadspline_lm_param(robotPt);
        if(!param){
            printf("error: DirectedRRT::expandOneStep: no quadspline structure initialized for robot %s\n",robotPt->name);
            return 0;
        }
        configPt qi=expandNode->getConfiguration()->getConfigStruct();
        configPt qf=directionConf->getConfigStruct();
        for(int k=param->az_dof+1;k<robotPt->nb_dof;k++){
            qf[k]=qi[k];
        }
    }

    // Fixing position of the gripper in case of robot bonebreaker
    if(strstr(robotPt->name,"bonebreaker") && robotPt->grasped_objects){
        lm_quadspline_str *param=lm_get_quadspline_lm_param(robotPt);
        if(!param){
            printf("error: DirectedRRT::expandOneStep: no quadspline structure initialized for robot %s\n",robotPt->name);
            return 0;
        }
        configPt qi=expandNode->getConfiguration()->getConfigStruct();
        configPt qf=directionConf->getConfigStruct();
        qf[robotPt->nb_dof-2]=qi[robotPt->nb_dof-2];
        qf[robotPt->nb_dof-1]=qi[robotPt->nb_dof-1];
    }

    // directed expansion
    switch (ENV.getExpansionMethod())
    {
    case Env::Extend:
        return exp->extend(expandNode,directionConf,directionNode);
    case Env::Connect:
        return exp->connect(expandNode,directionConf,directionNode);
    default:
        printf("error: DirectedRRT::expandOneStep: method not implemented\n");
        return 0;
    }

}

bool DirectedRRT::checkStopConditions(){

    if(TreePlanner::checkStopConditions())
        return true;

    if(ENV.getBool(Env::expandToGoal) && !PlanEnv->getBool(PlanParam::anytime)){
        if(this->trajFound()){
            cout<<"Trajectory found"<<endl;
            return true;
        }
    }

    return false;
}

unsigned DirectedRRT::run(){

    p3d_rob *robotPt=p3d_get_current_robot();
    DirectedExpansion *exp=(DirectedExpansion*)expansion;

    if(ENV.getBool(Env::quadOnly)){
            lm_quadspline_str *param=lm_get_quadspline_lm_param(robotPt);
            bool precond=(param!=NULL);
            if(precond){
                configPt qi=_q_start->getConfigStruct();
                configPt qf=_q_goal->getConfigStruct();
                for(int k=param->az_dof+1;k<robotPt->nb_dof;k++){
                    precond=precond && (fabs(qi[k]-qf[k])<EPS10);
                }
            }
            if(!precond){
                printf("error: DirectedRRT::run: pre-conditions are not satisfied for quadOnly planning.\n");
                return 0;
            }
            else{
                printf("DirectedRRT::run: quadOnly precond OK\n");
            }
            robotPt->lpl_type=QUADSPLINE;
            ENV.setBool(Env::isHybrid,false);
    }

    this->chronoStart();

    // check whether the pre-conditions are satisfied
    if (!this->preConditions()){
        cout << "error: DirectedRRT::run: pre-conditions are not satisfied." << endl;
        return 0;
    }

    // check whether this is a trivial planning scenario
    localPathPtr_t directPath(new LocalPath(_q_start, _q_goal));
    if (ENV.getBool(Env::expandToGoal) && exp->isValidConnection(directPath)){
        _Graph->addOrientedEdge(_Start,_Goal, directPath);
        cout <<"DirectedRRT::run: direct connection"<< endl;
        return 0;
    }

    unsigned nbCreatedNodes(0); // number of nodes created by one expansion step
    unsigned totalCreatedNodes(0); // total number of nodes created so far
    ConnectedComponent * fromComp; // component that will be expanded during the current step
    ConnectedComponent * toComp; // component toward which to expand (when relevant)

    fromComp = _Start->getConnectedComponent();
    toComp = _Goal->getConnectedComponent();

    // variable used to keep track of the cost of the best trajectory found so far, in the anytime variant
    double bestTrajCost = numeric_limits<double>::max();
    bool firstTraj(true);

    // RRT loop: start by checking the stopping conditions
    while (!this->checkStopConditions())
    {


        // in the case of a balanced Bi-directional RRT, do not expand when components are unbalanced
        if (!ENV.getBool(Env::biDir) || !ENV.getBool(Env::expandBalanced) ||
                (fromComp->getNumberOfNodes() < toComp->getNumberOfNodes() + 5))
        {
            ++m_nbExpansion;

            // perform a single expansion step of RRT
            nbCreatedNodes = this->expandOneStep(fromComp);

            // success, if some nodes have been created
            if (nbCreatedNodes > 0)
            {
                totalCreatedNodes += nbCreatedNodes;
                m_nbConscutiveFailures = 0;

                Node * lastNode = _Graph->getLastNode();
                confPtr_t lastConf = lastNode->getConfiguration();

                // attempt a connection toward the goal / the other component (when relevant)
                if (ENV.getBool(Env::expandToGoal) && !this->trajFound())
                {
                    // Bi-directional RRT
                    if (ENV.getBool(Env::biDir))
                    {
                        // look for the node whose configuration is the closest to the last configuration
                        // within the 'toComp' connected component
                        Node * closestNode=toComp->nearestNeighbor(lastConf,exp->expandGoalTree());

                        // create local path
                        localPathPtr_t path;
                        if(exp->expandGoalTree())
                            path.reset(new LocalPath(closestNode->getConfiguration(),lastConf));
                        else
                            path.reset(new LocalPath(lastConf,closestNode->getConfiguration()));

                        // if the path between both components is valid, connect them
                        if (exp->isValidConnection(path)){
                            if(exp->expandGoalTree())
                                _Graph->addOrientedEdge(closestNode,lastNode,path);
                            else
                                _Graph->addOrientedEdge(lastNode,closestNode,path);
                            _Start->forwardConnectToGoal();
                            _Goal->backwardConnectToSource();
                            this->updateConnections();
                            toComp = fromComp = _Start->getConnectedComponent();
                        }
                    }
                    // mono-directional RRT (with goal)
                    else
                    {
                        // if the path between the last node and the goal is valid, connect them
                        localPathPtr_t path(new LocalPath(lastConf,_q_goal));
                        if (exp->isValidConnection(path)){
                            _Graph->addOrientedEdge(lastNode,_Goal,path);
                            _Start->forwardConnectToGoal();
                            _Goal->backwardConnectToSource();
                            toComp = fromComp = _Start->getConnectedComponent();
                            this->updateConnections();
                        }
                    }
                }

                // anytime variant
                bool drawNewTraj = false;
                if (PlanEnv->getBool(PlanParam::anytime) && this->trajFound()){

                    bool extractNewTraj(true);

                    // try to add cycles (when relevant)
                    if (ENV.getBool(Env::addCycles))
                        extractNewTraj = (firstTraj || expansion->addCycles());

                    // extract the new trajectory from the graph (when the display is enabled)
                    if (extractNewTraj && !ENV.getBool(Env::drawDisabled))
                    {
                        Trajectory * trajectory = this->extractTrajectory();
                        if (trajectory)
                        {
                            double trajCost = trajectory->cost();
                            if (trajCost < bestTrajCost)
                            {
                                // keep track of the best trajectory found so far
                                bestTrajCost = trajCost;
                                trajectory->replaceP3dTraj();
                                drawNewTraj = ENV.getBool(Env::drawTraj);
                                cout << "best trajectory cost = " << bestTrajCost << endl;

                                this->trajCosts.push_back(bestTrajCost);
                            }
                            delete trajectory;
                            firstTraj = false;
                        }
                    }
                }

                // redraw the scene (when relevant)
                if (!ENV.getBool(Env::drawDisabled) && (ENV.getBool(Env::drawExploration) || drawNewTraj))
                    _draw_func();

            }
            // failure, otherwise
            else {
                ++m_nbFailedExpansion;
                ++m_nbConscutiveFailures;
            }
        }
        this->chronoUpdate();

        // in the case of the Bi-directional RRT, swap the roles of the connected components
        if (ENV.getBool(Env::biDir)){
            swap(fromComp,toComp);
            exp->switchTrees();
        }
    }

    this->chronoStop();
    return totalCreatedNodes;
}

void DirectedRRT::updateConnections(){

    _Graph->unflagAll();
    _Start->updateForward();

    _Graph->unflagAll();
    _Goal->updateBackward();

    cout<<"Connections updated"<<endl;
}

void DirectedRRT::pause(){
    bool was_running=ENV.getBool(Env::isRunning);
    g3d_draw_allwin_active();
    ENV.setBool(Env::isRunning,false);
    int c;
    printf(" ... [PAUSED] ...");
    fflush( stdout );
    do c = getchar(); while ((c != '\n') && (c != EOF));
    ENV.setBool(Env::isRunning,was_running);
}
