/*
 * Transition-RRT.cpp
 *
 *  Created on: Nov 20, 2012
 *      Author: ddevaurs
 */

#include "API/Roadmap/compco.hpp"
#include "API/Roadmap/graph.hpp"

#include "planner/Diffusion/Variants/Transition-RRT.hpp"

using namespace std;


/**
 * Constructor.
 */
TransitionExpansion::TransitionExpansion(Graph * graph) : RRTExpansion(graph)
{
	// set the temperature
	temperature = ENV.getDouble(Env::initialTemperature);

	// set the lowest and highest costs
	lowestCost = numeric_limits<double>::max();
	highestCost = -1. * numeric_limits<double>::max();
	vector<Node *> nodes = m_Graph->getNodes();
	for (vector<Node *>::iterator n = nodes.begin(); n != nodes.end(); ++n)
		this->updateCostRange((*n)->cost());
}


/**
 * T-RRT Extend.
 * @param fromNode: node from which the expansion is performed
 * @param directionConf: configuration toward which the expansion is going
 * @param directionNode: node toward which the expansion is going (when relevant)
 * @return the number of created nodes
 */
unsigned TransitionExpansion::extend(Node * fromNode, confPtr_t directionConf, Node * directionNode)
{
	confPtr_t fromConf = fromNode->getConfiguration();
	ConnectedComponent * fromComp = fromNode->getConnectedComponent();

	// path along which the extension is performed
	LocalPath directionPath(fromConf, directionConf);
	double pathLength = directionPath.getParamMax();
	double param = min(this->step(), pathLength);

	// Refinement control:
	// ((param == pathLength) && !directionNode) means that we may be about to add a new refinement node;
	// in that case, if the number of refinement nodes already present in the component is too high
	// (i.e. more than 10% of the total number of nodes), the expansion attempt is canceled
	if (ENV.getBool(Env::refinementControl) && ((param == pathLength) && !directionNode)
			&& (fromComp->getNumberOfRefinementNodes() > 0.1 * fromComp->getNumberOfNodes()))
		return 0;

	// construct an 'extension path', by linear interpolation along the 'direction path'
	localPathPtr_t extensionPath(new LocalPath(fromConf, directionPath.configAtParam(param)));

	// check the validity of the extension path, in terms of collisions and cost conditions
	bool success;
	if (ENV.getBool(Env::costBeforeColl))
		success = this->isCostValid(extensionPath) && extensionPath->isValid();
	else
		success = extensionPath->isValid() && this->isCostValid(extensionPath);

	// if the validation succeeds, add a new node to the graph
	if (success)
	{
		this->expansionSucceeded(fromNode);
		bool existingNode = ((param == pathLength) && directionNode);
		this->addNode(extensionPath, fromNode, existingNode ? directionNode : NULL);

		// if a new node has been created
		if (!existingNode)
		{
			// update the cost range, based on the cost of the new node
			this->updateCostRange(extensionPath->getEnd()->cost());

			// Refinement control:
			// if a new refinement node has been built, update the number of refinement nodes of the component
			if (ENV.getBool(Env::refinementControl) && (param == pathLength))
				fromComp->updateNumberOfRefinementNodes();
		}

		return (existingNode ? 0 : 1);
	}
	else
	{
		this->expansionFailed(fromNode);
		return 0;
	}
}


/**
 * T-RRT Connect.
 * @param fromNode: node from which the expansion is performed
 * @param directionConf: configuration toward which the expansion is going
 * @param directionNode: node toward which the expansion is going (when relevant)
 * @return the number of created nodes
 */
unsigned TransitionExpansion::connect(Node * fromNode, confPtr_t directionConf, Node * directionNode)
{
    confPtr_t fromConf = fromNode->getConfiguration();
	ConnectedComponent * fromComp = fromNode->getConnectedComponent();

	// path along which the connection is performed
	LocalPath directionPath(fromConf, directionConf);
	double pathLength = directionPath.getParamMax();

	// size of the validation step
	double stepSize = min(this->step(), pathLength);
	double param = stepSize;

	// Refinement control:
	// ((param == pathLength) && !directionNode) means that we may be about to add a new refinement node;
	// in that case, if the number of refinement nodes already present in the component is too high
	// (i.e. more than 10% of the total number of nodes), the expansion attempt is canceled
	if (ENV.getBool(Env::refinementControl) && ((param == pathLength) && !directionNode)
			&& (fromComp->getNumberOfRefinementNodes() > 0.1 * fromComp->getNumberOfNodes()))
		return 0;

	double totalCost(0.); // cost of the connection path
	double reverseCost(0.); // cost of its reverse path

	// attempt the connection step by step
	confPtr_t currentConf = fromConf;
	bool success(true);
	while (success && (param <= pathLength))
	{
		confPtr_t newConf = directionPath.configAtParam(param);
		localPathPtr_t pathPortion(new LocalPath(currentConf, newConf));

		// check the validity of the path portion, in terms of collisions and cost conditions
		if (ENV.getBool(Env::costBeforeColl))
			success = this->isCostValid(pathPortion) && pathPortion->isValid();
		else
			success = pathPortion->isValid() && this->isCostValid(pathPortion);

		if (success)
		{
			// update the cost of the connection path
			totalCost += pathPortion->cost();
			reverseCost += pathPortion->getReverseCost();

			// go to the next portion
			currentConf = newConf;
			if (param == pathLength)
				param = pathLength + 1.;
			else if (pathLength - param > stepSize + 1e-3)
				param += stepSize;
			else
				param = pathLength;
		}
	}

    // if "reached" or "advanced", add a node to the graph
	if (success || (param > stepSize))
	{
		this->expansionSucceeded(fromNode);
		localPathPtr_t connectPath(new LocalPath(fromConf, currentConf));
		connectPath->setCost(totalCost);
		connectPath->setReverseCost(reverseCost);
		bool existingNode = ((currentConf == directionPath.getEnd()) && directionNode);
		this->addNode(connectPath, fromNode, existingNode ? directionNode : NULL);

		// if a new node has been created
		if (!existingNode)
		{
			// update the cost range, based on the cost of the new node
			this->updateCostRange(currentConf->cost());

			// Refinement control:
			// if a refinement node has been built, update the number of refinement nodes of the component
			if (ENV.getBool(Env::refinementControl) && (param == pathLength))
				fromComp->updateNumberOfRefinementNodes();
		}

		return (existingNode ? 0 : 1);
	}
	else
	{
		this->expansionFailed(fromNode);
		return 0;
	}
}


/**
 * Check whether the given path is a valid connection, and compute its cost.
 * @return TRUE if the connection is valid, and FALSE otherwise
 */
bool TransitionExpansion::isValidConnection(localPathPtr_t path)
{
	// N.B.: the connection is attempted only if the path length is less than a specified threshold
	// (which is some multiple of the extension step)
	if (path->getParamMax() <= ENV.getDouble(Env::minimalFinalExpansionGap) * this->step())
	{
		// check whether the path is valid and satisfies the cost conditions
		if (ENV.getBool(Env::costBeforeColl))
			return (this->isCostValid(path, false) && path->isValid());
		else
			return (path->isValid() && this->isCostValid(path, false));
	}
	else
		return false;
}


/**
 * Try to add cycles to the graph.
 * @return TRUE if at least one cycle has been added, and FALSE otherwise
 */
bool TransitionExpansion::addCycles()
{
	unsigned nbCycles(0);
	vector<Node *> nodes = m_Graph->getNodes();
	Node * lastNode = m_Graph->getLastNode();
	confPtr_t lastConf = lastNode->getConfiguration();
	double distanceThreshold = ENV.getDouble(Env::minimalFinalExpansionGap) * this->step();
	double squareDistanceThreshold = distanceThreshold * distanceThreshold;

	// compute the cost-distance (in the graph) between the last node and every other node
	vector<double> graphDistance = m_Graph->computeDijkstraShortestPaths(lastNode);

	for (unsigned i(0); i < nodes.size() - 1; ++i)
	{
		// do not consider nodes that are too far away from the last node
		// as target candidates for a new edge
		if (lastConf->squareDistance(nodes[i]->getConfiguration().get()) <= squareDistanceThreshold)
		{
			// for a cycle to be considered useful, the cost of the associated path (in C-space)
			// has to be smaller than the cost of the alternative path in the graph
			localPathPtr_t path(new LocalPath(lastConf, nodes[i]->getConfiguration()));
			bool success;
			if (ENV.getBool(Env::costBeforeColl))
				success = (max(path->cost(), path->getReverseCost()) < graphDistance[i]) && path->isValid();
			else
				success = path->isValid() && (max(path->cost(), path->getReverseCost()) < graphDistance[i]);

			if (success)
			{
				m_Graph->addEdges(lastNode, nodes[i], path);
				++nbCycles;
			}
		}
	}

	// return true if at least one cycle has been added, and false otherwise
	return (nbCycles > 0);
}

bool TransitionExpansion::addCycles1()
{//Instead of trying to connect only the last added node at its closer neighbors
    //we try to connect all the nodes to k nodes
    //k is function of the successRate of adding cycle in the tree

    double successRate = 0;
    vector<Node *> nodes = m_Graph->getSortedNodes();

    static int k = 3;
    static double successRatePrev = 0;
    
    double distanceThreshold = ENV.getDouble(Env::minimalFinalExpansionGap) * this->step();
    double squareDistanceThreshold = distanceThreshold * distanceThreshold;
    int position = rand() % (nodes.size() - k);
    Node * current = nodes.at(position);

        for(uint j = 1 ; j < k;j++)
        {
            Node * newLink = nodes.at(position+j);
            if (current->getConfiguration()->squareDistance(newLink->getConfiguration().get()) <= squareDistanceThreshold)
            {
                    if(!m_Graph->isEdgeInGraph(current, newLink))
                    {
                        
                                double graphDistance = this->m_Graph->extractAStarShortestPath(current, newLink)->getRangeMax();

                                localPathPtr_t path(new LocalPath(current->getConfiguration(), newLink->getConfiguration()));
                                bool success;
                                if (ENV.getBool(Env::costBeforeColl))
                                    success = (max(path->cost(), path->getReverseCost()) < graphDistance) && path->isValid();
                                else
                                    success = path->isValid() && (max(path->cost(), path->getReverseCost()) < graphDistance);

                                if (success)
                                {
                                    // if the path is valid, add a useful cycle
                                    m_Graph->addEdges(current, newLink, path);
                                    successRate += 2/(current->getNodeStruct()->dist_Nnew + newLink->getNodeStruct()->dist_Nnew);
                                }
                        
                    }
            }

    }

    if(successRate == 0 && successRatePrev == 0)
    {
        k += 1;
    }
    else if(successRate > 0 && successRatePrev > 0)
    {
        k -= 1;
    }

    k = (k <= 2)?3:k;

    successRatePrev = successRate;

    // return TRUE if at least one cycle has been added, and FALSE otherwise
    return (successRate > 0);
}

bool TransitionExpansion::addCycles2()
{//Instead of trying to connect only the last added node at its closer neighbors
    //we try to connect all the nodes of the current trajectory to a random node of the graph
    //we limit at k try (function of the successRate)

    double successRate = 0;
    vector<Node *> nodes = m_Graph->getNotTrajNodes();

    static int k = 3;
    static double successRatePrev = 0;
    
    double distanceThreshold = ENV.getDouble(Env::minimalFinalExpansionGap) * this->step();
    double squareDistanceThreshold = distanceThreshold * distanceThreshold;
    int position = rand() % (nodes.size());
    Node * current = nodes.at(position);
    map<double, Node *> nodesToConnect = m_Graph->getNodesOnTraj(current);
    
	std::map<double,Node *>::iterator it;
	uint j = 0;
    for(it = nodesToConnect.begin() ; it!=nodesToConnect.end();++it)
    {
        if( j == k )
        {
                break;
        }
        else
        {
                j++;
        }
            
        Node * newLink = it->second;
        if (it->first <= squareDistanceThreshold)
        {
                if(!m_Graph->isEdgeInGraph(current, newLink))
                {
                    
                            double graphDistance = this->m_Graph->extractAStarShortestPath(current, newLink)->getRangeMax();

                            localPathPtr_t path(new LocalPath(current->getConfiguration(), newLink->getConfiguration()));
                            bool success;
                            if (ENV.getBool(Env::costBeforeColl))
                                success = (max(path->cost(), path->getReverseCost()) < graphDistance) && path->isValid();
                            else
                                success = path->isValid() && (max(path->cost(), path->getReverseCost()) < graphDistance);

                            if (success)
                            {
                                // if the path is valid, add a useful cycle
                                m_Graph->addEdges(current, newLink, path);
                                successRate += 2/(current->getNodeStruct()->dist_Nnew + newLink->getNodeStruct()->dist_Nnew);
                            }
                    
                }
        }

    }

    if(successRate == 0 && successRatePrev == 0)
    {
        k += 1;
    }
    else if(successRate > 0 && successRatePrev > 0)
    {
        k -= 1;
    }

    k = (k <= 2)?3:k;

    successRatePrev = successRate;

    // return TRUE if at least one cycle has been added, and FALSE otherwise
    return (successRate > 0);
}


bool TransitionExpansion::addCycles3()
{   //Instead of trying to connect only the last added node at its closer neighbors
    //we try to connect all the nodes of the current trajectory to a random node of the graph
    //we do not limit the number of try
    double distanceThreshold = ENV.getDouble(Env::minimalFinalExpansionGap) * this->step();
    double squareDistanceThreshold = distanceThreshold * distanceThreshold;
    
    Node* current = m_Graph->getRandomNodeNearTraj();

    map<double, Node *> nodesToConnect = m_Graph->getNodesOnTraj(current,true);
	std::map<double,Node *>::iterator it;
	bool success = false;
	int i = 0;
    for(it = nodesToConnect.begin() ; it!=nodesToConnect.end();++it)
    {
        Node * newLink = it->second;
        
        if (it->second != current && it->first <= squareDistanceThreshold)
        {
                if(!m_Graph->isEdgeInGraph(current, newLink) && !m_Graph->isEdgeInGraph(newLink, current))
                {

                            localPathPtr_t path(new LocalPath(current->getConfiguration(), newLink->getConfiguration()));
                            double graphDistance = this->m_Graph->extractAStarShortestPath(current, newLink)->getRangeMax();
                            if (ENV.getBool(Env::costBeforeColl))
                                success = (max(path->cost(), path->getReverseCost()) < graphDistance) && path->isValid();
                            else
                                success = path->isValid() && (max(path->cost(), path->getReverseCost()) < graphDistance);

                            if (success)
                            {
                                // if the path is valid, add a useful cycle
                                m_Graph->addEdges(current, newLink, path);
                            }
                    
                }
        }
        i++;
    }

    // return TRUE if at least one cycle has been added, and FALSE otherwise
    return success;
}


/**
 * Check whether the given path satisfies the cost constraints, and compute its cost.
 * @param temperatureTuning: with or without temperature tuning in the transition test
 * @return TRUE if this the case, and FALSE otherwise
 */
bool TransitionExpansion::isCostValid(localPathPtr_t path, bool temperatureTuning)
{
	double pathCost(0.);
	double reverseCost(0.);
	confPtr_t previousConf = path->getBegin();

	double pathLength = path->getParamMax();
	double stepSize = min(this->getDMax(), pathLength);
	double param = stepSize;
	bool valid(true);
    unsigned int n(1);
    double max_cost(0), averageCost(0);
    if(ENV.getInt(Env::costDeltaMethod) == AVERAGE_CONFIG_COST){
        max_cost = previousConf->cost();
    }

	// iterate over the path to check it and compute its cost
	while (valid && (param <= pathLength))
	{
		confPtr_t currentConf = path->configAtParam(param);

		// perform the transition test
		valid = this->transitionTest(previousConf, currentConf, temperatureTuning);
		if (valid)
		{
			// delta cost: choose between mechanical work and integral of cost
			double deltaCost(0.);
			double deltaReverseCost(0.);
			if (ENV.getInt(Env::costDeltaMethod) == MECHANICAL_WORK)
			{
				deltaCost = max(0., currentConf->cost() - previousConf->cost());
				deltaReverseCost = max(0., previousConf->cost() - currentConf->cost());
			}
			else if (ENV.getInt(Env::costDeltaMethod) == INTEGRAL)
			{
				deltaCost = stepSize * (previousConf->cost() + currentConf->cost()) / 2.;
				deltaReverseCost = deltaCost;
            }else if (ENV.getInt(Env::costDeltaMethod) == AVERAGE_CONFIG_COST)
            {
                double currentConfCost = currentConf->cost();
                deltaCost += currentConfCost;
                n++;
            }else if(ENV.getInt(Env::costDeltaMethod) == AVERAGE_AND_MAX)
            {
                double currentConfCost = currentConf->cost();
                deltaCost += currentConfCost;
                max_cost = std::max(currentConfCost,max_cost);
                n++;
            }

			// update the total cost of the path
			pathCost += deltaCost;
			reverseCost += deltaReverseCost;

			// go to the next portion
			previousConf = currentConf;
			if (param == pathLength)
				param = pathLength + 1.;
			else if (pathLength - param > stepSize + 1e-3)
				param += stepSize;
			else {
				stepSize = pathLength - param;
				param = pathLength;
			}
		}
	}
    if(ENV.getInt(Env::costDeltaMethod) == AVERAGE_CONFIG_COST){
        pathCost = averageCost = pathCost /n;
    }else if(ENV.getInt(Env::costDeltaMethod) == AVERAGE_AND_MAX){
        averageCost = pathCost /n;
        pathCost = (averageCost + max_cost)/2;
    }

	// set the cost of the path and the cost corresponding to its reverse path
	path->setCost(pathCost);
    path->setAverageCost(averageCost);
    path->setMaxCost(max_cost);
	path->setReverseCost(reverseCost);

	return valid;
}


/**
 * Transition test based on the cost difference between the two given configurations.
 * @param temperatureTuning: with or without temperature tuning
 * @return TRUE if the transition is accepted, and FALSE otherwise
 */
bool TransitionExpansion::transitionTest(confPtr_t fromConf, confPtr_t toConf, bool temperatureTuning)
{
	double fromCost = fromConf->cost();
	double toCost = toConf->cost();

	// failure, when cost is higher than maximal value
	if ((ENV.getDouble(Env::costMax) > 0.) && (toCost > ENV.getDouble(Env::costMax)))
		return false;

	// success, for down-hill or flat slopes
	if (toCost <= fromCost)
		return true;

	// Metropolis test
	if (exp((fromCost - toCost) / temperature) > 0.5)
	{
		// if success, decrease the temperature
		if (temperatureTuning)
		{
			double costRange = max(1., highestCost - lowestCost);
			temperature /= pow(2, (toCost - fromCost) / (0.1 * costRange));
			if (temperature < ENV.getDouble(Env::initialTemperature))
				temperature = ENV.getDouble(Env::initialTemperature);
		}
		return true;
	}
	else
	{
		// if failure, increase the temperature
		if (temperatureTuning)
			temperature *= pow(2, ENV.getDouble(Env::temperatureRate));
		return false;
	}
}

