//
// C++ Implementation: prm
//
// Description: 
//
//
// Author: Florian Pilardeau,B90,6349 <fpilarde@jolimont>, (C) 2009
//
// Copyright: See COPYING file that comes with this distribution
//
//

#include "Planner-pkg.h"

#include "API/Roadmap/graph.hpp"

#include "planner/planEnvironment.hpp"
#include "planner/PRM/PRM.hpp"

using namespace std;
using namespace API;


/**
 * Initialize the PRM.
 * @return the number of nodes added to the graph
 */
unsigned PRM::init()
{
	int ADDED = 0;
  
	Planner::init();
	ADDED += Planner::setInit(_Robot->getInitialPosition());
	ADDED += Planner::setGoal(_Robot->getGoTo());
	_Init = true;

	this->loadWaypoints();
  
	return ADDED;
}


/**
 * Check the stopping conditions of PRM.
 * @return TRUE if one of the conditions is satisfied, and FALSE otherwise
 */
bool PRM::checkStopConditions()
{
    if (PlanEnv->getBool(PlanParam::stopPlanner)) {
        cout << "Stop: planning canceled by user." << endl;
        return true;
    }

    if (PlanEnv->getBool(PlanParam::planWithTimeLimit) &&
            m_time >= PlanEnv->getDouble(PlanParam::timeLimitPlanning)) {
        cout << "Stop: the time limit has been reached." << endl;
        return true;
    }

    if (m_nbConscutiveFailures >= (unsigned) ENV.getInt(Env::NbTry)) {
        cout << "Stop: the maximum number of consecutive failures has been reached." << endl;
        return true;
    }

    if (_Graph->getNumberOfNodes() >= (unsigned) ENV.getInt(Env::maxNodeCompco)) {
        cout << "Stop: the maximum number of nodes has been reached." << endl;
        return true;
    }

	if (ENV.getBool(Env::expandToGoal) && !PlanEnv->getBool(PlanParam::anytime))
		return this->trajFound();

	return false;
}


/**
 * Check the pre-conditions of PRM.
 * @return TRUE if one of the conditions is satisfied, and FALSE otherwise
 */
bool PRM::preConditions()
{
	if (ENV.getBool(Env::expandToGoal) && _Start->getConfiguration()->equal(*_Goal->getConfiguration())) {
		cout << "graph creation failed: start and goal are the same" << endl;
		return false;
	}

	if (!ENV.getBool(Env::isCostSpace) && ENV.getBool(Env::expandToGoal))
	{
		localPathPtr_t direct(new LocalPath(_q_start, _q_goal));
		if (direct->isValid()) {
			_Graph->addEdges(_Start, _Goal, direct);
			cout << "Direct connection" << endl;
			return true;
		}
	}
	return true;
}


/**
 * Perform a single expansion step of PRM.
 */
void PRM::expandOneStep()
{
	confPtr_t q = _Robot->shoot();

	if (q->setConstraintsWithSideEffect() && (!q->isInCollision()))
	{
		m_nbConscutiveFailures = 0;
		++m_nbAddedNode;

		Node * N = new Node(_Graph, q);
		_Graph->addNode(N);
		_Graph->linkNode(N);
	}
	else
		++m_nbConscutiveFailures;
}


/**
 * Main function of the PRM.
 * @return the number of nodes added to the graph
 */
unsigned int PRM::run()
{
	double ts(0.);
	m_time = 0.;
	ChronoOn();

	m_nbAddedNode = 0;
	m_nbExpansions = 0;

	// check whether the pre-conditions are satisfied
	if (!this->preConditions()) {
		cout << "Failure: pre-conditions are not satisfied." << endl;
		return 0;
	}

	// variable used to keep track of the cost of the best trajectory found so far, in the anytime variant
	double bestTrajCost = numeric_limits<double>::max();

	// PRM loop: start by checking the stopping conditions
	while (!this->checkStopConditions())
	{
		++m_nbExpansions;

		this->expandOneStep();

		// if the expansion is successful
		if (m_nbConscutiveFailures == 0)
		{
			// anytime variant
			bool drawNewTraj = false;
			if (!ENV.getBool(Env::drawDisabled) && PlanEnv->getBool(PlanParam::anytime) && this->trajFound())
			{
				Trajectory * trajectory = this->extractTrajectory();
				if (trajectory)
				{
					double trajCost = trajectory->cost();
					if (trajCost < bestTrajCost)
					{
						// keep track of the best trajectory found so far
						bestTrajCost = trajCost;
						trajectory->replaceP3dTraj();
						drawNewTraj = ENV.getBool(Env::drawTraj);
						cout << "best trajectory cost = " << bestTrajCost << endl;
					}
					delete trajectory;
				}
			}

			// draw the scene
			if (!ENV.getBool(Env::drawDisabled) && (ENV.getBool(Env::drawExploration) || drawNewTraj))
				_draw_func();
		}

		ChronoTimes(&m_time, &ts);
	}

	ChronoOff();
	return m_nbAddedNode;
}


/**
 * Check whether PRM has found a trajectory.
 * @return TRUE if the criteria required for a trajectory to be found are met, and FALSE otherwise
 */
bool PRM::trajFound()
{
    // success if the graph contains a single connected component
    if (_Graph->getNumberOfCompco() == 1)
    	return true;

    // success if all the waypoints belong to the same connected component
    unsigned i(1);
    bool connected(true);
    ConnectedComponent * component = waypointNodes[0]->getConnectedComponent();
    while (connected && i < waypointNodes.size())
    	connected = (component == waypointNodes[i++]->getConnectedComponent());
    return connected;
}


/**
 * Extract a trajectory going through all the waypoints.
 */
Trajectory * PRM::extractTrajectory()
{
	// simple case: only two waypoints (start and goal)
    if (waypoints.size() == 2)
        return _Graph->extractAStarShortestPath(_q_start, _q_goal);

    // general case: multiple waypoints
    else if (waypoints.size() > 2)
    {
    	// compute the pairwise distance matrix of the waypoints
    	double distances[waypoints.size()][waypoints.size()];
    	for (unsigned i(0); i < waypoints.size(); ++i)
    	{
    		distances[i][i] = 0.;
    		for (unsigned j(i + 1); j < waypoints.size(); ++j)
    		{
    			Trajectory * trajectory = _Graph->extractAStarShortestPath(waypoints[i], waypoints[j]);
    			if (trajectory)
    			{
					distances[i][j] = trajectory->cost();
					delete trajectory;
    			}
				if (ENV.getBool(Env::isCostSpace) &&
						(ENV.getInt(Env::costDeltaMethod) == MECHANICAL_WORK))
				{
					trajectory = _Graph->extractAStarShortestPath(waypoints[j], waypoints[i]);
					if (trajectory)
					{
						distances[j][i] = trajectory->cost();
						delete trajectory;
					}
				}
				else
					distances[j][i] = distances[i][j];
    		}
    	}

    	// create a container for this distance matrix
    	vector<vector<double> > distanceMatrix;
    	for (unsigned i(0); i < waypoints.size(); ++i)
    	{
    		vector<double> distanceRow;
    		for (unsigned j(0); j < waypoints.size(); ++j)
    			distanceRow.push_back(distances[i][j]);
    		distanceMatrix.push_back(distanceRow);
    	}

    	// solve the TPS problem to find the shortest path
		vector<unsigned> shortestPath;
		this->TPSbyExhaustiveSearch(distanceMatrix, shortestPath);

    	// extract a trajectory going through all the waypoints, following the shortest path
		vector<unsigned>::iterator start = shortestPath.begin();
    	vector<unsigned>::iterator goal = start;
    	++goal;
    	Trajectory * trajectory = _Graph->extractAStarShortestPath(waypoints[*start], waypoints[*goal]);
		for (unsigned i(2); i < waypoints.size(); ++i)
		{
			start = goal++;
			Trajectory * traj = _Graph->extractAStarShortestPath(waypoints[*start], waypoints[*goal]);
			if (!trajectory->concat(*traj))
				return NULL;
    	}

        return trajectory;
    }

    // error: less than two waypoints are present
    else
        return NULL;
}


/**
 * Solve the Traveling Salesman Problem by exhaustive search.
 * @param distances: pairwise distance matrix given as input
 * @param solution: shortest path returned as output
 * @return length of the shortest path
 */
double PRM::TPSbyExhaustiveSearch(vector<vector<double> > & distances, vector<unsigned> & solution)
{
	// initialization
	unsigned N = distances.size();
	unsigned permutation[N];
	for (unsigned i(0); i < N; ++i) {
		permutation[i] = i;
		solution.push_back(i);
	}
	double minLength = 0.;
	for (unsigned i(1); i < N; ++i)
		minLength += distances[i - 1][i];

	// go through each permutation and check whether it corresponds to a shorter path
	while (next_permutation(permutation, permutation + N))
	{
		double length = 0.;
		for (unsigned i(1); i < N; ++i)
			length += distances[permutation[i - 1]][permutation[i]];

		if (length < minLength) {
			minLength = length;
			for (unsigned i(0); i < N; ++i)
				solution[i] = permutation[i];
		}
	}
	return minLength;
}


/**
 * Load the user-defined waypoints (start, goal, and potential intermediate waypoints).
 */
void PRM::loadWaypoints()
{
    // initialize the lists of waypoints
	waypoints.push_back(_q_start);
	waypointNodes.push_back(_Start);

    // load the additional user-defined configurations as waypoints (when relevant)
    for (unsigned i(0); i < (unsigned) _Robot->getRobotStruct()->nconf; ++i)
    {
        confPtr_t conf(new Configuration(_Robot, _Robot->getRobotStruct()->conf[i]->q));
        Node * node = new Node(_Graph, conf);
        _Graph->addNode(node);

        waypoints.push_back(conf);
        waypointNodes.push_back(node);
    }

    // finalize the lists of waypoints
    waypoints.push_back(_q_goal);
    waypointNodes.push_back(_Goal);
}
