/*
 *  HandoverReplanner.cpp
 *  libmove3D-planners 
 *
 *  Created by Benjamin Vadant on 01/02/13.
 *  Based on Jim Mainprice replanningAlgorithms.hpp/cpp
 *  Copyright 2013 LAAS/CNRS. All rights reserved.
 *
 */

#include "HandoverReplanner.hpp"
#include "ReplanningSimulator.hpp"

#include "API/project.hpp"

#include "planner/planEnvironment.hpp"
#include "planner/plannerFunctions.hpp"
#include "planner/TrajectoryOptim/Stomp/stompOptimizer.hpp"
#include "planner/TrajectoryOptim/trajectoryOptim.hpp"
#include "planner/TrajectoryOptim/Classic/costOptimization.hpp"
#include "planner/cost_space.hpp"
#include "planner/Diffusion/Variants/RRT-Star.hpp"

#include "HRICS_costspace.hpp"

#include "LightPlanner-pkg.h"
#include "Graphic-pkg.h"
#include "Collision-pkg.h"

#include "move3d-headless.h"

#include <boost/thread/thread.hpp>
#include <sys/time.h>

using namespace std;
using namespace tr1;

HandoverReplanner::HandoverReplanner(Robot* r) : SimpleReplanner(r)
{
    
}

HandoverReplanner::~HandoverReplanner()
{
    
}

bool HandoverReplanner::init()
{
    SimpleReplanner::init();
    
    goalFromList = true;
    human = global_Project->getActiveScene()->getRobotByName("HERAKLES_HUMAN1");//TODO: change this to a general ver
    
    const char* home = getenv("HOME_MOVE3D");
    if( home == NULL ) {
        cout << "ERROR home is not defined for config generator in " << __func__ << endl;
    }
    
    string dir(home);
    string file("/statFiles/OtpComputing/confHerakles.xml");
    
    goalFromList = ( ENV.getInt(Env::setOfActiveJoints) == 2 );
    
    handoverGenerator = new ConfGenerator( robot, human );
    
    if(  HRICS_activeNatu == NULL )                                         //HRICS_costspace
    {
        return false;
    }
    else 
    {
        handoverGenerator->initialize( dir+file, HRICS_activeNatu );
    }
    
    return true;
}

std::pair<bool,confPtr_t> HandoverReplanner::newGoalFromList()
{
    double bestCost=0.0;
    pair<confPtr_t,confPtr_t> bestHandoverConf;
    std::pair<bool,confPtr_t> result;
    
    // Parse list to find
    // the best feasible hand-over configuration
    result.first = handoverGenerator->computeHandoverConfigFromList( bestHandoverConf, bestCost ); 
    
    if( result.first ) {
        result.second = bestHandoverConf.second;
        p3d_update_virtual_object_config_for_arm_ik_constraint(robot->getRobotStruct(), 0, bestHandoverConf.second->getConfigStruct() );
    }
    
    return result;
}

std::pair<bool,confPtr_t> HandoverReplanner::newGoalFromIK()
{  
    //Time of Execution
    timeval tim;
    gettimeofday(&tim, NULL);
    double tInit = tim.tv_sec+(tim.tv_usec/1000000.0);
    double timeExecution = 0.0;
    //Time of Execution
    
    configPt q = NULL; 

    Eigen::Vector3d point = human->getJoint("rPalm")->getVectorPos();
    point[2] += 0.10;
    
    // if the generator finds a configuration, return it otherwise return null configuration
    if( handoverGenerator->computeRobotIkForGrabing( q, point ) )
    {
        //Time of Execution
        gettimeofday(&tim, NULL); 
        timeExecution = tim.tv_sec+(tim.tv_usec/1000000.0) - tInit;
        cout << "Valid robotIk computed in : " << timeExecution << " sec" << endl;
        //Time of Execution  
        
        confPtr_t q_rob( new Configuration( robot, q ) );
        return std::make_pair( true, q_rob );
    }
    else
    {
        //Time of Execution
        gettimeofday(&tim, NULL); 
        timeExecution = tim.tv_sec+(tim.tv_usec/1000000.0) - tInit;
        cout << "Could not compute a valid robot IK configuration in : " << timeExecution << " sec" << endl;
        //Time of Execution   
        
        p3d_destroy_config(robot->getRobotStruct(), q);
        confPtr_t q_rob( new Configuration( robot, NULL ) );
        return std::make_pair( false, q_rob );
    }
}

void HandoverReplanner::run()
{
    isPlanning = true;
    cout << "HandoverReplanner :: Start replanning step : " << __func__ << endl;
    
    //Time of Execution
    timeval tim;
    gettimeofday(&tim, NULL);
    double tInit = tim.tv_sec+(tim.tv_usec/1000000.0);
    double timeExecution = 0.0;
    //Time of Execution
    
    planningSucceeded = false;
    
    confPtr_t initConfig = currentTrajectory.configAtParam( positionSwitch );
    confPtr_t goalConfig = currentTrajectory.getEnd();
    
    idRun++;
    
    std::pair<bool,confPtr_t> goal;
    
    confPtr_t human_conf_tmp = human->getCurrentPos();          //v non sens
    
    goal = (goalFromList)?newGoalFromList():newGoalFromIK();
    
    human->setAndUpdate( *human_conf_tmp );                     //^ non sens
    
    if( goal.first ) {
        goalConfig = goal.second;
        cout << "New goal found!!!" << endl;
    }
    
    PlanEnv->setBool( PlanParam::planWithTimeLimit, true );
    PlanEnv->setDouble(PlanParam::timeLimitPlanning, timeLimitRePlanning-0.3 );// Set 0.3 seconds to do the concatenated traj
    
    p3d_traj* path = p3d_planner_function( robot->getRobotStruct(), 
                                           initConfig->getConfigStruct(), 
                                           goalConfig->getConfigStruct() );
  
    //  Set the time left for smoothing
    //Time of Execution
    gettimeofday(&tim, NULL); 
    timeExecution = tim.tv_sec+(tim.tv_usec/1000000.0) - tInit;
    //Time of Execution
    
    if( path != NULL && 
            !PlanEnv->getBool(PlanParam::stopPlanner) && 
            PlanEnv->getBool(PlanParam::withSmoothing) )
    {
        p3d_smoothing_function( robot->getRobotStruct(), path, 100, timeLimitRePlanning-0.3-timeExecution);
    }
    
    if( path )
    {
        // Cut the new trajectory in lp average piecese
        API::Trajectory newPortion( robot, path );
        cout << "newPortion.getNbOfPaths() : " << newPortion.getNbOfPaths() << endl;
        cout << "newPortion.getRangeMax() : " << newPortion.getRangeMax() << endl;
        
        std::pair<bool,API::Trajectory> concatenatedTrajectory = concatenateToCurrentTrajectory( newPortion );
        planningSucceeded = concatenatedTrajectory.first;
        
        globalRePlanningEnv->storeTrajectoryToDraw( concatenatedTrajectory.second, 0 );                                                          //simulator
              
        if( planningSucceeded ) {
            currentTrajectory = concatenatedTrajectory.second;
        }
        else {
            cout << "Could not concatenate trajectory" << endl;
        }
    }
    
    if( !planningSucceeded ) {
        cout << "HandoverReplanner :: Replanning failed" << endl;
    }
    
    cout << "HandoverReplanner :: End replanning step : " << __func__ << endl;
    isPlanning = false;
}
