/*
 *  ReplanningSimulator.cpp
 *  libmove3D-planners 
 *
 *  Created by Benjamin Vadant on 01/02/13.
 *  Based on Jim Mainprice replanningAlgorithms.hpp/cpp
 *  Copyright 2013 LAAS/CNRS. All rights reserved.
 *
 */
#include <iostream>

#include "ReplanningSimulator.hpp"
#include "AStarReplanner.hpp"
#include "HandoverReplanner.hpp"
#include "Replanner.hpp"
#include "RRTReplanner.hpp"
#include "SimpleReplanner.hpp"
#include "SoftmotionReplanner.hpp"
#include "StompReplanner.hpp"


#include "API/project.hpp"

#include "planner/planEnvironment.hpp"
#include "planner/plannerFunctions.hpp"
#include "planner/TrajectoryOptim/Classic/costOptimization.hpp"
#include "planner/cost_space.hpp"

#include "LightPlanner-pkg.h"
#include "Graphic-pkg.h"
#include "Collision-pkg.h"

#include "move3d-headless.h"

#include <boost/thread/thread.hpp>
#include <sys/time.h>

using namespace std;
using namespace tr1;

ReplanningSimulator* globalRePlanningEnv=NULL;
Graph* globalRePlanningGraph=NULL;

//! Soft motion trajectory for planner
SM_TRAJ globalManipPlannerLastTraj;

// Replanning drawing function
void replan_current_draw()
{
    if( globalRePlanningEnv )
        globalRePlanningEnv->draw();
}


ReplanningSimulator::ReplanningSimulator()
{
    //! Local variables
    initDone=false;
    
    drawFinalConfig = true;
    
    robot=NULL;
    robotSimulated=NULL;
    human=NULL;
       
    globalRePlanningEnv = this;
}

ReplanningSimulator::~ReplanningSimulator()
{
    globalRePlanningEnv = NULL;
}

//! Finds the robot in the P3D file
//! that also have a SIMUL robot
bool ReplanningSimulator::initSimulator()
{  
    string robotBaseName;
    if( !initFindRobotBasename( robotBaseName ) )
    {
        cout << "Error : p3d file not supported for replanning in " << __FILE__ << " at " << __LINE__ << endl;
        return false;
    }
    
    if( !initRobotSimulatedCntrtsAndCollisions() )
    {
        cout << "Error : p3d file not supported for replanning in " << __FILE__ << " at " << __LINE__ << endl;
        return false;
    }
    
    switch ( PlanEnv->getInt( PlanParam::replanningAlgorithm ) )
    {
    case 0:
        replanner = new SimpleReplanner( robot ); break;
    case 1:
        replanner = new SoftmotionReplanner( robot ); break;
    case 2:
        replanner = new RRTReplanner( robot ); break;
    case 3:
        replanner = new HandoverReplanner( robot ); break;
    case 4:
        replanner = new AStarReplanner( robot ); break;
    case 5:
        replanner = new StompReplanner( robot ); break;
    default :
        cout << "No replanner selected defaulting to SimpleReplanner" << endl;
        replanner = new SimpleReplanner( robot ); break;
    }
    
    // Create one line traj from current traj for robot simul
    executedTraj = API::Trajectory( robotSimulated );
    executedTraj.clear();
    
    // Set the simulator as initialized and put set the drawing variables
    initDone = true;
    isWritingDisplay=false;
    isReadingDisplay=false;
    return true;
}

//! Finds the robot in the P3D file
//! that also have a SIMUL robot
bool ReplanningSimulator::initFindRobotBasename( string& robotBaseName )
{
    Scene* sce = global_Project->getActiveScene();
    
    Robot* rob = sce->getRobotByNameContaining( "_ROBOT" );
    if( rob == NULL ) {
        return false; 
    }
    
    size_t found = rob->getName().find("_ROBOT");
    if (found!=string::npos)
    {
        robotBaseName = rob->getName().substr(0,found);
        cout << "Robot base name is : " << robotBaseName << endl;
    }
    else {
        return false;
    }
    
    if ( sce->getRobotByName( robotBaseName + string("_SIMUL") ) == NULL ) {
        return false;
    }
    
    human = global_Project->getActiveScene()->getRobotByNameContaining("HERAKLES");
    if( human == NULL ) {
        cout << "Error : no human found in " << __FILE__ << " at " << __LINE__ << endl;
        return false;
    }
    
    robot = sce->getRobotByName( robotBaseName + string("_ROBOT") );
    robotSimulated = sce->getRobotByName( robotBaseName + string("_SIMUL") );
    robotCylinder = sce->getRobotByNameContaining( string("CYLINDER") );
    if( robot == NULL || robotSimulated == NULL )
    {
        return false;
    }
    
    return true;
}

//! initialize the simulation robot
bool ReplanningSimulator::initRobotSimulatedCntrtsAndCollisions()
{
    if( robot == NULL || robotSimulated == NULL )
    {
        cout << "Error : the robots are not initialized in "  << __FILE__ << " at " << __LINE__ << endl;
        return false;
    }
    
    // Deactivate robot to simulation robot collision checking
    p3d_col_deactivate_rob_rob( robot->getRobotStruct(), robotSimulated->getRobotStruct() );
    
    if( robotCylinder ) {
        p3d_col_deactivate_rob_rob( robotCylinder->getRobotStruct(), robot->getRobotStruct() );
        p3d_col_deactivate_rob_rob( robotCylinder->getRobotStruct(), robotSimulated->getRobotStruct() );
    }
    
    // Deactivate all kinematic constraints
    // p3d_deactivate_all_cntrts( robotSimulated );
    
    // Fix all the joints but unfixes the arm 0
    //if( robot->getName() == "PR2_ROBOT" || robot->getName() == "BOB_ROBOT" || robot->getName() == "MAX_ROBOT")
    if( robot->getHriAgent() && robot->getHriAgent()->type == HRI_PR2)
        fixAllJointsWithoutArm( robotSimulated->getRobotStruct(), 0 );
    
    return true;
}

//-----------------------------------------------------------------------------------
//-----------------------------------------------------------------------------------
//-----------------------------------------------------------------------------------
//-----------------------------------------------------------------------------------
//-----------------------------------------------------------------------------------
void ReplanningSimulator::setMultithreadGraphical(bool enable)
{
    if( robot == NULL )
    {
        cout << "Replanner not initiailzed" << endl;
        return;
    }
    
    cout << "set graphic mode to multi thread" << endl;
    g3d_set_multi_thread_mode( enable );
    
    if (enable) 
    {
        robot->getRobotStruct()->display_mode = P3D_ROB_NO_DISPLAY;
        ext_g3d_draw_multi_thread = replan_current_draw;
    }
    else
    {
        robot->getRobotStruct()->display_mode = P3D_ROB_DEFAULT_DISPLAY;
        ext_g3d_draw_multi_thread = NULL;
    }
}

//! Stores the trajectory in a vector
void ReplanningSimulator::storeTrajectoryToVect(SM_TRAJ& smTraj, double current, double step)
{
    lastLine = currentLine;
    //lastLine.clear();
    currentLine.clear();
    
    for( double t=current; t<smTraj.getDuration(); t += step)
    {
        std::vector<SM_COND> cond;
        smTraj.getMotionCond(t,cond);
        
        vector<double> point(3);
        point[0] = cond[0].x;
        point[1] = cond[1].x;
        point[2] = 0.20;
        
        currentLine.push_back( point );
    }
}

void print_point( const vector<double>& p )
{
    cout << "p = ( "<<p[0]<<" , "<<p[1]<<" , "<<p[2]<<" )"<< endl;
}

void ReplanningSimulator::storeHumanPosition()
{
    confPtr_t qHuman = human->getCurrentPos();
    (*qHuman)[6] =  PlanEnv->getDouble(PlanParam::env_futurX);
    (*qHuman)[7] =  PlanEnv->getDouble(PlanParam::env_futurY);
    (*qHuman)[8] =  PlanEnv->getDouble(PlanParam::env_futurZ);
    (*qHuman)[11] = PlanEnv->getDouble(PlanParam::env_futurRZ);
    (*qHuman)[19] = PlanEnv->getDouble(PlanParam::env_futurArmZ);
    
    human->setAndUpdate(*qHuman);
    //  cout << "PlanEnv->getDouble(PlanParam::env_futurX) = " << PlanEnv->getDouble(PlanParam::env_futurX) << endl;
    //  cout << "PlanEnv->getDouble(PlanParam::env_futurY) = " << PlanEnv->getDouble(PlanParam::env_futurY) << endl;
    //  cout << "PlanEnv->getDouble(PlanParam::env_futurZ) = " << PlanEnv->getDouble(PlanParam::env_futurZ) << endl;
    //  cout << "PlanEnv->getDouble(PlanParam::env_futurRZ) = " << PlanEnv->getDouble(PlanParam::env_futurRZ) <<  endl;
    
    //  cout << "(*qHuman)[6] = " << (*qHuman)[6] << endl;
    //  cout << "(*qHuman)[7] = " << (*qHuman)[7] << endl;
    //  cout << "(*qHuman)[8] = " << (*qHuman)[8] << endl;
    //  cout << "(*qHuman)[11] = " << (*qHuman)[11] <<  endl;
}

//! Stores the trajectory in a vector
void ReplanningSimulator::storeTrajectoryToVect(API::Trajectory& traj, double step)
{
    if( !initDone )
        return;
    
    vector<double> point(3);
    //lastLine = currentLine;
    lastLine.clear();
    currentLine.clear();
    
    Configuration q(*traj.getBegin());
    point[0] = q[6];
    point[1] = q[7];
    point[2] = 0.20;
    
    // print_point( point );
    currentLine.push_back( point );
    
    step = traj.getRangeMax()/20;
    
    for( double t=step; t<traj.getRangeMax(); t += step)
    {
        q = *traj.configAtParam(t);
        point[0] = q[6];
        point[1] = q[7];
        point[2] = 0.20;
        
        // print_point( point );
        currentLine.push_back( point );
    }
    
    q = *traj.getEnd();
    point[0] = q[6];
    point[1] = q[7];
    point[2] = 0.20;
    
    // print_point( point );
    currentLine.push_back( point );
}

//! Stores the trajectory to a current trajectory
void ReplanningSimulator::storeTrajectoryToDraw(const API::Trajectory& traj, double step)
{
    if( !initDone )
        return;
    
    isWritingDisplay = true;
    
    while (isReadingDisplay);
    
    vector<double> point(3);
    //lastLine = currentLine;
    lastLine.clear();
    currentLine.clear();
    
    Joint* drawJoint = robot->getJoint( p3d_get_user_drawnjnt() );
    
    //  cout << "user_drawnjnt : " << p3d_get_user_drawnjnt() << endl;
    //  cout << "drawJoint : " << drawJoint << endl;
    
    if( drawJoint != NULL )
    {
        //    for( int i=0; i<traj.getNbOfViaPoints(); i++ )
        //    {
        //      robot->setAndUpdate(*traj[i]);
        //      
        //      Eigen::Vector3d vect = drawJoint->getVectorPos();
        //      point[0] = vect[0];
        //      point[1] = vect[1];
        //      point[2] = vect[2];
        //      
        //      currentLine.push_back( point );
        //    }
        
        double step = drawStep;
        //double step = traj.getRangeMax()/20;
        double range_max = traj.getRangeMax();
        
        for( double t=0.0; t<(range_max+step); t += step)
        {
            robot->setAndUpdate(*traj.configAtParam(t));
            Eigen::Vector3d vect = drawJoint->getVectorPos();
            point[0] = vect[0];
            point[1] = vect[1];
            point[2] = vect[2];
            
            // print_point( point );
            currentLine.push_back( point );
        }
    }
    
    // Store also human position
    storeHumanPosition();
    
    isWritingDisplay = false;
}

//! Stores the graph to a local graph
void ReplanningSimulator::storeGraphToDraw(const Graph& graph)
{
    isWritingDisplay = true;
    
    while (isReadingDisplay);
    
    cout << "Copy graph" << endl;
    delete globalRePlanningGraph;
    globalRePlanningGraph = new Graph(graph);
    
    // Store also human position
    storeHumanPosition();
    
    isWritingDisplay = false;
}

//! Stores the trajectory to a current trajectory
void ReplanningSimulator::storeExploration(const API::Trajectory& traj, double lPrev, double lNext, shared_ptr<
                                            Configuration> qNew)
{
    if( !initDone )
        return;
    
    isWritingDisplay = true;
    
    while (isReadingDisplay);
    
    //  cout << "ReplanningSimulator::storeExploration" << endl;
    
    deviateLine.clear();
    
    vector< shared_ptr<Configuration> > vectConf(3);
    vectConf.at(0) = traj.configAtParam(lPrev);
    vectConf.at(1) = qNew;
    vectConf.at(2) = traj.configAtParam(lNext);
    
    Joint* drawJoint = robot->getJoint( p3d_get_user_drawnjnt() );
    
    if( drawJoint != NULL )
    {
        vector<double> point(3);
        
        for( int i=0; i<int(vectConf.size()); i++ )
        {
            robot->setAndUpdate(*vectConf[i]);
            
            Eigen::Vector3d vect = drawJoint->getVectorPos();
            point[0] = vect[0];
            point[1] = vect[1];
            point[2] = vect[2];
            
            deviateLine.push_back( point );
        }
    }
    
    // Store also human position
    storeHumanPosition();
    
    isWritingDisplay = false;
}

//! Set current traj as executed
bool ReplanningSimulator::setExecutedTrajectoryToCurrent(API::Trajectory& traj)
{
    executedTraj.clear();
    
    for( int i=0; i<traj.getNbOfViaPoints(); i++ ) {
        confPtr_t q(new Configuration( robotSimulated, traj[i]->getConfigStruct()));
        executedTraj.push_back( q );
    }
    
    /*if( traj.getNbOfPaths() != executedTraj.getNbOfPaths() ) {
        cout << "Error: The number of path differs" << endl;
        cout << "traj.getNbOfPaths() : " << traj.getNbOfPaths() << endl;
        cout << "executedTraj.getNbOfPaths() : " << executedTraj.getNbOfPaths() << endl;
        return false;
    }
    
    for( int i=0; i<traj.getNbOfViaPoints(); i++ ) {
        if( *traj[i] != *executedTraj[i] ) {
            cout << "Error: config differ" << endl;
            traj[i]->print();
            executedTraj[i]->print();
            return false;
        }
    }
    
    for( int i=0; i<executedTraj.getNbOfPaths(); i++ )
    {
        double para1 = traj.getLocalPathPtrAt(i)->getParamMax();
        double para2 = executedTraj.getLocalPathPtrAt(i)->getParamMax();
        
        if( fabs( para1 - para2 ) > EPS6 ) {
            
            cout << "param 1 : " << para1 << endl;
            cout << "param 2 : " << para2 << endl;
            
            confPtr_t q1_1 = traj.getLocalPathPtrAt(i)->getBegin();
            confPtr_t q2_1 = traj.getLocalPathPtrAt(i)->getEnd();
            cout << "Current traj. conf" << endl;
            q1_1->equal( *q2_1, true );
            
            confPtr_t q1_2 = executedTraj.getLocalPathPtrAt(i)->getBegin();
            confPtr_t q2_2 = executedTraj.getLocalPathPtrAt(i)->getEnd();
            cout << "Executed traj. conf" << endl;
            q1_2->equal( *q2_2, true );
            cout << "Error : The localpath " << i << " are not the same length" << endl;
            
            LocalPath path3(q1_1,q2_1);
            LocalPath path4(q1_2,q2_2);
            
            if( *q1_1 != *q1_2 ) {
                q1_1->print();
                q1_2->print();
            }
            if( *q2_1 != *q2_2 ) {
                q2_1->print();
                q2_2->print();
            }
            cout << "param 3 : " << path3.getParamMax() << endl;
            cout << "param 4 : " << path4.getParamMax() << endl;
            
            cout << "q1_1->dist( *q2_1, true ) : " << q1_1->dist( *q2_1, true ) << endl;
            cout << "q1_2->dist( *q2_2, true ) : " << q1_2->dist( *q2_2, true ) << endl;
            
            return false;
        }
    }*/
    
    qEnd = executedTraj.getEnd();
    return true;
}

//! Draws replanning features
void ReplanningSimulator::draw()
{
    while (isWritingDisplay);
    
    isReadingDisplay = true;
    
    if(currentLine.size()>=2)
    {
        double color[4];
        color[0] = 1.0;
        color[1] = 0.8;
        color[2] = 0.2;
        color[3] = 1.0;
        
        for( int i=0; i<int(currentLine.size()-1); i++)
        {
            //      cout <<"currentLine["<<i<<"][0]="<<currentLine[i][0]<<endl;
            //      cout <<"currentLine["<<i<<"][1]="<<currentLine[i][1]<<endl;
            //      cout <<"currentLine["<<i<<"][2]="<<currentLine[i][1]<<endl;
            g3d_set_color(Any,color);
            //g3d_drawSphere(currentLine[i+0][0],currentLine[i+0][1],currentLine[i+0][2],0.01);
            g3d_draw_solid_sphere(currentLine[i+0][0],currentLine[i+0][1],currentLine[i+0][2],0.02,10);
            
            g3d_drawOneLine(currentLine[i+0][0],currentLine[i+0][1],currentLine[i+0][2],
                            currentLine[i+1][0],currentLine[i+1][1],currentLine[i+1][2],Any,color);
            
            if( i == int(currentLine.size()-2))
            {
                //g3d_drawSphere(currentLine[i+1][0],currentLine[i+1][1],currentLine[i+1][2],0.01);
                g3d_draw_solid_sphere(currentLine[i+1][0],currentLine[i+1][1],currentLine[i+1][2],0.01,10);
            }
        }
    }
    
    if(lastLine.size()>=2)
    {
        double color[4];
        color[0] = 0.8;
        color[1] = 1.0;
        color[2] = 0.2;
        color[3] = 1.0;
        
        for( int i=0; i<int(lastLine.size()-1); i++)
        {
            g3d_set_color(Any,color);
            //g3d_drawSphere(lastLine[i+0][0],lastLine[i+0][1],lastLine[i+0][2],0.01);
            g3d_draw_solid_sphere(lastLine[i+0][0],lastLine[i+0][1],lastLine[i+0][2],0.02,10);
            
            g3d_drawOneLine(lastLine[i+0][0],lastLine[i+0][1],lastLine[i+0][2],
                            lastLine[i+1][0],lastLine[i+1][1],lastLine[i+1][2],Any,color);
            
            if( i == int(lastLine.size()-2))
            {
                //g3d_drawSphere(lastLine[i+1][0],lastLine[i+1][1],lastLine[i+1][2],0.01);
                g3d_draw_solid_sphere(lastLine[i+1][0],lastLine[i+1][1],lastLine[i+1][2],0.01,10);
            }
        }
    }
    
    if(deviateLine.size()>=2 && PlanEnv->getBool(PlanParam::showExploration))
    {
        double color[4];
        color[0] = 1.0;
        color[1] = 0.0;
        color[2] = 0.0;
        color[3] = 1.0;
        
        for( int i=0; i<int(deviateLine.size()-1); i++)
        {
            g3d_set_color(Any,color);
            
            g3d_draw_solid_sphere(deviateLine[i+0][0], deviateLine[i+0][1], deviateLine[i+0][2],0.02,10);
            
            g3d_drawOneLine(deviateLine[i+0][0], deviateLine[i+0][1], deviateLine[i+0][2],
                            deviateLine[i+1][0], deviateLine[i+1][1], deviateLine[i+1][2],Any,color);
            
            if( i == int(deviateLine.size()-2))
            {
                g3d_draw_solid_sphere(deviateLine[i+1][0],deviateLine[i+1][1],deviateLine[i+1][2],0.02,10);
            }
        }
    }
    
    if (ENV.getBool(Env::drawGraph) && globalRePlanningGraph) 
    {
        globalRePlanningGraph->setRobot( robotSimulated );
        globalRePlanningGraph->draw();
    }
    
    if (ext_g3d_traj_debug) {
        ext_g3d_traj_debug();
    }
    
    //  if( drawFinalConfig ) {
    //    confPtr_t q_tmp = robotSimulated->getCurrentPos();
    //    G3D_Window* win = g3d_get_cur_win();
    //    win->vs.transparency_mode= G3D_TRANSPARENT_AND_OPAQUE;
    //    
    //    robotSimulated->setAndUpdate( *q_end );
    //    robotSimulated->getRobotStruct()->draw_transparent = false;
    //    g3d_draw_robot( robotSimulated->getRobotStruct()->num, win, 0);
    //    robotSimulated->getRobotStruct()->draw_transparent = false;
    //    robotSimulated->setAndUpdate( *q_tmp );
    //  }
    
    isReadingDisplay = false;
}

//-----------------------------------------------------------------------------------
//-----------------------------------------------------------------------------------
//-----------------------------------------------------------------------------------
//-----------------------------------------------------------------------------------
//-----------------------------------------------------------------------------------

//! first call stores the time of day in t_init
//! the next calls will returns the substraction
//! between timofday and the argument t_init
double ReplanningSimulator::timeSinceLastCall(bool& is_first_call, double& t_init)
{
    timeval tim;
    gettimeofday(&tim, NULL);
    double tu=tim.tv_sec+(tim.tv_usec/1000000.0);
    
    if (is_first_call) {
        t_init = tu;
        is_first_call = false;
    }
    
    double dt = tu - t_init;
    t_init = tu;
    return dt;
}

bool ReplanningSimulator::timeSwitchAndId(double s, double s_rep, int& id_switch, 
                                             API::Trajectory& traj, double &s_switch)
{
    double p = 0.0;
    
    for (int i=0; i<traj.getNbOfPaths(); i++)
    {
        p += traj.getLocalPathPtrAt(i)->getParamMax();
        
        if( p > (s+s_rep) )
        {
            switchId = i;
            s_switch = p;
            return true;
        }
    }
    
    return false;
}
// ---------------------------------------------------------------------------------
// Execute simulated replanning phase
// ---------------------------------------------------------------------------------
int ReplanningSimulator::executeSoftmotionSimulation( int (*fct)(p3d_rob* robot, p3d_localpath* localpathPt) ) 
{
    string robotname;
    initFindRobotBasename( robotname );
    
    if (!initSimulator()) {
        cout << "initSimulator did not work!!!" << endl;
        return false;
    }
    
    if (!initDone)
    {
        //    if(dynamic_cast<SoftmotionReplanner*>(replanner)->init_manipulationPlanner())
        //    {
        //      init = true;
        //    }
    }
    else 
    {
        cout << "ReplanningSimulator::init_execution did not work!!!" << endl;
        return false;
    }
    
    if( global_costSpace == NULL )
    {
        cout << "Error : Cost Space not initialized" << endl;
        return false;
    }
    
    p3d_rob* robotPt;
    p3d_localpath* localpathPt;
    SM_TRAJ smTraj;
    
    bool RunShowTraj=true;
    
    Configuration q(robotSimulated);
    
    ChronoOn();
    
    // Set global variables
    confPtr_t qSwitch =   replanner->getQSwitch();
    confPtr_t qGoal =     replanner->getQGoal();
    API::Trajectory& CurrentTraj = replanner->getCurrentTrajectory();
    
    lastLine.clear();
    currentLine.clear();
    setMultithreadGraphical( true );
    
    //  if( !replanner_manip() )
    //  {
    //    cout << "Failed to find an inital plan" << endl;
    //    return false;
    //  }
    
    // Initial values
    smTraj = globalManipPlannerLastTraj;
    
    // Simulator variables
    bool isFirstLoop=true;
    bool StopRun=false;
    int count=0;
    bool planningSucceded = false;
    bool is_switch_done = true;
    double tu_tmp = 0.0; //, ts = 0.0;
    double t_max = smTraj.getDuration();
    double t = 0.0;
    double t_init = 0.0;
    double lastFloor=0.0; /* parameters along the local path */
    double t_rep = 5.0;
    int initial_step =0;
    double lp_avera_length = 0.0;
    
    // Switch
    switchId = 0;
    
    // Store trajectory for drawing
    storeTrajectoryToVect(smTraj, 0.0, 1.0);
    
    // Replanning values
    double t_switch= t_max;
    
    while ( !StopRun )
    { 
        // Switch to new trajectory
        if( (t>t_switch) && !replanner->isPlanningUndergo() && planningSucceded )
        {
            smTraj = globalManipPlannerLastTraj;
            t_init = 0.0;
            t_max = smTraj.getDuration();
            is_switch_done = true;
            
            // Store trajectory for drawing
            storeTrajectoryToVect( smTraj, 0, 1.0 );
        }
        
        // Launch new planning
        if( is_switch_done && !replanner->isPlanningUndergo() )
        {
            is_switch_done = false;
            StopRun = true;
            
            if ( planningSucceded ) 
            {
                p3d_getQSwitchIDFromMidCVS( t, t_rep, &switchId );
                p3d_getMidCVSTimeOnTraj( switchId, t_switch );
                
                cout << "q switch id : " << switchId << endl;
                cout << "CurrentTraj.geMaxParam() : " << CurrentTraj.getRangeMax() << endl;
                
                if ( switchId < (CurrentTraj.getNbOfViaPoints()-1) )
                {
                    qSwitch = CurrentTraj[switchId];
                    
                    // Set data for switch in replanner
                    replanner->setSwitchData( qSwitch, switchId, t_switch, t_rep, lp_avera_length, initial_step );
                    
                    // Launch replanning thread
                    boost::thread workerThread( boost::bind( &Replanner::run, replanner) );
                    StopRun = false;
                }
                else if ( switchId == (CurrentTraj.getNbOfViaPoints()-1) )
                {
                    cout << "Last replanning!!!" << endl;
                    StopRun = false;
                }
            }
        }
        
        // Show current trajectory
        std::vector<SM_COND> cond;
        smTraj.getMotionCond(t-t_init,cond);
        
        q = *robotSimulated->getCurrentPos();
        
        q[6] = cond[0].x;
        q[7] = cond[1].x;
        q[11] = cond[5].x;
        // torso
        q[12] = cond[6].x;
        //head
        q[13] = cond[7].x;
        q[14] = cond[8].x;
        //leftarm
        q[25] = cond[19].x;
        q[26] = cond[20].x;
        q[27] = cond[21].x;
        q[28] = cond[22].x;
        q[29] = cond[23].x;
        q[30] = cond[24].x;
        q[31] = cond[25].x;
        //rightarm
        q[16] = cond[10].x;
        q[17] = cond[11].x;
        q[18] = cond[12].x;
        q[19] = cond[13].x;
        q[20] = cond[14].x;
        q[21] = cond[15].x;
        q[22] = cond[16].x;
        
        // set and update robot from configuration
        robotSimulated->setAndUpdate(q);
        
        if (fct) if (((*fct)(robotPt, localpathPt)) == FALSE) return(count);
        count++;
        
        RunShowTraj = (*fct_stop)();
        
        t += timeSinceLastCall( isFirstLoop, tu_tmp );
        
        if (floor(t) > lastFloor) {
            cout << "t = " << t << endl;
            lastFloor = floor(t);
        }
        
        if (t > t_max - EPS6) {
            t = t_max;
            StopRun = true;
        }
        
        if (PlanEnv->getBool(PlanParam::stopPlanner)) {
            StopRun = true;
        }
        
        if (!planningSucceded) 
        {
            cout << "Planning Failed!!!" << endl;
            StopRun = true;
        }
    }
    
    cout << "Final time(sec) on traj : " << t_max << endl;
    ChronoPrint("");
    ChronoOff();
    PlanEnv->setBool(PlanParam::stopPlanner,true);
    return 0;
}

// ---------------------------------------------------------------------------------
// Execute simulated replanning phase
// ---------------------------------------------------------------------------------
int ReplanningSimulator::executeSimpleSimulation( int (*fct)(p3d_rob* robot, p3d_localpath* localpathPt) )
{  
    if(!initSimulator())
    {
        cout << "Error : in " << __FILE__ << " at " << __LINE__ << endl;
        return false;
    }
    
    if(!replanner->init())
    {
        cout << "Error : in " << __FILE__ << " at " << __LINE__ << endl;
        return false;
    }
    
    // If HERAKLES exists in the sceen, set it
    replanner->setHuman( human );
    
    if( global_costSpace == NULL )
    {
        cout << "Warning : costspace not initialized in " << __FILE__ << " at " << __LINE__ << endl;
        //return false;
    }
    
    bool RunShowTraj=true;
    bool StopRun=false;
    bool isFirstLoop=true;
    bool do_switch=false;
    
    int count=0;
    p3d_rob* robotPt;
    p3d_localpath* localpathPt;
    
    // Replanning variables
    double t=0.0, t_init = 0.0;
    double s = 0.0,  s_switch=0.0, s_max=0.0, s_rep=0.0;
    double lastFloor=0.0;
    double t_rep = PlanEnv->getDouble( PlanParam::trajReplanningWindow );
    double t_total = PlanEnv->getDouble( PlanParam::trajReplanningTotalTime );
    double initial_step = 0.0;
    
    // Retreive Current traj & q_switch from replanner
    confPtr_t q_switch;
    API::Trajectory& CurrentTraj = replanner->getCurrentTrajectory();
    initial_step = CurrentTraj.getRangeMax()/1;
    qEnd = CurrentTraj.getEnd();
    
    // Set the executed traj & multi-thread display mode
    setExecutedTrajectoryToCurrent(CurrentTraj);
    setMultithreadGraphical(true);
    
    // we want the trajectory to execute in 30 seconds
    // the replanning window is estimated with the param factor
    double paramFactor = CurrentTraj.getRangeMax() / t_total;
    s_rep = t_rep * paramFactor;
    s_max = CurrentTraj.getRangeMax();
    s_switch = s+s_rep;
    
    cout << "q switch id : " << switchId << endl;
    cout << "s_max : " << s_max << endl;
    cout << "s_switch : " << s_switch << endl;
    
    // Set data for switch in replanner
    replanner->setSwitchData( q_switch, 0, s_switch, t_rep, 0.0, initial_step );
    
    // Launch replanning thread
    boost::thread workerThread( boost::bind(&Replanner::run,replanner) );
    
    while ( !StopRun )
    { 
        // Switch to new trajectory
        if( s>s_switch )
        {
            if( replanner->isPlanningUndergo() && (!replanner->isPlanSucceeded()) )
            {
                cout << "replanner still planning or not succeeded!!!" << endl;
            }
            else 
            {
                if( setExecutedTrajectoryToCurrent(CurrentTraj) ) {
                    do_switch = true;
                    s_max = executedTraj.getRangeMax();
                }
                else {
                    cout << "Executed traj and current traj differ" << endl;
                    StopRun = true;
                }
            }
            
            // Launch new planning
            if( do_switch && ((s_max-s) > (5*paramFactor)) )
            {
                do_switch = false;
                StopRun = true;
                
                if ( replanner->isPlanSucceeded() )
                {
                    StopRun = false;
                    s_switch = s+s_rep;
                    cout << " ------------- timeSwitchAndId ------------- " << endl;
                    cout << "s : " << s << endl;
                    cout << "s_rep : " << s_rep << endl;
                    cout << "s_switch : " << s_switch << endl;
                    
                    if ( s < s_max-s_rep )
                    {          
                        // Set data for switch in replanner
                        replanner->setSwitchData( q_switch, 0, s_switch, t_rep, 0.0, initial_step );
                        
                        // Launch thread
                        boost::thread workerThread( boost::bind( &Replanner::run, replanner ) );
                    }
                    else
                    {
                        cout << "Last replanning!!!" << endl;
                    }
                }
            }
        }
        
        RunShowTraj = (*fct_stop)();
        
        // Timer using real time &  s is the parameter on the path
        t += timeSinceLastCall( isFirstLoop, t_init );
        s = t*paramFactor;
        robotSimulated->setAndUpdate( *executedTraj.configAtParam(s) );
        //cout << "s : " << s << endl;
        
        // Set and update human pos
        storeHumanPosition();
        
        if (fct) if (((*fct)(robotPt, localpathPt)) == FALSE) return(count);
        count++;
        
        if(s>s_max) {
            StopRun = true;
        }
        
        if (PlanEnv->getBool(PlanParam::stopPlanner)) {
            StopRun = true;
        }
        
        if (floor(t) > lastFloor) {
            cout << "t = " << t << endl;
            lastFloor = floor(t);
        }
    }
    
    //ReplanningSimulator::setMultithreadGraphical(false);
    return 0;
}
